/*
 * ActionsGUIInterface.C
 * John Davin
 *
 * Sends commands to the Actions display GUI through the message queue
 *
 */

#include <unistd.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <iostream.h>
#include "ActionsGUIInterface.h"


/*
 * Declare the message structure.
 */
typedef struct msgbuf {
    long    mtype;
    char    mtext[GUI_MSGSIZE];
} message_buf;


/*
 * Open the message queue
 *
 */
ActionsGUIInterface::ActionsGUIInterface(bool writelogs, bool usegui)
  : WriteLogs(writelogs), useGUI(usegui)
{
#ifdef USE_GUI
  if(!useGUI) return;

  int msgflg = IPC_CREAT | 0666;
  key_t key;


  /*
   * Get the message queue id for the
   * "name" 1234, which was created by
   * the Actions GUI.
   */
  key = 1234;

  if ((msgid = msgget(key, msgflg )) < 0) {
    perror("[ActionsGUIInterface]: failed to open message queue\n");
    return;
  }

  //cout << "[ActionsGUIInterface]: Message queue opened with id "<<msgid<<endl;

#endif
}

/*
 * The player uses this to set the playernum when it receives it from the server.
 * We will also open the logfile here (if applicable). 
 */
void ActionsGUIInterface::setPlayernum(int num){
#ifdef USE_GUI
  if(!useGUI) return;

  char logfilename[LOG_FILENAME_LEN];

  if(playernum >= 1 && playernum <= 11) // playernum is already set, just return. 
    return;

  if(num < 1 || num > 11){
    //cerr << "[ActionsGUIInterface:setPlayernum]: num "<<playernum<<" invalid."<<endl;
    return;
  }
  playernum = num;
  //cout <<"[ActionsGUIInterface:setPlayernum]: set playernum to "<<playernum<<endl;

  if(WriteLogs){ // open the log file
    snprintf(logfilename, LOG_FILENAME_LEN, "%s%d_%s", LOG_DIR, playernum, LOG_FILENAME);
    logfile.open(logfilename);
    
    if(! logfile.is_open()){
      cerr << "[ActionsGUIInterface]: failed to open logfile "<<logfilename<<endl;
      WriteLogs = false;
    }
  }

#endif
}



/*
 * Send a string to the message queue. 
 * Messages should be of the format: "%d %s %s %s message" where first %d is the playernum
 * the 1st %s action source, 2nd is foreground color, and the 3rd %s is background color. 
 */
void ActionsGUIInterface::send(int cyclenum, char *actionsource, char *msg, 
			       char *fgcolor, char *bgcolor, int pnum){
#ifdef USE_GUI
  if(!useGUI) return;

  char logline[GUI_MAX_LOG_LINE];
  int rtn;
  struct msgbuf buf;
  size_t buf_length;
  buf.mtype = 1;

  if(pnum != -1) // playernum is manually specified (probably from logplayer)
    playernum = pnum;

  if(playernum < 1 || playernum > 11){
    cerr << "[ActionsGUIInterface:send]: playernum not set yet: "<<playernum<<endl;
    return;
  }


  /* if colors were not set, set them to "-", which indicates default color. */
  if(fgcolor == NULL){
    fgcolor = new char[15];
    snprintf(fgcolor, 15, "-");
  }

  if(bgcolor == NULL){
    bgcolor = new char[15];
    snprintf(bgcolor, 15, "-");
  }

  rtn = snprintf(buf.mtext, GUI_MSGSIZE, "%2d %s %s %s %s", playernum, actionsource, 
		 fgcolor, bgcolor, msg); 
  if(rtn == -1){
    cerr << "[ActionsGUIInterface:send]: msg exceed max message size ("<<GUI_MSGSIZE<<")"<<endl;
    return;
  }
  buf_length = strlen(buf.mtext) + 1;
    
  /*
   * Send a message.
   */
  if (msgsnd(msgid, &buf, buf_length, IPC_NOWAIT) < 0) {
    errorcnt++;

    if(errorcnt >= 5){
      if(errorcnt == 5){
	cout << "Could not send to msg queue. Maybe Actions GUI is not started?"<<endl;
	errorcnt++;
      }
    }
    else{
      //printf ("%d: \"%s\", len %d\n", msgid, buf.mtext, buf_length);
      //perror("[ActionsGUIInterface:send]: msgsnd failed");
    }
  }
  else
    errorcnt = 0;
  //else 
  //printf("[ActionsGUIInterface]: Sent <%s> at %d\n", buf.mtext, (int)time(NULL));


  if(!(cyclenum >= 0 && cyclenum < GUI_MAX_CYCLES)){
    cerr << "[ActionsGUIInterface:send]: ERROR: cyclenum "<<cyclenum
	 <<" is outside legal range. Not writing to log."<<endl;
    return;
  }

  // write message to actions log
  if(WriteLogs  && logfile.is_open()){
    rtn = snprintf(logline, GUI_MAX_LOG_LINE, "%d %d %s %s %s %s",
	     cyclenum, playernum, actionsource, fgcolor, bgcolor, msg );
    if(rtn == -1){
      cerr << "[ActionsGUIInterface:send]: ERROR: logline exceed max log line."<<endl;
      return;
    }

    logfile << logline <<endl;
  }


#endif
}


ActionsGUIInterface::~ActionsGUIInterface(){
#ifdef USE_GUI
  if(WriteLogs){ // close the log file
    logfile.close();
  }
#endif
}
