/* Dongryeol Lee
 *
 * ActionPredicate.h
 *
 * This is the header file for the ActionPredicate class. Each object
 * represents a predicate that pertains to action properties. This class
 * is provided as a virtual base class for other action predicates to
 * extend from.
 *
 */


#ifndef ACTION_PREDICATE
#define ACTION_PREDICATE


#include "VarBindings.h"
#include "clangutil.h"
#include "clangaction.h"
#include "WorldStatePredicate.h"
#include "Geometry.h"

#include "Logger.h"

/* Our access to the player's world model */
extern WorldModelAccessor MyWorldModel;



/* ActionPredicate virtual base class */
class ActionPredicate
{
 public:

  /* Constructor */
  ActionPredicate(bool positive) 
    { positiveFlag = positive; }

  /* Virtual destructor */
  virtual ~ActionPredicate() {}


  /* boolean funciton for other classes to implement: tells whether
     the given action satisfies the action specific properties of the
     Predicate */
  virtual bool isSatisfiable(rcss::clang::Action *,
			     const VarBindings&  ) const = 0;
  
  
  /* Return whether the current predicate is positive or negative */
  bool getPositiveFlag() const { return positiveFlag; }
  
  
 private:
  
  /* Tells whether the Predicate P represents P or ~P. 
     True if the Predicate represents P
     False if the Predicate represents ~P */
  bool positiveFlag;
  
};



/* Now onto some classes extending from the virtual base class, ActionPredicate
   .............. */



/* ForwardTeammate: True if the player in consideration is a teammate in the 
   forward position with respect the agent */
class ForwardTeammate: public ActionPredicate
{
 public:

  /* Constructor */
  ForwardTeammate(bool positive): ActionPredicate(positive) {}

  
  /* This predicate is involved when the action involves our teammate */
  bool isSatisfiable(rcss::clang::Action *act,
		     const VarBindings & ) const
    {

      using namespace rcss::clang;

      /* Passing to our player is the only action involved with this 
	 predicate */
      ActPassUNum *theact = (ActPassUNum *) act;


      /* Get the lone player in the UNumSet */
      UNumSet players = theact->getPlayers();


      /* Point to the beginning of the UNumSet to retrieve our player's UNum */
      UNum::unum_t our = (*(players.begin())).getUNum();

      
      /* First get the location of the agent */
      VecPosition agentPos = MyWorldModel.getAgentGlobalPosition();


      /* Retrieve the location of the teammate */
      VecPosition ourPos = MyWorldModel.getGlobalPosition(our, true);


      if( agentPos.getX() < ourPos.getX() )
	actionlog(40) << "Teammate is foward position\n";
      else
	actionlog(40) << "Teammate is nonforward position\n";



      /* And compare their X coordinates */
      return agentPos.getX() < ourPos.getX();
      
    }
};


/* ForwardOpponent: True if the player in consideration is a opponent in the 
   forward position with respect the agent. Use templated class since there
   are three actions involving opponents: mark, markl, tackle */
template <class T>
class ForwardOpponent: public ActionPredicate
{
 public:

  /* Constructor */
  ForwardOpponent(bool positive): ActionPredicate(positive) { }


  /* This predicate will be included when mark and tackle actions are involved
     (actions involving opponent UNums) */
  bool isSatisfiable(rcss::clang::Action *act,
		     const VarBindings &) const
    {
      using namespace rcss::clang;

      /* Type cast to appropriate action type */
      T *theact = (T *) act;

      /* Get the lone player in the UNumSet */
      UNumSet players = theact->getPlayers();

      /* Point to the beginning of the UNumSet to retrieve the passe's UNum */
      UNum::unum_t opp = (*(players.begin())).getUNum();

      
      /* First get the location of the agent */
      VecPosition agentPos = MyWorldModel.getAgentGlobalPosition();


      /* Retrieve the location of the opponent */
      VecPosition oppPos = MyWorldModel.getGlobalPosition(opp, false);

      /* And compare their X coordinates */
      return agentPos.getX() < oppPos.getX();
      
    }
};


/* KnowsTeammatePos: True if the agent knows the location of its teammate */
class KnowsTeammatePos: public ActionPredicate
{
 public:

  /* Constructor */
  KnowsTeammatePos(bool positive): ActionPredicate(positive) {}


  /* Test satisfiability on the current predicate: Don't have to check types of
     actions here as well since it will all involve UNums */
  bool isSatisfiable(rcss::clang::Action *act,
		     const VarBindings &) const
    {
      using namespace rcss::clang;

      /* Type cast to PassUNum action */
      ActPassUNum *theact = (ActPassUNum *) act;


      /* Get the lone player in the UNumSet */
      UNumSet players = theact->getPlayers();

      /* Point to the beginning of the UNumSet to retrieve the passe's UNum */
      UNum::unum_t passee = (*(players.begin())).getUNum();
     
      /* Just check if the agent is confident in the location of our
	 player */
      /* jdavin: But if playmode is not PLAY_ON (such as a kick_in), then it 
       * doesn't matter if we know where the teammate is because 
       * we have time to look for him (the pass action in CoachablePlayer will do this).
       * This is to fix a bug where players are not doing setplay passes b/c they 
       * lost sight of the passee. 
       */
      if(MyWorldModel.getPlayMode() != PlayOn)
	return true;


      if( MyWorldModel.playerConfidenceGood( passee, true ) )
	actionlog(30) << "KnowsTeammatePos of " << (int) passee << spades::ende;
      else
	actionlog(30) << "Doesn't know where teammate is: " << (int) passee <<
	  spades::ende;


      return MyWorldModel.playerConfidenceGood(passee, true);
    }
};


/* KnowsOpponentPos: True if the agent knows the location of the opponent */
template <class T>
class KnowsOpponentPos: public ActionPredicate
{
 public:

  /* Constructor */
  KnowsOpponentPos(bool positive): ActionPredicate(positive) {}


  /* Test satisfiability on the current predicate: Don't have to check types of
     actions here as well since it will all involve UNums */
  bool isSatisfiable(rcss::clang::Action *act,
		     const VarBindings &) const
    {
      using namespace rcss::clang;


      /* Type cast to appropriate action type */
      T *theact = (T *) act;

      /* Get the lone player in the UNumSet */
      UNumSet players = theact->getPlayers();

      /* Point to the beginning of the UNumSet to retrieve the passe's UNum */
      UNum::unum_t our = (*(players.begin())).getUNum();
     
      if( MyWorldModel.playerConfidenceGood( our, false ) )
	actionlog(30) << "KnowsOpponentPos of " << (int) our << spades::ende;
      else
	actionlog(30) << "Doesn't know where opponent is: " << (int) our <<
	  spades::ende;


      /* Just check if the agent's confidence in the location of the 
	 opponent */
      return MyWorldModel.playerConfidenceGood(our, false);
    }
};


/* OpponentCanKickBall: True if the opponent in the UNumSet can kick the ball
   at this instant. */
template <class T>
class OpponentCanKickBall: public ActionPredicate
{
 public:

  /* Constructor */
  OpponentCanKickBall(bool positive): ActionPredicate(positive) {}


  /* Test satisfiability on the current predicate: Don't have to check types of
     actions here as well since it will all involve UNums */
  bool isSatisfiable(rcss::clang::Action *act,
		     const VarBindings &) const
    {
      using namespace rcss::clang;


      /* Type cast to appropriate action type */
      T *theact = (T *) act;

      /* Get the lone player in the UNumSet */
      UNumSet players = act->getPlayers();

      /* Point to the beginning of the UNumSet to retrieve the passe's UNum */
      UNum::unum_t our = (*(players.begin())).getUNum();
     
      /* Just check if the ball can be kicked by the opponent */
      return MyWorldModel.isBallKickableByPlayer(our, false);
    }
};


/* RegionUncongested: Calculates the opponent density of the region and
   determines whether it is below some threshold ( players / m^2). True if it
   does, false otherwise */
class RegionUncongested: public ActionPredicate
{

 public:

  /* Constructor */
  RegionUncongested(bool positive): ActionPredicate(positive) {}


  /* Test satisfiability on the current predicate: Don't have to check types of
     actions here as well since it will all involve UNums */
  bool isSatisfiable(rcss::clang::Action *act,
		     const VarBindings &v) const
    {
      using namespace rcss::clang;

      /* REGION_UNCONGESTED macro is based on the 4m * 4m square with only two
	 players in it. 2 players / 16 m^2 = 0.125 players / m^2. I think more
	 than 2 will be a problem */

      #ifndef REGION_UNCONGESTED
      #define REGION_UNCONGESTED 0.125
      #endif

      /* Type cast to appropriate action type */
      ActDribble *theact = (ActDribble *) act;
      
      /* Get the region in consideration */
      Region *reg = (Region *) theact->getRegion();

      
      /* Opponent count in the region */
      int count = 0;

      /* Count up the number of opponents in the region */
      for(UNum u = UNum::u1; u.getUNum() <= UNum::u11; u++){

	/* Retrieve the UNum of the current teammate */
	UNum::unum_t currUNum = u.getUNum();


	/* Check the confidence of the location of the opponents we want to
	   check first.... If it is low, then don't bother with computation */
	if( !MyWorldModel.playerConfidenceGood(currUNum, false) )
	  continue;


	/* Retrieve the position of the teammate */
	VecPosition pos = MyWorldModel.getGlobalPosition(currUNum, true);

	/* Check if the opponent is in the region */
	if( reg->isPtIn(pos, &MyWorldModel, v) )
	  count++;

      }   /* End of for-loop */
   

      /* Check if the opponent density is lower than the area threshold */
      return (reg->getArea(&MyWorldModel, v) / (double) count) 
	< REGION_UNCONGESTED;
    }

};


/* GoodPass: the Predicate is true if there is no opponent in the arc region
   defined by the ball as the center and the radius of (the ball to the
   passee + 1.0) meters, with the angle of ballToPasseeAngle - 30.0 ~
   ballToPasseeAngle + 30.0.
   Also, the pass must have at least 5 units distance between the ball and passee.
*/
class GoodPass: public ActionPredicate
{
 public:

  /* Constructor */
  GoodPass(bool positive): ActionPredicate(positive) {}
  
  
  /* Test satisfiability on the current predicate: Don't have to check types of
     actions here as well since it will all involve UNums */
  bool isSatisfiable(rcss::clang::Action *act,
		     const VarBindings &v) const
    {
      using namespace rcss::clang;

      /* Change the cone angle width for defining the arc region */
      #ifndef CONE_ANGLE
      #define CONE_ANGLE 60.0
      #endif

      // the distance for which a pass is considered to be too short
      #ifndef CLOSE_PASS_DIST
      #define CLOSE_PASS_DIST 5.0
      #endif

      /* Typecast to ActPassUNum action */
      ActPassUNum *theact = (ActPassUNum *) act;

      
      /* Get the lone player in the UNumSet */
      UNumSet players = theact->getPlayers();

      /* Point to the beginning of the UNumSet to retrieve the passe's UNum */
      UNum::unum_t our = (*(players.begin())).getUNum();

      
      /* Get the ball position */
      VecPosition ballPos = MyWorldModel.getBallPos();

      /* Get the position of the passee */
      VecPosition passeePos = MyWorldModel.getGlobalPosition(our, true);


      /* Calculate the vector from the ball to the passee */
      VecPosition ballToPassee = passeePos - ballPos;


      /* Calculate the angle direction of the vector */
      double startAngle = ballToPassee.getDirection() - CONE_ANGLE / 2.0;


      /* Calculate the magnitude of the vector */
      double magnitude = ballToPassee.getMagnitude();

      // this pass is too short to be considered a good pass. 
      if(magnitude < CLOSE_PASS_DIST)
	return false;

      /* Construct a RegArc for testing */
      RegArc *reg = new RegArc(std::auto_ptr<Point>
			       (new PointSimple(ballPos.getX(),
						ballPos.getY())),
			       0.0, magnitude + 1.0,
			       //startAngle, 60.0);
			       startAngle, CONE_ANGLE/2.);


      /* Count how many opponent players are within this arc region */
      int count = 0;

      VecPosition pos;

      /* Count up the number of opponents in the region */
      for(UNum u = UNum::u1; u.getUNum() <= UNum::u11; u++){

	/* Retrieve the UNum of the current teammate */
	UNum::unum_t currUNum = u.getUNum();


	/* Check the confidence of the location of the opponents we want to
	   check first.... If it is low, then don't bother with computation */
	if( !MyWorldModel.playerConfidenceGood(currUNum, false) )
	  continue;


	/* Retrieve the position of the opponent */
	pos = MyWorldModel.getGlobalPosition(currUNum, false);

	/* Check if the opponent is in the region */
	if( reg->isPtIn(pos, &MyWorldModel, v) )
	  count++;

      }   /* End of for-loop */
         

      /* Delete the arc region to prevent memory leak */
      delete reg;
      

      if(count == 0)
	actionlog(30) << "Good pass."<<spades::ende;
      else 
	actionlog(30) << "Not Good pass."<<spades::ende;

      /* Check if the count = 0 (no one is in the cone region) */
      return count == 0;
    }
};


/* NotClosePass: the Predicate is true if the pass is at least 5 units 
 * distance between the ball and passee.
 */
class NotClosePass: public ActionPredicate
{
 public:

  NotClosePass(bool positive): ActionPredicate(positive) {}
  
  bool isSatisfiable(rcss::clang::Action *act,
		     const VarBindings &) const
    {
      using namespace rcss::clang;

      ActPassUNum *theact = (ActPassUNum *) act;
      
      /* Get the lone player in the UNumSet */
      UNumSet players = theact->getPlayers();

      /* Point to the beginning of the UNumSet to retrieve the passe's UNum */
      UNum::unum_t our = (*(players.begin())).getUNum();
      
      /* Get the ball position */
      VecPosition ballPos = MyWorldModel.getBallPos();

      /* Get the position of the passee */
      VecPosition passeePos = MyWorldModel.getGlobalPosition(our, true);

      /* Calculate the vector from the ball to the passee */
      VecPosition ballToPassee = passeePos - ballPos;

      /* Calculate the magnitude of the vector */
      double magnitude = ballToPassee.getMagnitude();

      // the pass is not a close pass 
      if(magnitude > CLOSE_PASS_DIST)
	return true;
      else
	return false;
      return false;
    }
};


/* 
   OpponentIsClosestToAgent: True if the opponent in the UNumSet is the
   closest one to the agent, false otherwise
*/
template <class T>
class OpponentIsClosestToAgent: public ActionPredicate
{
 public:

  /* Constructor */
  OpponentIsClosestToAgent(bool positive): ActionPredicate(positive) {}
  
  
  /* Test satisfiability on the current predicate: Don't have to check types of
     actions here as well since it will all involve UNums */
  bool isSatisfiable(rcss::clang::Action *act,
		     const VarBindings &) const
    {
      using namespace rcss::clang;

      /* Type cast to appropriate action type */
      T *theact = (T *) act;

      /* Get the lone player in the UNumSet */
      UNumSet players = theact->getPlayers();

      /* Point to the beginning of the UNumSet to retrieve the 
	 oppponent's UNum */
      UNum::unum_t opp = (*(players.begin())).getUNum();
     

      /* Retrieve the agent position */
      VecPosition agentPos = MyWorldModel.getAgentGlobalPosition();


      /* Closest distance found so far */
      double closestDist = 99999;
      
      /* UNum of the closest opponent */
      UNum::unum_t closestUNum;


      /* Position of the current opponent in consideration */
      VecPosition pos;

      /* Distance from the current opponent to the ball */
      double dist;


      /* Go through each opponent */
      for(UNum u = UNum::u1; u.getUNum() <= UNum::u11; u++){

	/* Retrieve the UNum of the current teammate */
	UNum::unum_t currUNum = u.getUNum();


	/* Check the confidence of the location of the opponent we want to
	   check first.... If it is low, then don't bother with computation */
	if( !MyWorldModel.playerConfidenceGood(currUNum, false) )
	  continue;


	/* Retrieve the position of the opponent */
	pos = MyWorldModel.getGlobalPosition(currUNum, false);

	/* Calculate the distance from the agent to the opponent */
	dist = pos.getDistanceTo(agentPos);

	
	/* Replace the result if better than previous */
	if(dist < closestDist) {

	  closestDist = dist;
	  closestUNum = currUNum;
	}

      }   /* End of for-loop */

      
      /* Get the boolean result of the positive
	 Just check the UNum of the closest player equals the opponent UNum */
      return opp == closestUNum;
   
    }  
};


/* OpponentCanKickBall: True if the opponent in the UNumSet can kick the ball
   at this instant. */
template <class T>
class OpponentCloseToBall: public ActionPredicate
{
 public:

  /* Constructor */
  OpponentCloseToBall(bool positive): ActionPredicate(positive) {}


  /* Test satisfiability on the current predicate: Don't have to check types of
     actions here as well since it will all involve UNums */
  bool isSatisfiable(rcss::clang::Action *act,
		     const VarBindings &) const
    {
      using namespace rcss::clang;

      /* Currently, closeness is defined to be 5 m */
      #ifndef OPPONENT_CLOSE_TO_BALL
      #define OPPONENT_CLOSE_TO_BALL  5.0
      #endif


      /* Type cast to appropriate action type */
      T *theact = (T *) act;

      /* Get the lone player in the UNumSet */
      UNumSet players = theact->getPlayers();

      /* Point to the beginning of the UNumSet to retrieve the opponent UNum */
      UNum::unum_t opp = (*(players.begin())).getUNum();
     

      /* First check the confidence values of the location of the opponent
	 and the ball. If either of them has low confidence value, return 
	 false (can't be determined)
      */
      if( !MyWorldModel.ballConfidenceGood() || 
	  !MyWorldModel.playerConfidenceGood( opp, false ) )
	return false;


      /* Otherwise, check the ball location and the opponent location */
      VecPosition ballPos = MyWorldModel.getBallPos();
      VecPosition oppPos = MyWorldModel.getGlobalPosition( opp, false );


      /* Just check if the ball to opponent distance is within the
         threshold */
      return ballPos.getDistanceTo(oppPos) < OPPONENT_CLOSE_TO_BALL;
    }
};


/* OpponentIsMarkedByTeammate: the Predicate is true if there is at least one 
   teammate "marking the opponent"
*/
class OpponentIsMarkedByTeammate: public ActionPredicate
{
 public:

  /* Constructor */
  OpponentIsMarkedByTeammate(bool positive): ActionPredicate(positive) {}
  
  
  /* Test satisfiability on the current predicate: Don't have to check types of
     actions here as well since it will all involve UNums */
  bool isSatisfiable(rcss::clang::Action *act, const VarBindings &) const
    {
      using namespace rcss::clang;

      /* Change the cone angle width for defining the arc region */
      #ifndef MARK_RADIUS
      #define MARK_RADIUS   4.5
      #endif



      /* TypeCast */
      ActMark *theact = (ActMark *) act;

      
      /* Get the lone player in the UNumSet */
      UNumSet players = theact->getPlayers();

      /* Point to the beginning of the UNumSet to retrieve the passe's UNum */
      UNum::unum_t opp = (*(players.begin())).getUNum();

      
      /* Get the position of the opponent */
      VecPosition oppPos = MyWorldModel.getGlobalPosition(opp, false);


      /* Check the confidence value of the ball; if the agent is not confident
	 about it, then we can assume the opponent is not marked */
      if( !MyWorldModel.ballConfidenceGood() )
	return false;


      /* Get the position of the ball */
      VecPosition ballPos = MyWorldModel.getBallPos();


      /* Determine if any of our teammates is marking the opponent */
      for(UNum u = UNum::u1; u.getUNum() <= UNum::u11; u++){

	/* Retrieve the UNum of the current teammate */
	UNum::unum_t currUNum = u.getUNum();


	/* Check the confidence of the location of the teammates we want to
	   check first.... If it is low, then don't bother with computation */
	if( !MyWorldModel.playerConfidenceGood(currUNum, true) )
	  continue;


	/* Retrieve the position of the teammate */
	VecPosition pos = MyWorldModel.getGlobalPosition(currUNum, true);


	/* Definition of marking:
	   
	   1. The teammate is close to the line made from the ball to the
	   opponent.
	   2. On the line, the teammate is within some radius of a distance
	*/

	/* Check if the current teammate meets this criteria */
	if( !(currUNum == MyWorldModel.getMyNum()) ) {

	  /* Construct a line from the ball to the opponent */
	  Line l = Line::makeLineFromTwoPoints( ballPos, oppPos );
	  
	  
	  /* Check criterion number 1, if the current teammate is within
	     50 cm on the marking line
	     
	     Check criterion number 2, if the current teammate is within
	     MARK_RADIUS m of the opponent
	  */

	  if( l.getDistanceWithPoint(pos) < 0.5 &&
	      pos.getDistanceTo( oppPos ) < MARK_RADIUS){
	    actionlog(30) << "OpponentIsMarkedByTeammate = true" <<spades::ende;
	    return true;
	  }
	}

      }   /* End of for-loop */


      actionlog(30) << "OpponentIsMarkedByTeammate = false " <<spades::ende;

      /* If you are here, no teammate is marking the opponent */
      return false;
    }
};


/* OpponentIsMarkedLineByTeammate: the Predicate is true if there is at least 
   one teammate "markline"ing the opponent
*/
class OpponentIsMarkedLineByTeammate: public ActionPredicate
{
 public:

  /* Constructor */
  OpponentIsMarkedLineByTeammate(bool positive): ActionPredicate(positive) {}
  
  
  /* Test satisfiability on the current predicate: Don't have to check types of
     actions here as well since it will all involve UNums */
  bool isSatisfiable(rcss::clang::Action *act, const VarBindings &) const
    {
      using namespace rcss::clang;


      /* TypeCast */
      ActMarkLinePlayer *theact = (ActMarkLinePlayer *) act;

      
      /* Get the lone player in the UNumSet */
      UNumSet players = theact->getPlayers();

      /* Point to the beginning of the UNumSet to retrieve the passe's UNum */
      UNum::unum_t opp = (*(players.begin())).getUNum();

      
      /* Get the position of the opponent */
      VecPosition oppPos = MyWorldModel.getGlobalPosition(opp, false);



      /* Check the confidence value of the ball; if the agent is not confident
	 about it, then we can assume the opponent is not marked */
      if( !MyWorldModel.ballConfidenceGood() )
	return false;



      /* Get the position of the ball */
      VecPosition ballPos = MyWorldModel.getBallPos();


      /* Determine if any of our teammates is marking the opponent */
      for(UNum u = UNum::u1; u.getUNum() <= UNum::u11; u++){

	/* Retrieve the UNum of the current teammate */
	UNum::unum_t currUNum = u.getUNum();


	/* Check the confidence of the location of the teammates we want to
	   check first.... If it is low, then don't bother with computation */
	if( !MyWorldModel.playerConfidenceGood(currUNum, true) )
	  continue;


	/* Retrieve the position of the teammate */
	VecPosition pos = MyWorldModel.getGlobalPosition(currUNum, true);


	/* Definition of marking the passing lane:
	   
	   1. The teammate is close to the line made from the ball to the
	   opponent.
	*/

	/* Check if the current teammate meets this criteria */
	if( !(currUNum == MyWorldModel.getMyNum()) ) {

	  /* Construct a line from the ball to the opponent */
	  Line l = Line::makeLineFromTwoPoints( ballPos, oppPos );
	  
	  
	  /* Check criterion number 1, if the current teammate is within
	     50 cm on the marking line	     
	  */

	  if( l.getDistanceWithPoint(pos) < 0.5 )
	    return true;
	}

      }   /* End of for-loop */
         

      /* If you are here, no teammate is marking the opponent */
      return false;
    }
};


/* IsBallOwner: True if the agent is the ball owner */
class IsBallOwner: public ActionPredicate
{
 public:

  /* Constructor */
  IsBallOwner(bool positive): ActionPredicate(positive) {}


  /* Test satisfiability on the current predicate: Don't have to check types of
     actions here as well since it will all involve UNums */
  bool isSatisfiable(rcss::clang::Action *,
		     const VarBindings &) const
    {
      rcss::clang::UNum::unum_t player;
      bool team = MyWorldModel.getBallOwner(&player);
      if(team == false) // opponent team
	return false;

      if( player == MyWorldModel.getMyNum() ){
	actionlog(30) << "Is ball owner" << spades::ende;
	return true;
      }
      else {
	actionlog(30) << "Is not ball owner" << spades::ende;
	return false;
      }

    }
};


/* BallIsKickable: True if the ball is kickable */
class BallIsKickable: public ActionPredicate
{
 public:

  /* Constructor */
  BallIsKickable(bool positive): ActionPredicate(positive) {}


  /* Test satisfiability on the current predicate: */
  bool isSatisfiable(rcss::clang::Action *,
		     const VarBindings &) const
    {
      if( MyWorldModel.isBallKickable() ){
	actionlog(30) << "Ball is kickable" << spades::ende;
	return true;
      }
      else {
	actionlog(30) << "Ball is not kickable" << spades::ende;
	return false;
      }

    }
};





#endif
