/*
Copyright (c) 2000-2002, Jelle Kok, University of Amsterdam
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
list of conditions and the following disclaimer. 

2. Redistributions in binary form must reproduce the above copyright notice, 
this list of conditions and the following disclaimer in the documentation 
and/or other materials provided with the distribution. 

3. Neither the name of the University of Amsterdam nor the names of its 
contributors may be used to endorse or promote products derived from this 
software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE 
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
/*! \file BasicPlayer.C
<pre>
<b>File:</b>          BasicPlayer.C
<b>Project:</b>       Robocup Soccer Simulation Team: UvA Trilearn
<b>Authors:</b>       Jelle Kok
<b>Created:</b>       10/12/2000
<b>Last Revision:</b> $ID$
<b>Contents:</b>      This file contains the class declaration for the
               BasicPlayer. The BasicPlayer is the class where the
               available skills for the agent are defined.
<hr size=2>
<h2><b>Changes</b></h2>
<b>Date</b>             <b>Author</b>          <b>Comment</b>:
10/12/2000       Jelle Kok       Initial version created
</pre>
*/

#include "BasicPlayer.h"
#include "Parse.h"        // parseFirstInt



#include "Logger.h"



/********************** LOW-LEVEL SKILLS *************************************/

/*! This skill enables an agent to align his neck with his body. It returns a
    turn neck command that takes the angle of the agent's body relative to his
    neck as its only argument.
   \return SoccerCommand turn_neck command that aligns neck with body */
SoccerCommand BasicPlayer::alignNeckWithBody( )
{
  return SoccerCommand( CMD_TURNNECK, WM->getAgentBodyAngleRelToNeck( ) );
}

/*! This skill enables an agent to turn his body towards a given point. It
    receives a global position 'pos' on the field and returns a turn command
    that will turn the agent's body towards this point. To this end the agent's
    global position in the next cycle is predicted based on his current
    velocity. This is done to compensate for the fact that the remaining
    velocity will move the agent to another position in the next cycle. The
    global angle between the given position and the predicted position is then
    determined after which the agent's global body direction is subtracted from
    this angle in order to make it relative to the agent's body. Finally,
    the resulting angle is normalized and adjusted to compensate for the
    inertia moment and speed of the agent. If it is impossible to turn towards
    the given position in a single cycle then the agent turns as far as
    possible.
    \param pos position to which body should be turned
    \param iCycles denotes the number of cycles that are used to update the
               the agent position. The resulting position is compared with
               'pos' to determine the desired turning angle.
    \return SoccerCommand turn command to turn body to the desired point */
SoccerCommand BasicPlayer::turnBodyToPoint( VecPosition pos, int iCycles )
{
  VecPosition posGlobal = WM->predictAgentPos(iCycles, 0);
  AngDeg angTurn        = (pos - posGlobal).getDirection();
  angTurn              -= WM->getAgentGlobalBodyAngle();
  angTurn               = VecPosition::normalizeAngle( angTurn );
  angTurn               = WM->getAngleForTurn( angTurn, WM->getAgentSpeed() );

  return SoccerCommand( CMD_TURN, angTurn );
}

/*! This skill enables an agent to turn his back towards a given point 'pos'.
    The only difference between this skill and turnBodyToPoint is that the
    angle between the given position and the predicted position of the agent
    in the next cycle is now made relative to the back of the agent by
    subtracting the agent's global back direction. This skill can for example
    be used by the goalkeeper in case he wants to move back to his goal while
    keeping sight of the rest of the field.
    \param pos position to which the agent's back should be turned
    \param iCycles denotes the number of cycles that are used to update the
               the agent position. The resulting position is compared with
               'pos' to determine the desired turning angle.
    \return SoccerCommand command to turn agent's back to the desired point */
SoccerCommand BasicPlayer::turnBackToPoint( VecPosition pos, int iCycles )
{
  VecPosition posGlobal = WM->predictAgentPos(iCycles, 0);
  AngDeg angTurn        = (pos - posGlobal).getDirection();
  angTurn              -= (WM->getAgentGlobalBodyAngle() + 180);
  angTurn               = VecPosition::normalizeAngle( angTurn );
  angTurn               = WM->getAngleForTurn( angTurn, WM->getAgentSpeed() );

  return SoccerCommand( CMD_TURN, angTurn );
}

/*! This skill enables an agent to turn his neck towards a given point. It
    receives a global position 'pos' on the field as well as a primary action
    command 'soc' that will be executed by the agent at the end of the current
    cycle and returns a turn neck command that will turn the agent's neck
    towards 'pos'. To this end the agent's global position and neck direction
    after executing the cmd command are predicted using methods from the world
    model. The global angle between the given position and the predicted
    position is then determined after which the predicted neck direction is
    subtracted from this angle in order to make it relative to the agent's
    neck. Finally, the resulting angle is normalized and directly passed as an
    argument to the turn neck command since the actual angle with which a
    player turns his neck is by definition equal to this argument.
    If the resulting turn angle causes the absolute angle between
    the agent's neck and body to exceed the maximum value, then the agent turns
    his neck as far as possible. Note that it is necessary to supply the
    selected primary command as an argument to this skill, since a turn neck
    command can be executed in the same cycle as a kick, dash, turn , move
    or catch command.
   \param pos position to which neck should be turned
   \param soc SoccerCommand that is executed in the same cycle
   \return SoccerCommand turn command to turn neck to the desired point */
SoccerCommand BasicPlayer::turnNeckToPoint(VecPosition pos, SoccerCommand soc)
{
  VecPosition posMe,   velMe;
  AngDeg      angBody, angNeck, angActual;
  Stamina     sta;

  // predict agent information after command 'soc' is performed
  // calculate the desired global angle of the neck
  // calculate the desired angle of the neck relative to the body
  WM->predictAgentStateAfterCommand(soc,&posMe,&velMe,&angBody,&angNeck,&sta);
  AngDeg angDesGlobNeck   = (pos - posMe).getDirection();
  AngDeg angNeckRelToBody = VecPosition::normalizeAngle(angDesGlobNeck-angBody);

  // calculate the current angle of the body relative to the neck
  // check if the desired neck angle relative to the body is possible:
  // if angle is smaller than the minimum or larger than the maximum neck angle
  //  turn neck to the minimum or maximum neck angle + the current neck angle
  // else calculate the desired angle relative to the body
  AngDeg angBodyRelToNeck = VecPosition::normalizeAngle(angBody-angNeck);
  if( angNeckRelToBody < SS->getMinNeckAng() )
    angActual = SS->getMinNeckAng() + angBodyRelToNeck;
  else if( angNeckRelToBody > SS->getMaxNeckAng() )
    angActual = SS->getMaxNeckAng() + angBodyRelToNeck;
  else
    angActual = angNeckRelToBody + angBodyRelToNeck;
  return SoccerCommand( CMD_TURNNECK, angActual );
}

/*! This skill enables an agent to search for the ball when he cannot see it.
    It returns a turn command that causes the agent to turn his body by an
    angle that equals the width of his current view cone (denoted by the
    ViewAngle attribute in the AgentObject class). In this way the agent will
    see an entirely different part of the field after the turn which maximizes
    the chance that he will see the ball in the next cycle. Note that the
    agent turns towards the direction in which the ball was last observed to
    avoid turning back and forth without ever seeing the ball. Furthermore
    the inertia moment of the agent is taken into account to compensate for
    the current speed of the agent.
    \return SoccerCommand that searches for the ball. */
SoccerCommand BasicPlayer::searchBall()
{
  static Time timeLastSearch;
  static int  iSign       = 1;
  
  //  if(WM->isConfidenceGood(OBJECT_BALL))
  //  assert(0);

  //  cout << "["<<WM->getCurrentTime().getTime()<<"]: {"<<WM->getPlayerNumber()<<"}: "<<"Searching for ball"<<endl;


  // calculate maximum turn angle
  AngDeg ang = WM->getAgentViewAngleValue( WM->getAgentViewAngle() );
  ang        = WM->getAngleForTurn( ang - ang/10.0, WM->getAgentSpeed() );

  // calculate whether expected ball position is too the left or right of body
  // do this by making angle in directon body and back of body and check
  // whether estimated ball prediction lies between these two angles
  // only change sign when not searched in previous cycles, otherwise could
  // turn left, right, left, right and never find ball
  AngDeg angBody = WM->getAgentGlobalBodyAngle();
  AngDeg angMax  = VecPosition::normalizeAngle( angBody + 180 );
  AngDeg angBall = (WM->getBallPos()-
                                  WM->getAgentGlobalPosition()).getDirection();

  if( WM->getCurrentTime() - timeLastSearch  > 3 )
    iSign = isAngInInterval( angBall, angBody, angMax ) ? 1: -1;
  timeLastSearch = WM->getCurrentTime();

  //  return SoccerCommand( CMD_TURN, iSign * ang );
  /*   -jdavin, 9/28/03 */
    return (isTimeToTurn())
    ? SoccerCommand( CMD_TURN, iSign * ang )
    : SoccerCommand( CMD_TURN , 1 );

}

SoccerCommand
BasicPlayer::scanWithNeck()
{
  //cout << "["<<WM->getPlayerNumber()<<"]: "<<"Scanning with neck"<<endl;

  /* first determine our scan ang, we'll go right as long as we can, then flip
     back to the left */
  AngDeg scan_ang = WM->getAgentViewAngleValue( WM->getAgentViewAngle() ) - 5.0;
  /* the first minus is because this is body to neck, not neck to body */
  AngDeg curr_neck_ang = -WM->getAgentBodyAngleRelToNeck();
  AngDeg future_neck_ang = curr_neck_ang + scan_ang;
  if (future_neck_ang > SS->getMaxNeckAng())
    {
      if (curr_neck_ang < SS->getMaxNeckAng() - 5.0)
	scan_ang = SS->getMaxNeckAng() - curr_neck_ang;
      else
	scan_ang = SS->getMinNeckAng() - curr_neck_ang;
    }

  //cout << WM->getCurrentTime() << ": scan_ang=" << scan_ang << endl;

  /* return SoccerCommand( CMD_TURNNECK, scan_ang );
      -jdavin, 9/28/03*/
  return (isTimeToTurn())
    ? SoccerCommand( CMD_TURNNECK, scan_ang )
    : SoccerCommand( CMD_TURNNECK , 0.0 );
  
}


/*! This method can be called to create a SoccerCommand that dashes to a point.
    This skill enables an agent to dash to a given point. It receives a global
    position 'pos' as its only argument and returns a dash command that causes
    the agent to come as close to this point as possible. Since the agent can
    only move forwards or backwards, the closest point to the target position
    that he can reach by dashing is the orthogonal projection of 'pos' onto the
    line that extends into the direction of his body (forwards and backwards).
    The power that must be supplied to the dash command is computed using the
    'getPowerForDash' method which takes the position of 'pos' relative to the
    agent as input.
    \param pos global position to which the agent wants to dash
    \return SoccerCommand dash command to move closer to 'pos' */
SoccerCommand BasicPlayer::dashToPoint( VecPosition pos )
{
  double dDashPower = WM->getPowerForDash(
                                 pos - WM->getAgentGlobalPosition(),
                                 WM->getAgentGlobalBodyAngle(),
                                 WM->getAgentGlobalVelocity(),
                                 WM->getAgentEffort()                );
  return SoccerCommand( CMD_DASH, dDashPower );

}


/*! This method can be called to create a SoccerCommand that tackles at the ball.
    \param power to tackle
    \return SoccerCommand tackle command to tackle */
SoccerCommand BasicPlayer::tackleBall( double power )
{
  return SoccerCommand( CMD_TACKLE, power );
}




/*! This skill enables an agent to freeze a moving ball, i.e. it returns a kick
    command that stops the ball dead at its current position. Since ball
    movement in the soccer server is implemented as a vector addition, the ball
    will stop in the next cycle when it is kicked in such a way that the
    resulting acceleration vector has the same length and opposite direction to
    the current ball velocity. The desired speed that should be given to the
    ball on the kick thus equals the current ball speed. Furthermore, the
    direction of the kick should equal the direction of the current ball
    velocity plus 180 degrees. Note that this direction must be made relative
    to the agent's global body angle before it can be passed as an argument to
    the kick command.
    \return SoccerCommand to freeze the ball. */
SoccerCommand BasicPlayer::freezeBall( )
{
  // determine power needed to kick the ball to compensate for current speed
  // get opposite direction (current direction + 180) relative to body
  // and make the kick command.

  double dPower = WM->getKickPowerForSpeed( WM->getBallSpeed() );
  if( dPower > SS->getMaxPower() )
  {
    Log.log( 552, "%d: freeze ball has to much power", WM->getCurrentCycle() );
    dPower = (double)SS->getMaxPower();
  }
  double dAngle = WM->getBallDirection() + 180 - WM->getAgentGlobalBodyAngle();
  dAngle        = VecPosition::normalizeAngle( dAngle );

      
  return SoccerCommand( CMD_KICK, dPower, dAngle );
}

/*! This skill enables an agent to kick the ball close to his body. It receives
    an angle 'ang' as its only argument and returns a kick command that causes
    the ball to move to a point at a relative angle of 'ang' degrees and at a
    close distance (kickable margin/6 to be precise) from the agent's body.
    To this end the ball has to be kicked from its current position to the
    desired point relative to the predicted position of the agent in the next
    cycle. In general, this skill will be used when the agent wants to kick the
    ball to a certain position on the field which cannot be reached with a
    single kick. Since the efficiency of a kick is highest when the ball is
    positioned just in front of the agent's body, calling this skill with
    'ang = 0' will have the effect that the agent can kick the ball with more
    power after it is executed.
    Note that this skill will only be executed if it is possible to actually
    reach the desired ball position with a single kick. If the required power
    does exceed the maximum then the ball is frozen at its current position
    using the freezeBall skill. In general, it will then always be possible to
    shoot the motionless ball to the desired point in the next cycle.
    \param 'ang' relative angle to body to which the ball should be kicked
    \return SoccerCommand to kick the ball close to the body */
SoccerCommand BasicPlayer::kickBallCloseToBody( AngDeg ang )
{
  AngDeg      angBody    = WM->getAgentGlobalBodyAngle();
  VecPosition posAgent   = WM->predictAgentPos( 1, 0 );
  double      dDist      = SS->getPlayerSize() +
                           SS->getBallSize()   +
                           SS->getKickableMargin()/6.0;
  AngDeg      angGlobal  = VecPosition::normalizeAngle( angBody + ang );
  VecPosition posDesBall = posAgent   + VecPosition( dDist, angGlobal, POLAR );
  VecPosition vecDesired = posDesBall - WM->getBallPos();
  VecPosition vecShoot   = vecDesired - WM->getGlobalVelocity( OBJECT_BALL );
  double      dPower     = WM->getKickPowerForSpeed( vecShoot.getMagnitude() );
  AngDeg      angActual  = vecShoot.getDirection() - angBody;
              angActual  = VecPosition::normalizeAngle( angActual );

  if( dPower > SS->getMaxPower() && WM->getBallSpeed() > 0.1 )
  {
    Log.log( 500, "kickBallCloseToBody: cannot compensate for ball speed, freeze");
    Log.log( 101, "kickBallCloseToBody: cannot compensate for ball speed, freeze");
    return freezeBall();
  }
  else if( dPower > SS->getMaxPower() )
  {
    Log.log( 101, "kickBallCloseToBody: ball has no speed, but far away" );
    dPower = 100;
  }
  else
    Log.log( 101, "(kick %f %f), vecDesired (%f,%f)", dPower, angActual,
    vecDesired.getX(), vecDesired.getY() );



  /* Power reduced by quarter */
  return SoccerCommand( CMD_KICK, dPower / 4.0, angActual );
}

/*! This skill enables an agent to accelerate the ball in such a way that it
    gets a certain velocity after the kick. It receives the desired velocity
    'vecDes' as its only argument and returns a kick command that causes the
    ball to be accelerated to this velocity. If the power that must be supplied
    to the kick command to get the desired result does not exceed the maximum
    kick power then the desired velocity can be realized with a single kick.
    The kick direction should then be equal to the direction of the
    acceleration vector relative to the agent's global body angle. However,
    if the desired velocity is too great or if the current ball velocity is
    too high then the required acceleration cannot be realized with a single
    kick. In this case, the ball is kicked in such a way that the acceleration
    vector has the maximum possible length and a direction that aligns the
    resulting ball movement with 'vecDes'. This means that after the kick the
    ball will move in the same direction as 'vecDes' but at a lower speed. To
    this end the acceleration vector has to compensate for the current ball
    velocity in the `wrong' direction (y-component).
    \param velDes desired ball velocity
    \return SoccerCommand that accelerates the ball to 'vecDes' */
SoccerCommand BasicPlayer::accelerateBallToVelocity( VecPosition velDes )
{
  AngDeg      angBody  = WM->getAgentGlobalBodyAngle();
  VecPosition velBall  = WM->getGlobalVelocity( OBJECT_BALL );
  VecPosition accDes   = velDes - velBall;
  double      dPower;
  double      angActual;

  // if acceleration can be reached, create shooting vector
  if( accDes.getMagnitude() < SS->getBallAccelMax() )
  {
    dPower    = WM->getKickPowerForSpeed   ( accDes.getMagnitude() );
    angActual = VecPosition::normalizeAngle( accDes.getDirection() - angBody );
    if( dPower <= SS->getMaxPower() ) {

      
      return SoccerCommand( CMD_KICK, dPower, angActual );

    }
  }

  // else determine vector that is in direction 'velDes' (magnitude is lower)
         dPower    = SS->getMaxPower();
  double dSpeed    = WM->getActualKickPowerRate() * dPower;
  double tmp       = velBall.rotate(-velDes.getDirection()).getY();
         angActual = velDes.getDirection() - asinDeg( tmp / dSpeed );
         angActual = VecPosition::normalizeAngle( angActual - angBody );
	 
	 


  return SoccerCommand( CMD_KICK, dPower, angActual );
}

/*! This skill enables an agent to catch the ball and can only be executed
    when the agent is a goalkeeper. It returns a catch command that takes the
    angle of the ball relative to the body of the agent as its only argument.
    The correct value for this argument is computed by determining the global
    direction between the current ball position and the agent's current
    position and by making this direction relative to the agent's global body
    angle.
    \return SoccerCommand to catch the ball */
SoccerCommand BasicPlayer::catchBall()
{
  // true means returned angle is relative to body instead of neck
  return SoccerCommand( CMD_CATCH, WM->getRelativeAngle( OBJECT_BALL, true ));
}


/*! This skill enables an agent to communicate with other players on the field.
    It receives a string message as its only argument and returns a say command
    that causes the message to be broadcast to all players within a certain
    distance from the speaker.
    \return SoccerCommand to say the specified string 'str' */
SoccerCommand BasicPlayer::communicate( char *str )
{
  return SoccerCommand( CMD_SAY, str );
}

/*! This skill enables an agent to issue an attentionto command, which 
    causes him to focus on the specified player to receive hear messages only
    from that player. 
    int team - team >= 0 specifies our team, < 0 specifies opponent team.
    int player - must be a player number >= 1 and <= MAX_PLAYER (11).
    \return SoccerCommand to focus attention on team,player. */
SoccerCommand BasicPlayer::setAttentionto( int team, int player )
{
  return SoccerCommand( CMD_ATTENTIONTO, team, player );
}


/*! This method returns a 'move' command to teleport the agent directly to the
    specified global position.
    \param pos global position to which should be moved.
    \return SoccerCommand to move directly to 'pos'. */
SoccerCommand BasicPlayer::teleportToPos( VecPosition pos )
{
  return SoccerCommand( CMD_MOVE, pos.getX(), pos.getY() );

}

/********************** INTERMEDIATE LEVEL SKILLS ****************************/

/*! This skill enables an agent to turn his body towards an object o which is
    supplied to it as an argument. To this end the object's global position
    o in the next cycle is predicted based on its current velocity.
    This predicted position is passed as an argument to the turnBodyToPoint
    skill which generates a turn command that causes the agent to turn his
    body towards the object.
    \param o object to which agent wants to turn
    \return SoccerCommand that turns to this object */
SoccerCommand BasicPlayer::turnBodyToObject( ObjectT o )
{
  return turnBodyToPoint( WM->predictPosAfterNrCycles(o, 1) );
}

/*! This skill enables an agent to turn his neck towards an object. It
    receives as arguments this object o as well as a primary action command
    'soc' that will be executed by the agent at the end of the current cycle.
    Turning the neck towards an object amounts to predicting the object's
    global position in the next cycle and passing this predicted position
    together with the 'soc' command as arguments to the turnNeckToPoint skill.
    This low-level skill will then generate a turn neck command that causes the
    agent to turn his neck towards the given object. Note that the 'soc'
    command is supplied as an argument for predicting the agent's global
    position and neck angle after executing the command. This is necessary
    because a turn neck command can be executed in the same cycle as a kick,
    dash, turn , move or catch command.
   \param o object to which the agent wants to turn his neck
   \param soc SoccerCommand that is performed in this cycle.
   \return SoccerCommand that turns the neck of the agent to this object */
SoccerCommand BasicPlayer::turnNeckToObject( ObjectT o, SoccerCommand soc )
{
  return turnNeckToPoint( WM->predictPosAfterNrCycles(o, 1), soc );
}

/*! This skill enables an agent to move to a global position 'pos' on the field
    which is supplied to it as an argument. Since the agent can only move
    forwards or backwards into the direction of his body, the crucial decision
    in the execution of this skill is whether he should perform a turn or a
    dash. Turning has the advantage that in the next cycle the agent will be
    orientated correctly towards the point he wants to reach. However, it has
    the disadvantage that performing the turn will cost a cycle and will reduce
    the agent's velocity since no acceleration vector is added in that cycle.
    Apart from the target position 'pos', this skill receives several
    additional arguments for determining whether a turn or dash should be
    performed in the current situation. If the target point is in front of the
    agent then a dash is performed when the relative angle to this point is
    smaller than a given angle 'angWhenToTurn'. However, if the target point is
    behind the agent then a dash is only performed if the distance to point is
    less than a given value 'dDistBack' and if the angle relative to the back
    direction of the agent is smaller than 'angWhenToTurn'. In all other cases
    a turn is performed. Note that in the case of the goalkeeper it is
    sometimes desirable that he moves backwards towards his goal in order to
    keep sight of the rest of the field. To this end an additional boolean
    argument 'bMoveBack' is supplied to this skill that indicates whether the
    agent should always move backwards to the target point. If this value
    equals true then the agent will turn his back towards the target point if
    the angle relative to his back direction is larger than 'angToTurn'. In
    all other cases he will perform a (backward) dash towards 'posTo'
    regardless of whether the distance to this point is larger than 'dDistBack'.
    \param posTo global target position to which the agent wants to move
    \param angWhenToTurn angle determining when turn command is returned
    \param dDistBack when posTo lies closer than this value to the back of
          the agent (and within angWhenToTurn) a backward dash is returned
    \param bMoveBack boolean determing whether to move backwards to 'posTo'
    \return SoccerCommand that determines next action to move to 'posTo' */
SoccerCommand BasicPlayer::moveToPos( VecPosition posTo, AngDeg angWhenToTurn,
                                      double dDistBack, bool bMoveBack )
{
  VecPosition posPred   = WM->predictAgentPos( 1, 0 );
  AngDeg      angBody   = WM->getAgentGlobalBodyAngle();
  VecPosition posAgent  = WM->getAgentGlobalPosition();
  AngDeg      angTo     = ( posTo - posPred ).getDirection();
              angTo     = VecPosition::normalizeAngle( angTo - angBody );
  AngDeg      angBackTo = VecPosition::normalizeAngle( angTo + 180 );
  double      dDist     = posAgent.getDistanceTo( posTo );


  if( bMoveBack )
  {
    if( fabs( angBackTo ) < angWhenToTurn ) {
      return dashToPoint( posTo );
    }
    else {
      return turnBackToPoint( posTo );
    }
  }
  else if(  fabs( angTo     ) < angWhenToTurn ||
	    (fabs( angBackTo ) < angWhenToTurn && dDist < dDistBack ) ) {
    return dashToPoint( posTo );
  }
  else {
    return turnBodyToPoint( posTo );
  }
}

/*! This skill enables an agent to intercept a ball which is close to him. The
    objective is to move in such a way that the ball will come within the
    kickable distance from the agent in one or two cycles. To this end the
    prediction methods from the world model are used to predict the ball
    position in the next cycle and two cycles into the future. It is then
    determined whether it is possible to move the agent within kickable
    distance from one of these positions using all logical combinations of
    turn and dash commands. If it is not possible to intercept the ball within
    two cycles then this skill returns an illegal command to indicate that
    it cannot be performed. First it is determined whether the agent can
    intercept the ball in one cycle. To this end the position of the ball in
    the next cycle is predicted and a calculation is performed to decide
    whether a single dash can move the agent within the kickable distance from
    this position. In order to be able to kick the ball efficiently after
    intercepting it, it is important that the agent moves to a good position
    relative to the ball (i.e. the ball must be in front of him). At the same
    time the agent must make sure that he does not collide with the ball when
    trying to intercept it. Let l be a line that runs forwards and backwards
    from the predicted position of the agent in the next cycle into the
    direction of his body. This line thus denotes the possible movement
    direction of the agent. Note that we have to use the agent's predicted
    position in the next cycle since his current velocity must be taken into
    account. In addition, let c be a circle which is centered on the predicted
    ball position and which has a radius equal to the sum of the radius of the
    agent, the radius of the ball and a small buffer (kickable margin/6). It
    is now determined whether the agent can intercept the ball in the next
    cycle by looking at the number of intersection points between l and c. If
    l and c have exactly one point in common then this point is the desired
    interception point for the next cycle. However, if the number of
    intersection points equals two then the desired point is the one for
    which the absolute angle of the ball relative to that point is the
    smallest. This amounts to the intersection point which is closest to
    the agent when the ball lies in front of him and to the furthest one when
    the ball is behind his back. As a result, the desired interception point
    will always be such that the agent has the ball in front of him in the
    next cycle. Then a dash command is generated
    that will bring the agent as close as possible to the desired point.
    Next, the position of the agent after executing this command is predicted
    and if it turns out that this predicted position lies within the kickable
    distance from the ball then the dash is performed. However, if the
    predicted position is not close enough to the ball or if l and c have no
    points in common then it is assumed that the ball cannot be intercepted
    with a single dash. In these cases, two alternatives are explored to see
    if the ball can be intercepted in two cycles.
    The first alternative is to determine whether the agent can intercept the
    ball by performing a turn followed by a dash. To this end the global
    position of the ball is predicted two cycles into the future and a turn
    command is generated that will turn
    the agent towards this point. The agent's position after executing this
    command is then predicted after which a dash command is generated
    that will bring the agent as close as possible to
    the predicted ball position in two cycles. If it turns out that the
    predicted position of the agent after the dash lies within kickable
    distance from the ball then the first command (i.e. the turn) in the
    sequence of two is performed. Otherwise, a second alternative is tried to
    determine whether the agent can intercept the ball by performing two dash
    commands. To this end two dash commands are generated to get closer to the
    predicted ball position after two cycles.
    If the predicted position of the agent after these two dashes lies within
    kickable distance from the ball then the first dash is performed.
    Otherwise, an illegal command is returned to indicate that the skill
    cannot be performed. The close interception procedure is heavily based on
    a similar method introduced in CMU'99 by Peter Stone.
    \return command to intercept ball in two cycles, CMD_ILLEGAL otherwise */
SoccerCommand BasicPlayer::interceptClose( )
{
  SoccerCommand soc;
  double        dPower, dDist;
  AngDeg        ang,    ang2;
  VecPosition   s1,     s2;

  // first determine whether the distance to the ball is not too large
  dDist = 2*SS->getPlayerSpeedMax()
          + (1.0 + SS->getBallDecay())*SS->getBallSpeedMax()
          + SS->getMaximalKickDist();
  if( WM->getRelativeDistance( OBJECT_BALL ) > dDist )
    return SoccerCommand( CMD_ILLEGAL );

  // initialize all variables with information from the worldmodel.
  VecPosition   posAgent = WM->getAgentGlobalPosition( );
  VecPosition   posPred  = WM->predictAgentPos( 1, 0 );
  VecPosition   posBall  = WM->predictPosAfterNrCycles( OBJECT_BALL, 1 );
  VecPosition   velMe    = WM->getAgentGlobalVelocity( );
  Stamina       sta      = WM->getAgentStamina( );
  AngDeg        angBody  = WM->getAgentGlobalBodyAngle( );

  // make a line from center of body in next cycle with direction of body
  // use next cycle since current velocity is always propogated to position in
  // next cycle.  Make a circle around the ball with a radius equal to the
  // sum of your own body, the ball size and a small buffer. Then calculate
  // the intersection between the line and this circle. These are the (two)
  // points that denote the possible agent locations close to the ball
  // From these two points we take the point where the body direction of the
  // agent makes the smallest angle with the ball (with backward
  // dashing we sometime have to dash "over" the ball to face it up front)
  Line line = Line::makeLineFromPositionAndAngle(posPred,angBody);
  dDist     = SS->getPlayerSize()+SS->getBallSize()+SS->getKickableMargin()/6;
  int  iSol = line.getCircleIntersectionPoints( Circle(posBall,dDist), &s1, &s2);
  if (iSol > 0)                                          // if a solution
  {
    if (iSol == 2)                                       // take the best one
    {
     ang  = VecPosition::normalizeAngle((posBall - s1).getDirection() -angBody);
     ang2 = VecPosition::normalizeAngle((posBall - s2).getDirection() -angBody);
     if ( fabs(ang2) < 90)
       s1 = s2;                                          // and put it in s1
    }

    // try one dash
    // now we have the interception point we try to reach in one cycle. We
    // calculate the needed dash power from the current position to this point,
    // predict were we will stand if we execute this command and check whether
    // we are in the kickable distance
    dPower  = WM->getPowerForDash(s1-posAgent, angBody, velMe,sta.getEffort() );
    posPred = WM->predictAgentPos( 1, (int)dPower);
    if ( posPred.getDistanceTo( posBall ) < SS->getMaximalKickDist() )
      return SoccerCommand( CMD_DASH, dPower );
  }

  // try one dash by getting close to ball
  // this handles situation where ball cannot be reached within distance
  // SS->getKickableMargin()/6
  soc = dashToPoint( posBall );
  WM->predictAgentStateAfterCommand(soc,&posPred,&velMe,&angBody,&ang,&sta );
  if ( posPred.getDistanceTo( posBall ) < SS->getMaximalKickDist() )
    return soc;
  
  // try turn and then dash
  // first predict the angle between the agent and the ball after two cycles,
  // is this larger than the angle when to turn, turn to position of the ball
  // after two cycles. Then perform a dash based on the position of the ball
  // after two cycles and the position of the agent after the turn. If after
  // these two command the agent is in the kickable distance, return turn
  // command.
  posBall = WM->predictPosAfterNrCycles( OBJECT_BALL, 2 );
  posPred = WM->predictAgentPos( 2, 0 );
  ang     = (posBall - posPred).getDirection();
  ang     = VecPosition::normalizeAngle( ang - angBody );
  if (fabs( ang ) > PS->getPlayerWhenToTurnAngle() ) // if we want to turn
  {
    soc = turnBodyToPoint( posBall, 2 );             // perform turn
    WM->predictAgentStateAfterCommand(soc,&posPred,&velMe,&angBody,&ang,&sta );
    dPower = WM->getPowerForDash(posBall-posPred,angBody,velMe,sta.getEffort());
    WM->predictStateAfterDash( dPower, &posPred, &velMe, &sta, angBody);
    if (posPred.getDistanceTo(posBall) < SS->getMaximalKickDist())
      return soc;
  }

  // try two dashes
  // first predict the position in the next cycle when dash with full power
  // is performed. Then calculate the dash power to reach the point where the
  // ball will be in two cycles and predict the global position of the agent
  // after a dash with this power. If the position is in the kickable distance
  // return a dash with full power.
  soc = SoccerCommand( CMD_DASH, SS->getMaxPower() );
  WM->predictAgentStateAfterCommand(soc,&posPred,&velMe,&angBody,&ang,&sta );
  dPower=WM->getPowerForDash(posBall-posPred,angBody,velMe,sta.getEffort());
  WM->predictStateAfterDash( dPower, &posPred, &velMe, &sta, angBody );
  if (posPred.getDistanceTo(posBall) < SS->getMaximalKickDist())
    return soc;

  // try three dashes
  // same as with two dashes, but now try two maximum dashes and
  posBall = WM->predictPosAfterNrCycles( OBJECT_BALL, 3 );
  soc = SoccerCommand( CMD_DASH, SS->getMaxPower() );
  WM->predictAgentStateAfterCommand(soc,&posPred,&velMe,&angBody,&ang,&sta );
  WM->predictStateAfterCommand( soc, &posPred, &velMe, &angBody, &ang,&sta );
  dPower=WM->getPowerForDash(posBall-posPred,angBody,velMe,sta.getEffort());
  WM->predictStateAfterDash( dPower, &posPred, &velMe, &sta, angBody );
  if (posPred.getDistanceTo(posBall) < SS->getMaximalKickDist())
    return soc;

  // cannot intercept ball in two cycles
  return SoccerCommand( CMD_ILLEGAL );
}

/*! This skill enables a goalkeeper to intercept a ball which is close to him.
    The objective is to move in such a way that the ball will come within the
    catchable distance from the agent in one or two cycles. To this end the
    prediction methods from the world model are used to predict the ball
    position in the next cycle and two cycles into the future. It is then
    determined whether it is possible to move the agent within the
    catchable area from one of these positions using all logical combinations
    of turn and dash commands. If it is not possible to intercept the ball
    within two cycles then this skill returns an illegal command to indicate
    that it cannot be performed. First it is determined whether the goalkeeper
    can intercept the ball in one cycle. To this end the position of the ball
    in the next cycle is predicted and a calculation is performed to decide
    whether a single dash can move the agent within catchable distance from
    this position. If it turns out that this is the case, the corresponding
    dash is performed. However, if the predicted position is not close enough
    to the ball then it is assumed that the ball cannot be intercepted with a
    single dash. In these cases, two alternatives are explored to see if the
    ball can be intercepted in two cycles.
    The first alternative is to determine whether the agent can intercept the
    ball by performing two dash commands. To this end two dash commands are
    generated. If the predicted position of the agent after these two
    dashes lies within catchable distance from the ball then the first dash is
    performed. Otherwise, a second alternative is tried to determine whether
    the agent can intercept the ball by performing a turn
    followed by a dash. To this end the global position of the ball is
    predicted two cycles into the future and a turn command is generated
    that will turn the agent towards this point. The agent's position after
    executing this command is then predicted after which a dash command is
    generated that will bring the agent as close as possible to the predicted
    ball position. If it turns out that the predicted position of the agent
    after the dash lies within catchable distance from the ball then the first
    command (i.e. the turn) in the sequence of two is performed.
    Otherwise, an illegal command is returned to indicate that the
    skill cannot be performed. The close interception procedure is heavily
    based on the method used by CMU'99 by Peter Stone.
    \return command to intercept ball in two cycles, CMD_ILLEGAL otherwise */
SoccerCommand BasicPlayer::interceptCloseGoalie( )
{
  SoccerCommand soc;
  double        dPower, dDist;
  AngDeg        ang;
  VecPosition   posClosestToBall;

  // initialize all variables with information from worldmodel.
  VecPosition   posPred   = WM->predictAgentPos( 1, 0 );
  VecPosition   posBall   = WM->predictPosAfterNrCycles( OBJECT_BALL, 1 );
  VecPosition   velMe     = WM->getAgentGlobalVelocity( );
  Stamina       sta       = WM->getAgentStamina( );
  AngDeg        angBody   = WM->getAgentGlobalBodyAngle( );
  Line          lineGoalie=Line::makeLineFromPositionAndAngle(posPred,angBody);

  // when it is theoretical possible
  // try one dash and check whether ball is in catchable area
  dDist = SS->getBallSpeedMax()+SS->getPlayerSpeedMax()+SS->getCatchableAreaL();
  if( WM->getRelativeDistance( OBJECT_BALL ) < dDist )
  {
    posClosestToBall = lineGoalie.getPointOnLineClosestTo( posBall );
    dPower           = WM->getPowerForDash(
                          posClosestToBall-posPred,
                          angBody,
                          velMe,
                          sta.getEffort()           );
    posPred          = WM->predictAgentPos( 1, (int)dPower);
    if ( posPred.getDistanceTo( posBall ) < SS->getCatchableAreaL() )
      return SoccerCommand( CMD_DASH, dPower );
  }

  // when it is theoretical possible
  // try two dashes and check whether ball is in catchable area
  // otherwise try first two  dashes and check whether ball is in catchable
  // area, thereafter for turn and dash.
  dDist = SS->getBallSpeedMax()*(1.0+SS->getBallDecay())
           + 2*SS->getPlayerSpeedMax()
           + SS->getCatchableAreaL();
  if( WM->getRelativeDistance( OBJECT_BALL ) < dDist )
  {
    // try two dashes
    // first predict the position in the next cycle when dash with full power
    // is performed. Then calculate the dash power to reach the point where the
    // ball will be in two cycles and predict the global position of the agent
    // after a dash with this power. If the position is in the catchable area
    // return a dash with full power.
    posBall = WM->predictPosAfterNrCycles( OBJECT_BALL, 2 );
    soc     = dashToPoint( posBall );
    WM->predictAgentStateAfterCommand(soc,&posPred,&velMe,&angBody,&ang,&sta );
    dPower=WM->getPowerForDash(posBall-posPred,angBody,velMe,sta.getEffort());
    WM->predictStateAfterDash( dPower, &posPred, &velMe, &sta, angBody );
    if( posPred.getDistanceTo(posBall) < SS->getCatchableAreaL() )
      return soc;

    // try one turn and a dash
    posBall = WM->predictPosAfterNrCycles( OBJECT_BALL, 2 );
    posPred = WM->predictAgentPos( 2, 0 );
    ang     = (posBall - posPred).getDirection();
    ang     = VecPosition::normalizeAngle( ang - angBody );
    if (fabs( ang ) > PS->getPlayerWhenToTurnAngle() ) // if we want to turn
    {
      soc = turnBodyToPoint( posBall, 2 );             // perform turn
      WM->predictAgentStateAfterCommand(soc,&posPred,&velMe,&angBody,&ang,&sta );
      dPower=WM->getPowerForDash(posBall-posPred,angBody,velMe,sta.getEffort());
      WM->predictStateAfterDash( dPower, &posPred, &velMe, &sta, angBody);
      if( posPred.getDistanceTo(posBall) < SS->getCatchableAreaL() ){
	//cout <<WM->getCurrentTime().getTime()<<": intercept: trying one turn and a dash"<<endl;
        return soc;
      }
    }
  }

  // did not succeed
  return SoccerCommand( CMD_ILLEGAL );
}

/*! This skill enables an agent to kick the ball from its current position to a
   given position 'posTarget' in such a way that it has a remaining speed equal
   to 'dEndSpeed' when it reaches this position. However, it is possible that
   the ball cannot reach this velocity with a single kick either because the
   needed magnitude of the generated velocity vector exceeds the maximum speed
   of the ball or due to the fact that the current ball speed in combination
   with the position of the ball relative to the agent make it impossible to
   achieve the required acceleration. If the magnitude of needed velocity vector
   is larger than ball speed max it is certain that even in the optimal
   situation (i.e. if the ball lies directly in front of the agent and has zero
   velocity) the agent will not be able to kick the ball to the target position
   at the desired speed. In this case the expected ball movement is computed
   after executing a kick with maximum power into the direction of 'posTarget'.
   If the magnitude of the resulting movement vector is larger than a given
   percentage ('getPlayerWhenToKick' defined in the PlayerSettings)  of the
   maximum ball speed then this kick is actually performed despite the fact
   that it cannot produce the wanted effect. Otherwise, the agent shoots the
   ball close to his body and directly in front of him using the
   kickBallCloseToBody skill. In this way he will be able to kick the ball with
   more power in the next cycle. However, if the magnitude of the desired
   velocity vector is smaller than ball speed max it is possible to reach the
   target point at the desired speed in the optimal situation.
   If the power that must be supplied to the kick command to achieve this
   acceleration is less than or equal to the maximum power the
   accelerateBallToVelocity skill is used to perform the desired kick.
   Otherwise, the agent uses the kickBallCloseToBody skill to put the
   ball in a better kicking position for the next cycle.
   \param posTarget target position where the ball should be shot to
   \param endSpeed desired speed ball should have in target position
   \return SoccerCommand that produces kick  */
SoccerCommand BasicPlayer::kickTo( VecPosition posTarget, double dEndSpeed )
{
  VecPosition posBall = WM->getBallPos();
  VecPosition velBall = WM->getGlobalVelocity(OBJECT_BALL);
  VecPosition posTraj = posTarget - posBall;
  VecPosition velDes  = VecPosition(
               WM->getKickSpeedToTravel( posTraj.getMagnitude(), dEndSpeed ),
               posTraj.getDirection(),
               POLAR                                                        );
  double      dPower;
  AngDeg      angActual;

  Log.log( 101, "ball (%f,%f), agent (%f,%f), to (%f,%f) ang %f %f" ,
          WM->getBallPos().getX(), WM->getBallPos().getY(),
          WM->getAgentGlobalPosition().getX(),
          WM->getAgentGlobalPosition().getY(),
          posTarget.getX(), posTarget.getY(),
          WM->getAgentGlobalBodyAngle(),
          WM->getAgentGlobalNeckAngle() );
  Log.log( 101, "rel pos body (%f,%f), velocity ball: (%f,%f) kick_dist: %f  (%f,%f,%f)" ,
          WM->getRelativeDistance( OBJECT_BALL ),
          WM->getRelativeAngle( OBJECT_BALL, true ),
           velBall.getX(), velBall.getY(),
           SS->getMaximalKickDist(), SS->getPlayerSize(), SS->getBallSize(), 
             SS->getKickableMargin() );


  if( WM->isKickInUs() || WM->isFreeKickUs())
  {
    Log.log( 101, "kick in or free kick, accelerate ball" );
    return accelerateBallToVelocity( velDes );  
  }

  if( velDes.getMagnitude() > SS->getBallSpeedMax() ) // can never reach point
  {
           dPower     = SS->getMaxPower();
    double dSpeed     = WM->getActualKickPowerRate() * dPower;
    double tmp        = velBall.rotate(-velDes.getDirection()).getY();
           angActual  = velDes.getDirection() - asinDeg( tmp / dSpeed );
    double dSpeedPred = ( WM->getGlobalVelocity(OBJECT_BALL)+
                         VecPosition(dSpeed,angActual, POLAR )).getMagnitude();

    // but ball acceleration in right direction is very high
    if( dSpeedPred > PS->getPlayerWhenToKick()*SS->getBallAccelMax() )
    {
      Log.log( 101, "point (%f,%f) too far, but can acccelerate ball good to %f k=%f,%f",
                     velDes.getX(), velDes.getY(), dSpeedPred, dSpeed, tmp );
      return accelerateBallToVelocity( velDes );    // shoot nevertheless
    }
    else if( WM->getActualKickPowerRate() >
        PS->getPlayerWhenToKick() * SS->getKickPowerRate() )
    {
      Log.log( 101, "point too far, freeze ball" ); // ball well-positioned
      return freezeBall();                          // freeze ball
    }
    else
    {
      Log.log( 101, "point too far, reposition ball (k_r = %f)",
                      WM->getActualKickPowerRate()/(SS->getKickPowerRate()) );
      return kickBallCloseToBody( 0 );            // else position ball better
    }
  }
  else                                            // can reach point
  {
    VecPosition accBallDes = velDes - velBall;
    dPower = WM->getKickPowerForSpeed(accBallDes.getMagnitude());
    if( dPower <= 1.05*SS->getMaxPower() )        // with current ball speed
    {                               // 1.05 since cannot get ball fully perfect
      Log.log( 101, "point good and can reach point" );
      return accelerateBallToVelocity( velDes );  // perform shooting action
    }
    else
    {
      Log.log( 101, "point good, but reposition ball since need %f",dPower );
      return kickBallCloseToBody( 0 );            // else position ball better
    }
  }
}

/*! This skill enables an agent to turn towards a global angle while keeping
    the ball in front of him. It is used, for example, when a defender has
    intercepted the ball in his defensive area and faces his own goal. In this
    situation the defender usually wants to pass the ball up the field into an
    area that is currently not visible to him and to this end he will first use
    this skill to turn with the ball towards the opponent's goal. Turning with
    the ball requires a sequence of commands to be performed. The ball first has
    to be kicked to a desired position relative to the agent, then it has to be
    stopped dead at that position and finally the agent must turn towards the
    ball again. Each time when this skill is called it has to be determined
    which part of the sequence still has to be executed. This is done as follows.
    If the absolute difference between the desired angle and the global angle
    of the ball relative to the position of the agent is larger than the
    value 'angKickThr' then the kickBallCloseToBody skill is used to kick
    the ball to a position close to the agent and at the desired angle.
    Otherwise, it is checked whether the ball still has speed from the previous
    action. If the remaining ball speed exceeds the given value
    'dFreezeThr' then the ball is stopped dead at its current position
    using the freezeBall skill. In all other cases the agent turns his body
    towards the specified angle 'ang'.
    \param ang global direction in which ball and player should be faced
    \param angKickThr when ball angle difference is larger than this value
                      ball is repositioned
    \param dFreezeThr when ball lies correct, but has speed higher than this
                      value, ball is frozen.
    \return Soccercommand to turn with the ball to global angle 'ang'.
   */

/*
SoccerCommand BasicPlayer::turnWithBallTo( AngDeg ang, AngDeg angKickThr,
                                           double dFreezeThr )
{
  // if opponent is close
  // if ball is located more than 'angKickThr' degrees from ang
  //  kick ball to point right in front of player in direction ang
  // else if ball has still speed higher than 'dFreezeThr'
  /// freezeBall
  // else
  //  turn to direction 'ang'
  VecPosition posAgent = WM->getAgentGlobalPosition();
  VecPosition posBall  = WM->getBallPos();
  AngDeg      angBody  = WM->getAgentGlobalBodyAngle();
  AngDeg      angDiff  = (posBall-posAgent).getDirection() - ang;
              angDiff  = VecPosition::normalizeAngle( angDiff );
  double      dDist;
  ObjectT     objOpp   = WM->getClosestInSetTo( OBJECT_SET_OPPONENTS,
                           WM->getAgentObjectType(), &dDist );
  VecPosition posOpp   = WM->getGlobalPosition( objOpp );

  if( objOpp != OBJECT_ILLEGAL && dDist < 2.5 )
  {
    if( posBall.getDistanceTo( posOpp ) < posBall.getDistanceTo( posAgent ) )
    {
      ang = (posOpp - posAgent).getDirection() + 180;
      return kickBallCloseToBody( VecPosition::normalizeAngle( ang - angBody ));
    }
  }
  else if( fabs( angDiff ) > angKickThr )
  {
    Log.log( 101, "turnWithBall: kick ball close %f", ang );
    return kickBallCloseToBody( VecPosition::normalizeAngle( ang - angBody ) );
  }

  // hier niet altijd freezen -> kan dan niet meer goed liggen omdat je zelf
  // doorschiet.
  if( WM->getBallSpeed() > dFreezeThr )
  {
    Log.log( 101, "turnWithBall: freeze ball" );
    return freezeBall();
  }

  ACT->putCommandInQueue( alignNeckWithBody() );
  return turnBodyToPoint( posAgent + VecPosition(1.0, ang, POLAR ) );
}
*/
SoccerCommand BasicPlayer::turnWithBallTo( AngDeg ang, AngDeg ,
                                           double )
{
  // if opponent is closer to the ball than I am
  //  kick ball away from his direction
  // if the ball will be in my kick_range in the next cycle
  //  turn to direction ang
  // if ball will be in kick range if frozen
  //   freeze_ball
  // else
  //   kickBallCloseToBody(ang)
  VecPosition   posAgent= WM->getAgentGlobalPosition();
  VecPosition   posBall = WM->getBallPos();
  AngDeg        angBody = WM->getAgentGlobalBodyAngle();
  double        dDist;
  ObjectT       objOpp  = WM->getClosestInSetTo( OBJECT_SET_OPPONENTS,
                          WM->getAgentObjectType(), &dDist );
  VecPosition   posOpp  = WM->getGlobalPosition( objOpp );
  SoccerCommand soc     = turnBodyToPoint(posAgent+VecPosition(1.0,ang,POLAR));

  if( objOpp != OBJECT_ILLEGAL && dDist < 2.5 )
  {
    AngDeg angBall = (posBall-posAgent).getDirection();
    AngDeg angOpp  = (posOpp -posAgent).getDirection();
    if( fabs( VecPosition::normalizeAngle( angBall - angOpp ) ) < 90  )
    {
      ang = angOpp + 180;
      Log.log( 101, "turnWithBall: kick ball away from opp at ang %f", angOpp );
      return kickBallCloseToBody( VecPosition::normalizeAngle( ang - angBody ));
    }
  }

  VecPosition posAgentPred = WM->predictAgentPosAfterCommand( soc );
  VecPosition posBallPred  = WM->predictPosAfterNrCycles( OBJECT_BALL, 1 );
  if( posAgentPred.getDistanceTo( posBallPred ) < 0.8*SS->getMaximalKickDist()
   || 
   ( posAgentPred.getDistanceTo(posBallPred) < 0.9*SS->getMaximalKickDist() 
     && WM->getBallSpeed() < 0.1 ))
  {
    Log.log( 101, "turnWithBall: turn since ball will be kickable in t+1" );
    return soc;
  }

  posAgentPred = WM->predictPosAfterNrCycles( WM->getAgentObjectType(), 1, 0 );

  // if ball will not be in kickable distance in next turn, we have to freeze
  // the ball. Do this only when current ball position (same as frozen ball
  // position) lies within kickable distance of predicted agent position in
  // next cycle. When ball lies at edge of the kickable distance, do not
  // freeze, since when vel.is not completely correct can move outside area
  if( posAgentPred.getDistanceTo( posBall ) < 0.8*SS->getMaximalKickDist() && 
      WM->getBallSpeed() > 0.1  )
  {
    Log.log( 101, "turnWithBall: freeze ball" );
    return freezeBall();
  }
  else
  {
    Log.log( 101, "turnWithBall: kick ball close to desired turn angle" );
    return kickBallCloseToBody( VecPosition::normalizeAngle( ang - angBody ) );
  }
}




/*! This skill enables an agent to move along a line to a given position 'pos'
    on this line. It is used, for example, by the goalkeeper who often wants to
    stay on a line in front of his goal and move to different positions on this
    line depending on where the ball is located. Furthermore, it can also used
    by defenders for marking an  opponent player by moving along a line between
    this player and the ball. The idea is that the agent must try to move as
    fast as possible to the desired point 'pos' along the line thereby
    keeping the number of turns to a minimum to avoid wasting cycles. Apart from
    the target position 'pos', this skill receives several additional arguments
    for determining whether the agent should turn or dash in the current
    situation. Since the agent can only move forwards or backwards into the
    direction of his body, it is important that he tries to keep the orientation
    of his body aligned with the direction of the line in order to be able to
    move quickly to the target point. A given angle 'ang' denotes the desired
    body angle (global) of the agent in the point 'pos'. The line l can thus be
    defined as going through 'pos' and having global direction 'ang'. Due to the
    noise that is added to the movement of the agent, the orientation of his
    body will never be exactly equal to 'ang' and as a result the agent'
    s position will start to deviate from the line. Each time when this skill
    is called, the agent's desired orientation is therefore slightly adjusted
    depending on his position with respect to l. If the distance d between the
    agent's current position and the line l is smaller than the given value
    'dDistThr' then 'ang' remains unchanged. However, if d exceeds 'dDistThr'
    then 'ang' is adjusted in such a way that the agent will move closer to l
    in subsequent cycles. This is done by either increasing or decreasing the
    desired orientation 'ang' by 'angCorr' degrees depending on which side of
    the line the agent is located and on a prediction of the agent's movement
    in the forthcoming cycles. This prediction is represented by a given
    value 'iSign' which equals 1 if the agent is expected to move in the same
    direction as 'ang' and -1 if he will move in the opposite  direction.
    Adjusting 'ang' in this way has the effect that in subsequent cycles the
    agent will move closer to the line again if this is necessary. The final
    decision whether to turn or dash is now made by comparing the agent's
    current body angle to the desired orientation. If the difference between
    these two angles is larger than 'angThr' degrees then the agent uses the
    turnBodyToPoint skill to turn in the desired direction.
    Otherwise, the dashToPoint skill is called to move towards the
    target position.
    \param pos global position to which the agent wants to move
    \param ang desired global body angle of agent in position 'pos'
    \param dDistThr threshold value that defines if desired angle is adjusted
    \param iSign indication whether agent predicts that he will move in the same
    direction as 'ang' in the subsequent cycles (iSign=1) or in the opposite
    direction (iSign=-1)
    \param angThr threshold value that specifies when agent will perform a turn
    command
    \param angCorr correction term with which angle is adjusted if necessary
    \return SoccerCommand that will move agent along line defined by position
    'pos' and angle 'ang'. */
SoccerCommand BasicPlayer::moveToPosAlongLine( VecPosition pos, AngDeg ang,
    double dDistThr, int iSign, AngDeg angThr, AngDeg angCorr )
{
  Line        l        = Line::makeLineFromPositionAndAngle( pos, ang );
  VecPosition posBall  = WM->getBallPos();
  VecPosition posAgent = WM->getAgentGlobalPosition();
  AngDeg      angBody  = WM->getAgentGlobalBodyAngle();
  VecPosition posProj  = l.getPointOnLineClosestTo( posAgent );
  double      dDist    = posAgent.getDistanceTo( posProj );
  double      dDiff    = pos.getDistanceTo     ( posProj );

  // if deviated too much from line, compensate
  if( dDist > dDistThr )
  {
    // check on which side of line agent is located
    VecPosition posOrg(0,0);
    Line        m            = Line::makeLineFromTwoPoints( posOrg, posAgent );
    VecPosition posIntersect = l.getIntersection( m );
    int         iSide;
    if( posAgent.getDistanceTo(posOrg) < posIntersect.getDistanceTo( posOrg ) )
      iSide = +1;
    else
      iSide = -1;

    // adjust desired turning angle to move back to line in coming cycles
    ang = ang + iSign * iSide * angCorr;
  }

  Log.log( 553, "y difference to defend point %f", dDiff );
  // if current body angle differs much from desired turning angle, turn body
  if( fabs( VecPosition::normalizeAngle( ang - angBody ) ) > angThr )
  {
    Log.log( 553, "angle differs too much body = %f, des = %f", angBody, ang );
    return turnBodyToPoint( posAgent + VecPosition( 1.0, ang, POLAR ) );
  }
  else if( ( posBall.getX() > -PITCH_LENGTH/4.0 && dDiff > 0.6 ) )
    return dashToPoint( pos );  // move later when ball is far from pen. area.
  else if( ( posBall.getX() < -PITCH_LENGTH/4.0 && dDiff > 0.3 ) )
    return dashToPoint( pos );  // move earlier when is ball near pen. area. 
  else
    return SoccerCommand( CMD_ILLEGAL );
}


/********************** HIGH LEVEL SKILLS ************************************/

/*! When the ball-interception skill is called, it is first determined whether
    it is possible for the agent to intercept the ball within two cycles using
    the intermediate player skill closeIntercept (for the goalkeeper
    closeIntereptGoalie). If it turns out that the ball cannot be intercepted
    within  two cycles then the agent uses an iterative scheme to compute the
    optimal interception point. This is done using the method
    'getInterceptionPointBall'.
    \param isGoalie indicates whether the current player is a goalkeeper or not
    \return SoccerCommand to intercept the ball. */
SoccerCommand BasicPlayer::intercept( bool isGoalie )
{
  SoccerCommand soc = (isGoalie) ? interceptCloseGoalie() : interceptClose();
  if( soc.commandType != CMD_ILLEGAL )
  {
    Log.log( 502, "intercept in two cycles" );
    return soc;
  }

  // else find the best interception point
  int         iSol;
  VecPosition posBall = getInterceptionPointBall( &iSol, isGoalie  );

  /*
    if(WM->getPlayerNumber() == 1 && !WM->isInOwnPenaltyArea(posBall)){
    cout << "[BasicPlayer:intercept]: "<<WM->getCurrentTime()<<": goalie intercepting outside penalty area: "
	 << posBall<<endl;

	 }*/


  Log.log( 502, "intercept ball in %d cycles", iSol );
  return moveToPos( posBall, PS->getPlayerWhenToTurnAngle(), 3.0 );
}

/*! This skill enables an agent to dribble with the ball, i.e. to move with the
    ball while keeping it within a certain distance. This amounts to repeatedly
    kicking  the ball at a certain speed into a desired direction and then
    intercepting it again.  Two arguments, the angle 'ang' and type 'dribbleT',
    are supplied to this skill which respectively denote the global direction
    towards which the agent wants to dribble and the kind of dribble that must
    be performed. We distinguish three kinds of dribbling:
    - DRIBBLE FAST: a fast dribble action in which the agent kicks the ball
                    relatively far ahead of him.
    - DRIBBLE SLOW: a slower dribble action in which the agent keeps the ball
                    closer than on a fast dribble.
    - DRIBBLE WITH BALL: a safe dribble action in which the agent keeps the
                    ball very close to his body.

    It is important to realize that this skill is only called when the ball is
    located within the agent's kickable range. This means that it is only
    responsible for the kicking part of the overall dribbling behavior, i.e.
    it only causes the ball to be kicked a certain distance ahead into the
    desired direction 'ang'. If the absolute angle between 'ang' and the agent's
    body direction is larger than DribbleTurnAngle (which currently has a value
    of 30 degrees) then the agent uses the turnWithBallTo skill to turn with the
    ball towards the global angle 'ang'. Otherwise, he uses the kickTo skill to
    kick the ball into the desired direction towards a point that lies a certain
    distance ahead depending on the type of dribble. After the kick, the ball
    will move out of the agent's kickable range and as a result the agent will
    try to intercept it using the intercept skill. The dribbling skill can
    then be called again once the agent has succeeded in intercepting the ball.
    This sequence of kicking and intercepting will repeat itself until the agent
    decides to perform another skill. Note that during the dribble the power of
    a kick depends on the distance that the ball should travel and on the speed
    that it should have when it reaches the target point.
    In our current implementation this speed equals 0.5 (=DribbleKickEndSpeed)
    for any type of dribble. Experiments have shown that lower end speed values
    cause the agent to intercept the ball before it reaches the target point
    which slows the dribble down significantly.
    \param ang global direction in which should be dribbled
    \param dribbleT type of dribble that should be performed
    \return SoccerCommand to dribble in direction 'ang' */
SoccerCommand BasicPlayer::dribble( AngDeg ang, DribbleT dribbleT )
{
  double        dLength;
  AngDeg        angBody  = WM->getAgentGlobalBodyAngle();
  VecPosition   posAgent = WM->getAgentGlobalPosition();
  SoccerCommand soc;

  // if not turned into correct direction, turn with the ball to that angle
  AngDeg angDiff = VecPosition::normalizeAngle( ang - angBody );
  if( fabs( angDiff ) > PS->getDribbleAngThr() ){
    //cerr <<"[dribble]: turning with ball first. ang="<<angDiff<<endl;
    //cerr <<fabs( angDiff )<<" "<< PS->getDribbleAngThr()<<endl;
    return turnWithBallTo( ang, PS->getTurnWithBallAngThr(),
                                PS->getTurnWithBallFreezeThr() );
  }

  switch( dribbleT )
  {
    case DRIBBLE_WITHBALL:
      dLength = 4.0;
      break;
    case DRIBBLE_SLOW:
      dLength = 5.0;
      break;
    case DRIBBLE_FAST:
      dLength = 10.0;
      break;
    default:
      dLength = 0.0;
      break;
  }

  // determine shooting point, relative to agent since moving in that dir.
  VecPosition posDribble = posAgent + VecPosition( dLength, angBody, POLAR );

  // adjust when point lies outside side of field
  // to a point that lies at distance 2.0 from the side of the field
  if( fabs( posDribble.getY() ) > PITCH_WIDTH/2.0  - 3.0 ) 
    posDribble.setY( (PITCH_WIDTH/2.0  - 3.0)*sign(posDribble.getY()) );
  if( fabs( posDribble.getX() ) > PITCH_LENGTH/2.0 - 3.0 )
    posDribble.setX( (PITCH_LENGTH/2.0  - 3.0)*sign(posDribble.getX()) );  

  soc = kickTo( posDribble, 0.5 );

  // if ball is kickable but already heading in the right direction, so only
  // small correction term is necessary, start intercepting
  SoccerCommand soc2      = intercept( false );
  if( soc.dPower < 7  && WM->isCollisionAfterDash( soc2 ) == false )
  {
    Log.log( 560, "wanted to dribble, but only small kick %f", soc.dPower );
    return soc2;
  }
  return soc;
}

/*! This skill enables an agent to pass the ball directly to another player. It
    receives two arguments, 'pos' and 'passType', which respectively denote the
    position (of usually a teammate) to which the agent wants to pass the ball
    and the kind of pass (either normal or fast) that should be given. This
    skill uses the kickTo skill to pass the ball to the specified position
    with a certain desired end speed depending on the type of pass.
    \param pos position of object to which a direct pass should be given
    \param passType kind of pass (either PASS_NORMAL or PASS_FAST )
    \return SoccerCommand to perform a direct pass to object 'o' */
SoccerCommand BasicPlayer::directPass( VecPosition pos, PassT passType)
{

  if( passType == PASS_NORMAL )
    return kickTo( pos, PS->getPassEndSpeed() );
  else if( passType == PASS_FAST )
    return kickTo( pos, PS->getFastPassEndSpeed() );
  else
    return SoccerCommand( CMD_ILLEGAL );
}

/*! This skill enables an agent to give a leading pass to another player. A
    leading pass is a pass into open space that `leads' the receiver, i.e.
    instead of passing the ball di:rectly to another player it is kicked just
    ahead of him. In this way the receiver is able to intercept the ball while
    moving in a forward direction and this will speed up the attack. This skill
    receives two arguments, an object o (usually a teammate) and dist, which
    respectively denote the intended receiver of the leading pass and the
    `leading distance' ahead of the receiver. It uses the kickTo skill to pass
    the ball to a point that lies dist in front of the current position of o.
    Here `in front of' means in positive x-direction, i.e. at a global angle of
    0 degrees. Note that the desired end speed for a leading pass is always
    equal to PassEndSpeed (currently 1.4) since the leading aspect of the pass
    might cause the receiver to miss the ball when its speed is higher.
    \param o object to which a leading pass should be given
    \param dDist distance in front of o to which is passed
    \return SoccerCommand to perform a leading pass to object 'o' */
SoccerCommand BasicPlayer::leadingPass( ObjectT o, double dDist )
{
  VecPosition posShoot = WM->getGlobalPosition( o )
                         + VecPosition( dDist, 0.0, POLAR );
  return kickTo( posShoot, PS->getPassEndSpeed() );
}

/*! This skill enables an agent to give a more advanced type of pass called a
    through pass. With a through pass the ball is not passed directly to another
    player or just ahead of him, but it is kicked into open space between the
    opponent defenders and the opponent goalkeeper in such a way that a teammate
    (usually an attacker) will be able to reach the ball before an opponent
    does. If a through pass is executed successfully it often causes a
    disorganization of the opponent's defense which will enable an attacker to
    get the ball close to the enemy goal. This skill takes an object o (usually
    a teammate) as an argument which denotes the intended receiver of the
    through pass. The position p on the field to which the ball should be kicked
    is determined by drawing a line l from the object's current position to a
    given point 'pos' also supplied as an argument) and by computing the safest
    trajectory for the ball to a point on this line. To this end the widest
    angle between opponents from the current ball position to a point p on l is
    calculated. After this, the speed that the ball should have when it
    reaches this point is determined based on the distance from the current
    ball position to p and on the number of cycles n that o will need to
    reach p. If it turns out that the required end speed falls outside the
    range [MinPassEndSpeed .. MaxPassEndSpeed] it is set to the closest
    boundary of this range. The kickTo skill is then used to kick the ball to
    the desired point p at the required speed.
    \param o Object to which through pass should be given
    \param posEnd position that together with ball position defines line
              segment on which shooting point should be determined.
    \param *angMax will be filled with the largest angle between the opponents
    \return SoccerCommand to give a throughPass to 'o' */
SoccerCommand BasicPlayer::throughPass( ObjectT o, VecPosition posEnd,
                                        AngDeg *angMax )
{
  VecPosition posShoot = getThroughPassShootingPoint( o, posEnd, angMax );
  double      dEnd     = getEndSpeedForPass( o, posShoot );

  return kickTo( posShoot, dEnd );
}

/*! This skill enables an agent to outplay an opponent. It is used, for
    example, when an attacker wants to get past an enemy defender. This is
    done by passing the ball to open space behind the defender in such a
    way that the attacker can beat the defender to the ball. Note that the
    attacker has an advantage in this situation, since he knows to which
    point he is passing the ball and is already turned in the right
    direction, whereas the defender is not. As result, the attacker has a
    headstart over the defender when trying to intercept the ball. Since a
    player can move faster without the ball, the main objective is to kick
    the ball as far as possible past the opponent while still being able to
    reach it before the opponent does. This skill receives two arguments,
    'pos' and 'o', which respectively denote the point to which the agent
    wants to move with the ball and the object (usually an opponent) that the
    agent wants to outplay in doing so. First it is determined if it is
    possible to outplay the opponent o in the current situation. Let l be the
    line segment that runs from the agent's current position to the given
    point 'pos'. The best point to kick the ball to will be the furthest point
    from on this line that the agent can reach before the opponent does. We
    use a simple geometric calculation to find the point s on l which has
    equal distance to the agent and to the opponent. Let o' be the
    perpendicular projection of the opponent's position onto l and let d1,
    d2 and d3 respectively denote the distance between the agent poisition and
    o', the distance between o' and o and the distance between o' and s. To
    determine s we need to compute the unknown value of d3 using the values
    for d1 and d2 which can be derived from the world model. Since the
    distance from the agent position to s will be equal to the distance
    from o to s. Using this value for d3 we can compute the coordinates of
    the shooting point s. However, in some situations it is not likely that
    shooting the ball to this point will eventually result in outplaying
    the given opponent o on the way to 'pos'. We therefore use the values
    for d1, d2 and d3 to determine whether it is possible to outplay o in
    the current situation, and if so, what the best shooting point will be.
    The following situations are distinguished:
    - d1 + d3 > OutplayMinDist.
    If this condition holds, the opponent is located at a relatively large
    distance from the agent which makes an attempt to outplay him likely to
    be successful. First it is checked whether the agent's body is turned
    suffiiently towards the point 'pos'. If this is not the case then the
    turnWithBallTo skill is used to turn with the ball in the right
    direction. Otherwise, the kickTo skill is used to kick the ball to a
    point on the line l where the agent will be able to intercept it first.
    Note that in general the agent will be able to reach the point s before
    the opponent despite the fact that both players need to travel the same
    distance to this point. This is because the agent has already turned his
    body more or less towards 'pos' (and thus towards s), whereas the opponent
    probably has not. However, the actual point to which the ball is kicked
    is chosen slightly closer to the agent than the point s in order to be
    absolutely sure that he can intercept the ball before the opponent does.
    This safety margin is represented by the parameter OutplayBuffer which has
    a value of 2.5 in our current implementation. For this skill the
    desired end speed when the ball reaches z equals OutplayKickEndSpeed
    (=0.5). Note that the value for this parameter cannot be chosen too
    low, since this will cause the agent to intercept the ball before it
    reaches the target point
    - d1 + d3 < OutplayMinDist and d1 < d2.
    If this condition holds, the opponent is located close to the agent
    which makes it dicult to outplay him. However, if the agent is already
    turned in the right direction (i.e. towards 'pos') then d1 < d2 implies
    that the distance between the opponent and the line l (denoting the
    agent's desired movement trajectory) is large enough for the agent to
    outplay this opponent when the ball is kicked hard in the direction of
    'pos' (i.e. further ahead than s). This is because the agent can start
    dashing after the ball immediately, whereas the opponent still has to
    turn in the right direction. As a result, the agent will have dashed
    past the opponent by the time the latter has turned correctly and this
    puts him in a good position to intercept the ball before the opponent.
    Therefore it is checked first whether the agent's body is sufficiently
    turned towards 'pos' and if this is not so then the turnWithBallTo skill
    is used toturn with the ball in the right direction. Otherwise, the
    kickTo skill is used to kick the ball past the opponent. In this case the
    point to which the ball is kicked either lies OutplayMaxDist (=20) metres
    ahead of the agent's current position into the direction of 'pos' or
    equals 'pos' when the distance to 'pos' is smaller than this value. In
    all other cases (i.e. d1 +d3 < OutplayMinDist and d1 > d2) this
    skill returns an illegal command to indicate that it is not
    possible to outplay the opponent o on the way to the point 'pos'.
    \param o opponent object that should be outplayed
    \param pos position to which agent wants to move while outplaying 'o'
    \param *posTo position to which ball will be shot
    \return SoccerCommand to outplay object o on the way to 'pos',
       CMD_ILLEGAL when this is not possible. */
SoccerCommand BasicPlayer::outplayOpponent( ObjectT o, VecPosition pos,
                                            VecPosition *posTo )
{
  // future: take more than one opponent into account

  VecPosition posAgent   = WM->getAgentGlobalPosition();
  AngDeg      angBody    = WM->getAgentGlobalBodyAngle();

  AngDeg      ang        = (pos - posAgent).getDirection();
  // if not heading in the desired direction, first turn with the ball
  AngDeg angTmp = VecPosition::normalizeAngle( ang - angBody ) ;  
  if( fabs( angTmp ) > PS->getDribbleAngThr() )
    return turnWithBallTo( ang, PS->getTurnWithBallAngThr(),
                                PS->getTurnWithBallFreezeThr() );
// NEW
//  ang = WM->getAgentGlobalBodyAngle();
  Line        l          = Line::makeLineFromPositionAndAngle(posAgent,ang);
  VecPosition posObj     = WM->getGlobalPosition( o );
  VecPosition posProj    = l.getPointOnLineClosestTo( posObj );
  double      dDistOpp   = posProj.getDistanceTo( posObj );
  double      dDistAgent = posProj.getDistanceTo( posAgent );
  VecPosition posShoot;

  // we want to know when distance from ball to point p equals distance
  // from opp to point p :
  // d1 + d3 = sqrt(d2^2 + d3^2) > (d1+d3)^2 = d2^2 + d3^2 =>
  // d1^2 + 2*d1*d3 = d2^2 -> d3 = (d2^2 - d1^2 ) / 2*d1
  double dCalcDist;
  if( o != OBJECT_ILLEGAL )
  {
    dCalcDist = (dDistOpp*dDistOpp-dDistAgent*dDistAgent)/(2*dDistAgent);
    dCalcDist += dDistAgent;
  }
  else
    dCalcDist = 20;


  Log.log( 552, "outplay opponent %d, calc: %f, opp: %f, agent:  %f",
     SoccerTypes::getIndex( o ) + 1, dCalcDist, dDistOpp, dDistAgent );
  Log.log( 560, "outplay opponent %d, calc: %f, opp: %f, agent:  %f",
     SoccerTypes::getIndex( o ) + 1, dCalcDist, dDistOpp, dDistAgent );

  if( dCalcDist > 7.0 ) // if point far away, kick there
  {
    dCalcDist = min( posAgent.getDistanceTo( pos ), dCalcDist - 2.5 );
    dCalcDist = min( 20, dCalcDist );
    posShoot  = posAgent + VecPosition( dCalcDist, ang, POLAR );
  }
  else if( dDistAgent < dDistOpp - 0.3 ) // point close and well-positioned
  {                                      // shoot far away and outplay opp
    dCalcDist = min( posAgent.getDistanceTo( pos ), 20);
    posShoot = posAgent + VecPosition( dCalcDist, ang, POLAR );
  }
  else                                   // opponent stands in line
    return SoccerCommand( CMD_ILLEGAL );

  if( posShoot.getDistanceTo( WM->getAgentGlobalPosition() ) < 2.5 )
  {
    Log.log( 552, "calculated point too close" );
    Log.log( 560, "calculated point too close (%f,%f)", posShoot.getX(),
     posShoot.getY() );
    return SoccerCommand( CMD_ILLEGAL );
  }
  else if( WM->getNrInSetInCone( OBJECT_SET_OPPONENTS,PS->getConeWidth(),
    posAgent, posShoot ) != 0 )
  {
    Log.log( 552, "outplay: is opponent in cone" );
    Log.log( 560, "outplay: is opponent in cone (%f,%f)", posShoot.getX(),
     posShoot.getY() );
    return SoccerCommand( CMD_ILLEGAL );
  }

  if( posTo != NULL )
    *posTo = posShoot;
  return kickTo( posShoot, 0.5 );
}

/*! This skill enables an agent to clear the ball to a certain area on the
    field. It is useful, for example, when a defender cannot dribble or pass
    the ball to a teammate in a dangerous situation. Using this skill he can
    then kick the ball up the field away from the defensive zone. It is
    important to realize that this skill is only called when the agent has no
    alternative options in the current situation. Clearing the ball amounts to
    kicking it with maximum power into the widest angle between opponents in
    a certain area. The shooting direction is determined using the function
    which returns the direction of the bisector of this widest angle. The area
    on the field from which this angle is selected depends on the type of
    clear which is supplied as an argument to this skill. We distinguish
    three types of clearing:
    - CLEAR BALL DEFENSIVE: clearing the ball away from the defensive zone
      into a triangular area which is defined by the current ball position
      and the center line on the field.
    - CLEAR BALL OFFENSIVE: clearing the ball towards the offensive zone into
      a triangular area which is defined by the current ball position and the
      line segment that coincides with the front line of the opponent's
      penalty area and extends to the left and right side lines.
    - CLEAR BALL GOAL: clearing the ball into a triangular area in front of the
      opponent's goal which is defined by the current ball position and the
      line segment that runs from the center of the opponent's goal to the
      center of the front line of the penalty area.
    \param type type of the clear ball skill
    \param s if specified indicates the part of the field the clearBall should
            be directed to.
    \param angMax if specified (and not NULL) will be filled with the angle
           between the opponents in the direction in which will be shot
    \return SoccerCommand kick command to clear the ball */
SoccerCommand BasicPlayer::clearBall( ClearBallT type, SideT s, AngDeg *angMax )
{
  VecPosition posBall = WM->getBallPos();
  VecPosition posLeft, posRight;
  double      clearDist = PS->getClearBallDist();

  double      dPitchY = PITCH_WIDTH / 2.0;
  if( type == CLEAR_BALL_DEFENSIVE )
  {
    posLeft.setVecPosition ( 0, - dPitchY + clearDist );
    posRight.setVecPosition( 0, + dPitchY - clearDist );
  }
  else if( type == CLEAR_BALL_OFFENSIVE )
  {
    posLeft.setVecPosition ( PENALTY_X - clearDist, - dPitchY + clearDist );
    posRight.setVecPosition( PENALTY_X - clearDist, + dPitchY - clearDist );
  }
  else if( type == CLEAR_BALL_OFFENSIVE_SIDE )
  {
    posLeft.setVecPosition ( PENALTY_X - clearDist, - dPitchY + 8 );
    posRight.setVecPosition( PENALTY_X - clearDist, + dPitchY - 8 );
  }
  else if( type == CLEAR_BALL_GOAL && posBall.getY() > 0 )
  {
    posLeft.setVecPosition ( max(PENALTY_X - 2.0, posBall.getX()-10),  0.0 );
    posRight.setVecPosition( PITCH_LENGTH/2.0 - 5.0, 0.0 );
  }
  else if( type == CLEAR_BALL_GOAL && posBall.getY() < 0 )
  {
    posLeft.setVecPosition ( PITCH_LENGTH/2.0 - 5.0, 0.0 );
    posRight.setVecPosition( max(PENALTY_X - 2.0, posBall.getX()-10),  0.0 );
  }
  else
    return SoccerCommand( CMD_ILLEGAL );

  if( type != CLEAR_BALL_GOAL && s == SIDE_RIGHT ) // take only right part of
  {
    if( type == CLEAR_BALL_OFFENSIVE_SIDE )
      posLeft.setY( 15.0 );
    else
      posLeft.setY ( 0.0 );                          // field into account
    
  }
  else if( type != CLEAR_BALL_GOAL && s == SIDE_LEFT )
  {
    if( type == CLEAR_BALL_OFFENSIVE_SIDE )
      posRight.setY( -15.0 );
    else
      posRight.setY( 0.0 );
  }

  // get angle of ball with left and right points
  // get the largest angle between these two angles
  AngDeg angLeft  = (posLeft  - posBall).getDirection();
  AngDeg angRight = (posRight - posBall).getDirection();
  double dDist;
  if( type != CLEAR_BALL_GOAL )
    dDist = PS->getClearBallOppMaxDist();
  else
    dDist = max(posBall.getDistanceTo(posLeft),posBall.getDistanceTo(posRight));
  AngDeg ang      = WM->getDirectionOfWidestAngle( posBall, angLeft, angRight,
                                        angMax, dDist );

  Line l1 = Line::makeLineFromPositionAndAngle( posBall, ang );
  Line l2 = Line::makeLineFromTwoPoints( posLeft, posRight );
  VecPosition posShoot = l1.getIntersection( l2 );
  Log.log( 560, "angLeft %f, right %f, best %f point (%f,%f)",
  angLeft, angRight, ang, posShoot.getX(), posShoot.getY() );
  if( type == CLEAR_BALL_GOAL  )
    return kickTo( posShoot, SS->getBallSpeedMax() );
  else if( type == CLEAR_BALL_OFFENSIVE )
    return kickTo( posShoot, 0.25 );
  else if( type == CLEAR_BALL_OFFENSIVE_SIDE )
  {
    return kickTo( posShoot, (posBall.getX()>20) ? 1.2 : 2.7);
  }
  else
    return kickTo( posShoot, 0.5 );
}

/*! This skill enables an agent to mark an opponent, i.e. to guard him
    one-on-one with the purpose to minimize his usefulness for the opponent
    team. It can be used, for example, to block the path from the ball to
    an opponent or from an opponent to the goal. In this way the agent can
    prevent this opponent from receiving a pass or from moving closer to the
    goal while also obstructing a possible shot. This skill amounts to
    calculating the desired marking position based on the given arguments
    and then moving to this position. It receives three arguments: an
    object o (usually an opponent) that the agent wants to mark, a distance
    'dDist' representing the desired distance between o and the marking
    position and a type indicator that denotes the type of marking that is
    required. We distinguish three types of marking:
    - MARK BALL: marking the opponent by standing at a distance 'dDist' away
                 from him on the line between him and the ball. This type of
                 marking will make it dicult for the opponent to receive a
                 pass.
    - MARK GOAL: marking the opponent by standing at a distance 'dDist' away
                 from him on the line between him and the center point of the
                 goal he attacks. This type of marking will make it difficult
                 for the opponent to score a goal.
    - MARK BISECTOR: marking the opponent by standing at a distance 'dDist'
                 away from him on the bisector of the ball-opponent-goal
                 angle. This type of marking enables the agent to intercept
                 both a direct and a leading pass to the opponent.

    After determining the marking position, the agent uses the moveToPos skill
    to move to this position. Note that the decision whether to turn or dash
    in the current situation depends on the angle of the marking position
    relative to the agent's body direction and on the distance to this position
    if this point lies behind the agent. In this case the moveToPos skill uses
    the threshold parameters MarkTurnAngle (=30) and MarkDistanceBack (=3) to
    make this decision. The values for these parameters are such that the
    condition which must hold for allowing a dash is fairly flexible. This is
    done because the marking position will be different in consecutive cycles
    due to the fact that the opponent and the ball move around from each
    cycle to the next. As a result, the agent will be able to actually
    progress towards a point that lies close to the marking position instead
    of constantly turning towards the newly calculated marking position in
    each cycle.
    \param o object that has to be marked
    \param dDist distance marking position is located from object position
    \param mark marking technique that should be used
    \return SoccerCommand to mark object 'o'. */
SoccerCommand BasicPlayer::mark( ObjectT o, double dDist, MarkT mark )
{

  VecPosition posMark  = getMarkingPosition( o, dDist, mark );
  VecPosition posAgent = WM->getAgentGlobalPosition();
  VecPosition posBall  = WM->getGlobalPosition( OBJECT_BALL );
  AngDeg      angBody  = WM->getAgentGlobalBodyAngle();

  if( o == OBJECT_BALL )
  {
    if( posMark.getDistanceTo( posAgent ) < 1.5 )
      return turnBodyToObject( OBJECT_BALL );
    else
     return moveToPos( posMark, 30.0, 3.0, false );
  }

  // if marking position is outside pen. area and marking position is a small
  // distance further to the center, do not move but wait for opponent.

  // PLAYERS WON'T MARK FORWARD FROM THEM!
  /*  if(0 && posMark.getX() > -PENALTY_X && posAgent.getX() < posMark.getX())
  {
    if( posMark.getX() - posAgent.getX() < 4.0 &&
        posMark.getDistanceTo( posAgent ) < 4.0 )
    {
      Log.log( 455, "%do not move upwards on field to opponent, but wait" );
      return turnBodyToPoint( posAgent - VecPosition ( 10, 0 ));
    }
    Log.log( 455, "opponent upwards on field, so change x" );
    posMark.setX( posAgent.getX() );
  }
  else
  {*/
    posAgent = WM->predictAgentPos( 1, 0 );
    Line lineMe = Line::makeLineFromPositionAndAngle( posAgent, angBody );
    VecPosition posOpp = WM->getGlobalPosition( o );
    posOpp.setX( posOpp.getX() - 1 );
    Line lineBall = Line::makeLineFromTwoPoints( posBall, posOpp );
    VecPosition posIntersect = lineMe.getIntersection( lineBall );
    if( posIntersect.getDistanceTo( posMark ) < 4.0 &&
        posIntersect.getDistanceTo( posAgent ) > 0.5 )
    {
      Log.log( 455, "move forward to approximated good marking point" );

      return moveToPos( posIntersect, 90, 3.0 );
    }
    //}

  double x = WM->getOffsideX();
  if( posMark.getX() >  x ) {
    posMark.setX( x - 2.0 );
  }

  if( WM->isOffsideUs( ) && posMark.getX() > posBall.getX() ) {
    posMark.setX( posBall.getX() - 1.0 );
  }

  if( posMark.getDistanceTo( posAgent )  < 2.0 )
  {
    Log.log( 455, "turn body to -180" );
    return turnBodyToPoint( posAgent - VecPosition ( 10, 0 ));
  }
  Log.log( 455, "%move to mark position (%f %f)",posMark.getX(),posMark.getY());

  return moveToPos( posMark, PS->getPlayerWhenToTurnAngle(), 3.0, false );
}


/*! This skill enables an agent (usually the goalkeeper) to defend his own goal
    line. To this end the agent moves to a position along a line l which runs
    parallel to the goal line at a small distance 'dDist' (supplied as an
    argument) from the goal. The guard point to which the agent moves depends
    on the predicted position of the ball in the next cycle and is chosen in
    anticipation of a future shot on goal. This means that the guard point is
    selected in such a way that it will be most difficult for the opponent team
    to pass the goalkeeper. To find this point we need to look at the angle that
    the ball makes with respect to the left and right goal posts and we need to
    determine which point on l covers this angle in the most optimal way, i.e.
    leaves an equal gap to the left and to the right of the goalkeeper. Let m
    be the line that goes through the center point of the goal line and through
    the predicted ball position in the next cycle. Since this line m bisects the
    above-mentioned angle, the optimal guard point on l can be computed by
    determining the intersection between l and m. Note that in our current
    implementation the goalkeeper always stays in front of the goal mouth to
    avoid leaving an open goal when the ball is passed to an opponent in the
    center of the penalty area. The computed guard point is therefore adjusted
    if it lies too far to the side . After computing the guard point the
    goalkeeper needs to move to this point while keeping sight of the ball.
    If the distance between the current goalkeeper position and the line l is
    larger than DefendGoalLineMaxDist (which has a value of 3.0 in our current
    implementation) the moveToPos skill is used to move directly towards the
    guard point. This can happen, for example, when the goalkeeper has moved
    forward from his line to intercept the ball and now has to move back to his
    line again. Note that the fourth argument supplied to the moveToPos skill
    equals true in this case to indicate that the goalkeeper wants to turn his
    back to the guard point in order to keep the ball in sight while moving.
    However, if the distance between the guard point and l is less than
    DefendGoalLineMaxDist then the moveToPosAlongLine skill is used to move
    along l to the guard point. This skill receives an argument `sign'
    representing a prediction of the agent's movement in the coming cycles.
    This value is used to adjust the agent's body direction when necessary.
    In this case it can be expected that the goalkeeper will move along l
    in the same direction as the ball and `sign' is therefore determined
    by looking at the ball velocity in cycle t.
    \param dDist distance before goal the goalkeeper will move along
    \return SoccerCommand to defend the goal line. */
SoccerCommand BasicPlayer::defendGoalLine( double dDist )
{
  // determine defending point as intersection keeper line and line ball-goal
  VecPosition posBall    = WM->getBallPos(); 
  VecPosition posAgent   = WM->getAgentGlobalPosition();
  Line        lineGoal   = Line::makeLineFromPositionAndAngle(
                            VecPosition( - PITCH_LENGTH/2.0 + dDist, 0 ), 90 );

  VecPosition posGoalLeft ( -PITCH_LENGTH/2.0, -SS->getGoalWidth()/2.0 );
  VecPosition posGoalRight( -PITCH_LENGTH/2.0,  SS->getGoalWidth()/2.0 );
  Line left    = Line::makeLineFromTwoPoints( posBall, posGoalLeft  );
  Line right   = Line::makeLineFromTwoPoints( posBall, posGoalRight );
  posGoalLeft  = left.getIntersection ( lineGoal );
  posGoalRight = right.getIntersection( lineGoal );
  double dDistLeft  = posGoalLeft.getDistanceTo( posBall );
  double dDistRight = posGoalRight.getDistanceTo( posBall );
  double dDistLine  = posGoalLeft.getDistanceTo( posGoalRight );
  VecPosition posDefend  = posGoalLeft+
      VecPosition( 0, (dDistLeft/(dDistLeft+dDistRight))*dDistLine);

  bool        bBallInPen = WM->isInOwnPenaltyArea( posBall );

  // do not stand further to side than goalpost
  if( fabs( posDefend.getY() ) > SS->getGoalWidth()/2.0 )
    posDefend.setY( sign(posDefend.getY())*SS->getGoalWidth()/2.0);

  // if too far away from line, move directly towards it
  double dDiff = ( bBallInPen == true ) ? 1.5 : 0.5;
  if( posDefend.getX() + dDiff < posAgent.getX()  )
  {
    Log.log( 553, "move backwards to guard point" );
    return moveToPos( posDefend, 30, -1.0, true ); // always backwards
  }
  else if( posDefend.getX() - dDiff > posAgent.getX() )
  {
    Log.log( 553, "move forward to guard point" );
    return moveToPos( posDefend, 30, -1.0 );       // always forward
  }

  // desired body angle is in direction of the ball
  // predicted movement direction in subsequent cycles is in moving dir. ball
  AngDeg  angDes;
  if( fabs( posBall.getY() - posDefend.getY() ) > 0.5 )
    angDes = sign( posBall.getY() - posDefend.getY() )*90.0;
  else
    angDes = sign( WM->getAgentGlobalBodyAngle() )*90.0;
  int     iSign     = sign( WM->getGlobalVelocity( OBJECT_BALL ).getY() );

  // move to position along line: when ball in penalty area, never adjust body
  // angle (value 3.0) and change trajectory when angle difference is > 7
  // when ball is outside pen. area, adjust angle to move to line when more
  // than 0.5 from desired line, adjust when angle diff > 2 with 12 degrees
  if( bBallInPen )
  {
    Log.log( 553, "move along line, with ball in penalty area" );
    return moveToPosAlongLine( posDefend, angDes, 3.0, iSign, 7.0, 12.0 );
  }
  else
  {
    Log.log( 553, "move along line, with ball outside penalty area (%f,%f) (%f,%f) %f",
      WM->getAgentGlobalPosition().getX(),
      WM->getAgentGlobalPosition().getY(),
      posBall.getX(), posBall.getY(), WM->getConfidence( OBJECT_BALL) );
    Log.log( 553, "%s", WM->strLastSeeMessage );
    return moveToPosAlongLine( posDefend, angDes, 0.5, iSign, 2.0, 12.0 );
  }
}

/*! This method returns a command to intercept a ball that is currently
    heading towards the goal. The current trajectory of the ball is determined
    and the goalkeeper positions himself on a point on this trajectory just
    before the goal.
    \return SoccerCommand to intercept a ball heading towards the goal. */
SoccerCommand BasicPlayer::interceptScoringAttempt( )
{
  SoccerCommand soc;
  VecPosition   posAgent = WM->getAgentGlobalPosition();
  VecPosition   posBall  = WM->getBallPos();
  Line          lineBall = Line::makeLineFromPositionAndAngle( posBall,
                                    WM->getBallDirection() );
  Line          lineKeeper = Line::makeLineFromPositionAndAngle( posAgent, 90 );
  bool          bInterceptAtSide = false;

  // first create intersection point between ball trajectory and rectangle
  // that is located just in front of the goal mouth
  VecPosition   posIntersect = lineBall.getIntersection( lineKeeper );
  if( fabs( posIntersect.getY() ) > SS->getGoalWidth()/2.0  )
  {
     VecPosition posGoalPost( -PITCH_LENGTH/2.0,
                         sign(posBall.getY())*(SS->getGoalWidth()/2.0 + 0.5));
     Line l = Line::makeLineFromPositionAndAngle( posGoalPost, 0 );
     posIntersect = lineBall.getIntersection( l );

     // if intersection point does not cross rectangle between the agent and
     // the goalpost, we just move to edge of the rectangle
     if( posIntersect.getX() < posGoalPost.getX()  ||
         posIntersect.getX() > posAgent.getX() )
       posIntersect.setVecPosition( posAgent.getX(), posGoalPost.getY() );
     else
       bInterceptAtSide = true; // interception point is at side of rectangle
  }

  // first try close intercept
  soc = interceptCloseGoalie();
  if( ! soc.isIllegal() )
  {
    Log.log( 553, "close intercept" );
// new na GO 2002
    return soc;
// end new na GO 2002    
  }

  // if far away from goal, just intercept ball
  if( fabs( -PITCH_LENGTH/2.0 - posAgent.getX() ) > 8.0 )
  {
    Log.log( 553, "I am far away from keeper line" );
    return intercept( true );
  }
  else if( posBall.getX() < posAgent.getX() &&
           fabs( posBall.getY() ) < SS->getGoalWidth()/2.0 + 2.0 )
  {
    Log.log( 553, "ball heading and ball behind me" );
    return intercept( true );
  }

  // move to interception point
  if( posIntersect.getDistanceTo( posAgent ) < 0.5 )
  {
     Log.log( 553, "close to intersection point keeperline" );
     soc = turnBodyToObject( OBJECT_BALL );
  }
  else if( sign( (posIntersect - posAgent ).getDirection() ) ==
           sign( WM->getAgentGlobalBodyAngle()             ) ||
       bInterceptAtSide == true)
  {
    Log.log( 553, "move forward to intersection point keeperline" );
    soc = moveToPos( posIntersect, 20, SS->getGoalWidth() );
  }
  else
  {
    Log.log( 553, "move backward to intersection point keeperline" );
    soc = moveToPos( posIntersect, 20, SS->getGoalWidth(), true );
  }

  return soc;
}

/********************** UTILITY METHODS **************************************/

/*! This method returns the shooting point for a through pass to objTeam.
    This point lies on the line l that is created of the estimated position of
    the teammate after 3 cycles and 'posEnd'. Between these two points the
    maximal angle with respect to the opponents is calculated and the widest
    angle is calculated (this value is stored in angMax afterwards). The
    intersection between this widest angle and the line l is returned as the
    shooting point.
    \param objTeam teammate to give a throughpass to
    \param posEnd end point to which throughpasses are considered
    \param angMax will be filled with the maximal angle between the opponents
    \return position to give a through pass to. */
VecPosition BasicPlayer::getThroughPassShootingPoint( ObjectT objTeam, VecPosition  
                                               posEnd, AngDeg  *angMax  )
{
  VecPosition posTeam = WM->getGlobalPosition( objTeam );
  return getShootPositionOnLine( posTeam, posEnd, angMax );
}

/*! This method uses an iterative scheme to compute the optimal interception
    point of the ball. A loop is executed in which the prediction
    methods are used to predict the position of the ball a number of cycles, say
    i, into the future and to predict the number of cycles, say n, that the
    agent will need to reach this position. This is repeated for increasing
    values of i until n < i in which case it is assumed thay the agent should be
    able to reach the predicted ball position before the ball does.
    This point is chosen as the interception point and the moveToPos skill is
    used to move towards this point.
    \param iCyclesToBall will be filled with the nr of cycles it will take the
                        ball to reach the returned position.
    \param isGoalie indicates whether the current player is a goalkeeper or not
    \return position to intercept the ball. */
VecPosition BasicPlayer::getInterceptionPointBall( int *iCyclesBall,
                                                   bool isGoalie )
{
  VecPosition posPred  = WM->getAgentGlobalPosition();
  VecPosition velMe    = WM->getAgentGlobalVelocity();
  double      dSpeed, dDistExtra;
  VecPosition posMe, posBall;
  AngDeg      ang, angBody, angNeck;
  Stamina     sta;
  double      dMaxDist;

  /*if( isGoalie && 
      !WM->isInOwnPenaltyArea(WM->predictPosAfterNrCycles( OBJECT_BALL, 45)))
  {
    *iCyclesBall = -1;
    return posBall;
    } jdavin: disabled this because it causes goalie to go to (0,0) sometimes
  *  moved this functionality to CoachablePlayer */

  dMaxDist = (isGoalie) ? SS->getCatchableAreaL() : SS->getMaximalKickDist();

  // predict the position of the agent when current velocity is propogated
  dSpeed     = WM->getAgentSpeed();
  dDistExtra = Geometry::getSumInfGeomSeries( dSpeed, SS->getPlayerDecay() );
  posPred   += VecPosition( dDistExtra, velMe.getDirection(), POLAR );

  // for each loop, check whether agent can reach ball in less cycles
  for ( int i = 0; i <= PS->getPlayerWhenToIntercept(); i++ )
  {
    // (re-)initialize all the needed variables
    // set ball prediction one further to get right in front of ball line
    velMe   = WM->getAgentGlobalVelocity();
    angBody = WM->getAgentGlobalBodyAngle();
    angNeck = WM->getAgentGlobalNeckAngle();
    posBall = WM->predictPosAfterNrCycles( OBJECT_BALL, i + 1 );
    posMe   = WM->getAgentGlobalPosition();
    ang     = (posBall - posPred).getDirection();
    ang     = VecPosition::normalizeAngle( ang - angBody );
    sta     = WM->getAgentStamina();
    int turn = 0;

    // as long as not correctly headed for point, simulate a turn command
    while (fabs(ang) > PS->getPlayerWhenToTurnAngle() && turn<5)
    {
      turn++;
      WM->predictStateAfterTurn( WM->getAngleForTurn( ang,velMe.getMagnitude() ),
                   &posMe, &velMe, &angBody, &angNeck, &sta             );
      ang      = (posBall - posPred).getDirection();
      ang      = VecPosition::normalizeAngle( ang - angBody );
    }

    if( turn > 1 )
    {
      Log.log( 502, "nr of turns needed: %d", turn );
    }

    // for cycles that are left over after turn(s), execute dash with full power
    for( ; turn < i; turn++ )
      WM->predictStateAfterDash(SS->getMaxPower(),&posMe,&velMe,&sta,angBody);

    // if in kickable distance or passed ball, we can reach the ball!
    if (posMe.getDistanceTo( posBall ) < dMaxDist  ||
       (posMe.getDistanceTo( posPred ) > posBall.getDistanceTo( posPred ) +
                                         dMaxDist) )
    {
      *iCyclesBall = i;
      return posBall;
    }
  }
  *iCyclesBall = -1;
  return posBall;
}

/*! This method returns the point on a line segment with which the ball has
    the largest angle with the surrounding opponents. It uses the method
    getDirectionOfWidestAngle. The line is determined by the two position
    p1 and p2. The returned position lies on this line and makes the largest
    angle with the opponents. The actual angle between the opponents is
    returned by 'angLargest'.
    \param p1 first position of line segment
    \param p2 second position of line segment
    \param *angLargest will be filled with the largest angle with the opponents
    \return VecPosition position on line that has the largest angle with the
            opponents */
VecPosition BasicPlayer::getShootPositionOnLine( VecPosition p1,
                                      VecPosition  p2, AngDeg *angLargest )
{
  VecPosition posBall  = WM->getBallPos();
  Line   line          = Line::makeLineFromTwoPoints( p1, p2 );
  double dRadius       = min( PS->getClearBallOppMaxDist(),
                              posBall.getDistanceTo( p2 )  );
  AngDeg angMin        = (p1 - posBall ).getDirection();
  AngDeg angMax        = (p2 - posBall ).getDirection();
  // not correct when line crosses -180 boundary, but will never happenk
  AngDeg angShoot      = WM->getDirectionOfWidestAngle(
                             posBall, min(angMin, angMax),
                                 max(angMin,angMax), angLargest, dRadius );
  Line   line2         = Line::makeLineFromPositionAndAngle( posBall,
                                                             angShoot );
  return line.getIntersection( line2 );
}

/*! This method returns the end speed for a pass. This end speed is
    determined using two parameters, 'o' and 'posPass' which
    respectively denote the object to which is passed and the position to
    which is passed. First it is determined how many server cycles 'o'
    needs to travel to position 'posPass'. Then the starting speed of the ball
    is determined when it wants to travel to 'posPass' in the same number of
    cycles. This starting speed is adjusted if the corresponding end speed
    of the ball in the passing point lies outside the range
    [MinPassEndSpeed .. MaxPassEndSpeed].
    \param o object to which the ball will be passed
    \param posPass desired passing point where o can intercept the ball
    \return end speed to give to the ball such that player can intercept
                the ball the best. */
double BasicPlayer::getEndSpeedForPass( ObjectT o, VecPosition posPass )
{
  // we want that the ball arrives at that point after length nr of cycles
  // where length is the nr of cycles it takes the player to get there.
  VecPosition posBall = WM->getBallPos();
  double      dDist   = posBall.getDistanceTo( posPass );
  double      dLength = WM->predictNrCyclesToPoint( o, posPass,
                                        PS->getPlayerWhenToTurnAngle() );
  double      dFirst  = WM->getFirstSpeedFromDist( dDist, dLength );
  if( dFirst > SS->getBallSpeedMax() )
    dFirst = SS->getBallSpeedMax();
  double dEnd         = WM->getEndSpeedFromFirstSpeed( dFirst, dLength );
  if( dEnd > PS->getPassEndSpeed() )
    dEnd = PS->getPassEndSpeed();
  else if( dEnd < 0.6 )
    dEnd = 0.6;
  else if( dLength > 10.0 )
    dEnd = 0.6;

  return dEnd;
}

/*! This method returns a global position on the field which denotes the position
    to mark object 'o'. It receives three arguments: an
    object o (usually an opponent) that the agent wants to mark, a distance
    'dDist' representing the desired distance between o and the marking
    position and a type indicator that denotes the type of marking that is
    required. We distinguish three types of marking:
    - MARK BALL: marking the opponent by standing at a distance 'dDist' away
                 from him on the line between him and the ball. This type of
                 marking will make it dicult for the opponent to receive a
                 pass.
    - MARK GOAL: marking the opponent by standing at a distance 'dDist' away
                 from him on the line between him and the center point of the
                 goal he attacks. This type of marking will make it difficult
                 for the opponent to score a goal.
    - MARK BISECTOR: marking the opponent by standing at a distance 'dDist'
                 away from him on the bisector of the ball-opponent-goal
                 angle. This type of marking enables the agent to intercept
                 both a direct and a leading pass to the opponent.
    \param o object that has to be marked
    \param dDist distance marking position is located from object position
    \param mark marking technique that should be used
    \return position that is the marking position. */
VecPosition BasicPlayer::getMarkingPosition( ObjectT o, double dDist, MarkT mark)
{
  VecPosition posBall  = WM->predictPosAfterNrCycles( OBJECT_BALL, 3 );
  VecPosition posGoal  = WM->getPosOwnGoal( );
  VecPosition pos      = WM->getGlobalPosition( o ) - VecPosition( 1.0, 0.0 );
  VecPosition posMark;
  AngDeg      ang, angToGoal, angToBall;

  if( mark == MARK_GOAL )                       // position in direction goal
  {
    angToGoal = (posGoal-pos).getDirection( );
    posMark   = pos + VecPosition( dDist, angToGoal, POLAR );
  }
  else if( mark == MARK_BALL )                  // position in direction ball
  {
    angToBall = (posBall-pos).getDirection( );
    posMark   = pos + VecPosition( dDist, angToBall, POLAR );
  }
  else if( mark == MARK_BISECTOR )             // position between ball and goal
  {
    angToBall = (posBall - pos).getDirection( );
    angToGoal = (posGoal - pos).getDirection( );
    ang       = getBisectorTwoAngles( angToBall, angToGoal );
    posMark   = pos + VecPosition( dDist, ang ,POLAR );
  }
  return posMark;
}



/* 
 * Returns true if it is time to turn, false otherwise
 * It is time to turn when we will be getting a new see message in the *next* cycle.
 * 
 * iTriCounter indicates whether we will be receiving a see msg in next cycle. 
 * based on trilearn code in SenseHandler::setTimeSignal
 */
bool BasicPlayer::isTimeToTurn(){
  int iTriCounter = SENSE->getiTriCounter();

  if( WM->getAgentViewFrequency() == 1.0 ) // VA_NORMAL AND VQ_HIGH (default)
    {
      if(iTriCounter % 3 == 0){ // receiving a see in first half of next cycle

	Log.log(544, "[isTimeToTurn]: %d: time to turn.", WM->getCurrentTime().getTime());
	return true;
      }
      else if(iTriCounter % 3 == 1){ // receiving see in 2nd half of next cycle
	Log.log(544, "[isTimeToTurn]: %d: time to turn.", WM->getCurrentTime().getTime());
	return true;
      }
      else{ // iTriCounter == 2 -> no see will arrive
	Log.log(544, "[isTimeToTurn]: %d: not time to turn.", WM->getCurrentTime().getTime());
	return false;
      }

    }
  else if( WM->getAgentViewFrequency() == 2.0 ) // VA_WIDE AND VQ_HIGH
    {
      // only get a see if iTriCounter == 0
      if(iTriCounter % 3 == 0){ 
	Log.log(544, "[isTimeToTurn]: %d: time to turn (wide).", WM->getCurrentTime().getTime());
	return true;
      }
      else{
	Log.log(544, "[isTimeToTurn]: %d: not time to turn (wide).", WM->getCurrentTime().getTime());
	return false;
      }
    }
  else                                     // VA_NARROW AND VQ_HIGH
    {
      Log.log(544, "[isTimeToTurn]: %d: time to turn (narrow).", WM->getCurrentTime().getTime());
      // we'll get a see every cycle.
      return true;
    }

}


/* 
 * Returns true a see message is arriving in the *current* cycle
 * 
 * iTriCounter indicates whether we will be receiving a see msg in next cycle. 
 * based on trilearn code in SenseHandler::setTimeSignal
 */
bool BasicPlayer::seeMsgArrivingThisCycle(){
  int iTriCounter = SENSE->getiTriCounter();

  if( WM->getAgentViewFrequency() == 1.0 ) // VA_NORMAL AND VQ_HIGH (default)
    {
      if(iTriCounter % 3 == 0){ // no see this cycle

	return false;
      }
      else if(iTriCounter % 3 == 1){ // receiving see in 1st half of this cycle
	return true;
      }
      else{ // iTriCounter == 2 -> receiving see in 2nd of half this cycle
	return true;
      }

    }
  else if( WM->getAgentViewFrequency() == 2.0 ) // VA_WIDE AND VQ_HIGH
    {
      // only get a see if iTriCounter == 1
      if(iTriCounter % 3 == 1){ 
	return true;
      }
      else{
	return false;
      }
    }
  else                                     // VA_NARROW AND VQ_HIGH
    {
      // we'll get a see every cycle.
      return true;
    }

}
