/*
 * CoachableAgentComm - class that is used by the coachable agent to communicate 
 *  with other agents through say messages. 
 *  Uses the coachable agent say message library (saymsglib) to encode and decode
 *  messages.
 *
 *
 * Creation Date: 12/3/03
 * Author: John Davin
 *
 */

#include "CoachableAgentComm.h"
#include "MySayMsgFactory.h"

#include "CommPicker.h"



CoachableAgentComm::CoachableAgentComm(WorldModel *wm){

  MyEncoder = new SayMsgEncoder();
  MyDecoder = new SayMsgDecoder(new MySayMsgFactory());
  WM = wm;

  attentionto = 0; // start out with focus on no players. 
  timeLastReceived = -1;

  //focusStrategy = new FocusTest(WM);
  focusStrategy = new FocusNormal(WM);

}

CoachableAgentComm::~CoachableAgentComm(){
  if(MyEncoder != NULL) delete MyEncoder;
  if(MyDecoder != NULL) delete MyDecoder;
  if(focusStrategy != NULL) delete focusStrategy;
}

/*
 * Return an encoded say message that we should say this cycle.
 *
 */
string CoachableAgentComm::getSayMsg(){
  string saymsg;

  CommPicker pickmsg(this);

  // getMaxSize() - getSize() tells the CommPicker how many units we have to use.
  pickmsg.pickMsgs(MyEncoder->getMaxSize() - MyEncoder->getSize());
  // adds messages to MyEncoder.


  saymsg = MyEncoder->getEncodedStr();
  MyEncoder->clear(); // so we can use the encoder again next time

  return saymsg;
}

/*
 * Return the number of the player we would like to listen to
 *
 * Return 100 if we do not need to change focus (our current attentionto 
 *   player is correct).
 * Return 0 if we want to turn off attention focus (attentionto off) in order
 *   to let the server randomly select a message for us.
 */
unsigned short CoachableAgentComm::getPlayerToFocusOn(){

  unsigned short newfocus = focusStrategy->getDesiredFocus();

  if( newfocus == (unsigned short)WM->getPlayerNumber())
    cerr << "ERROR: attempting to focus on self: "<<newfocus<<" "<<WM->getPlayerNumber()<<endl;
  assert(newfocus != (unsigned short)WM->getPlayerNumber()); // can't focus on self

  if(newfocus != attentionto)
    LOGOUT(51) << "[CoachableAgentComm]: switching focus to player "<<newfocus<<endl;

  attentionto = newfocus;

  // WorldModel::getFocusPlayerNum returns the player that sense_body told us we're focused on.
  if(WM->getFocusPlayerNum() != attentionto) // our focus player still isn't set.
    return attentionto;


  return 100; // don't change focus; keep current attentionto player.
}


/*
 * Input: an encoded say message received from player 'playernum'
 * Handle the message by decoding it and integrating the information
 * into our world model (or using the info in other ways). 
 */
void CoachableAgentComm::handleReceivedMsg(char *msg, unsigned short playernum){

  Log.logWithTime( 2, " processing say msg from %d: %s", playernum, msg);
  LOGOUT(51) << "\n[CoachableAgentComm]: "<<WM->getCurrentTime()<<": I ("<<WM->getPlayerNumber()
       <<") received from player "<<playernum<<": "<<msg<<endl;

  // check if we received this msg from the player that we are focused on
  if(playernum != attentionto && attentionto != 0)
    LOGOUT(51) << "[WARNING]: "<<WM->getCurrentTime()<<": I ("<<WM->getPlayerNumber()
	 << ") received from player "<<playernum<<" instead of "<<attentionto<<endl;


  // make sure we received a comm msg last cycle
  if( timeLastReceived > 0 && timeLastReceived < WM->getCurrentTime() - 1)
    cerr << "[WARNING]: player "<<WM->getPlayerNumber()<<" missed receiving a say (lastsay="
	 << timeLastReceived <<", curtime="<< WM->getCurrentTime()<<")\n"<<endl;

  timeLastReceived = WM->getCurrentTime();

  // decode and process the message: 
  string str(msg);

  bool success = MyDecoder->decodeStr(str, playernum);
  if(!success){ 
    cerr << "[CoachableAgentComm]: failed to decode string! "<<str<<endl; 
    // there may be usable msgs in the decoder, but we'll just return (and not use them)
    // because failure to decode the msg may mean it's corrupted or not following spec. 
    return;}


  SayMsgDecoder::DecodedMsgIterator iter = MyDecoder->getMsgIterator();
  for( ; iter != MyDecoder->getMsgIteratorEnd(); iter++){
    // iterates over SayMsgTypes* pointers. 
    SayMsgTypes *curunit = *iter;
    curunit->process();
    // calls the process function that you defined for the current SayMsgType 
  }


}

