/* Dongryeol Lee
 * 
 * CompareActionVisitor.C
 *
 * This class is the visitor class that will compare two actions and store
 * the result as a boolean truth value.
 *
 */


#include "clangutil.h"
#include "CompareActionVisitor.h"

using namespace rcss::clang;


/* Compare function for ActPos actions */
void CompareActionVisitor::visit(ActPos *)
{
  /* Equality isn't defined yet */
  result = false;
}

/* Compare function for ActHome actions */
void CompareActionVisitor::visit(ActHome *)
{
  /* This isn't implemented yet as well. May not make sense at all */
  result = false;
}

/* This action is obsolete, so ignore */
void CompareActionVisitor::visit(ActBallToReg *)
{

}

/* Same with this action */
void CompareActionVisitor::visit(ActBallToPlayer *)
{

}

/* Compare function for ActMark actions */
void CompareActionVisitor::visit(ActMark *act)
{
  /* First, check the type of action */
  if(actionType != ActionVisitor::MARK) {
    
    result = false;
    return;
  }

  /* Now, check if they share the same UNums. We can check this by
     using the inline subtraction function defined in the UNumSet
     A - B and B - A will be both empty if they are equal.
     
     However, just computing A - B is necessary since both UNumSets
     are guaranteed to have the same size after reduction */

  
  /* First retrieve the UNum from the ActPassUNum act */
  UNum::unum_t B = (act->getPlayers()).getIndexAt(0).getUNum();

  /* Then, retrieve the UNum from the player set */
  UNum::unum_t A = players.getIndexAt(0).getUNum();

  /* Check if the two UNums are equal */
  result = (A == B);

}

/* Compare function for ActMarkLinePlayer actions */
void CompareActionVisitor::visit(ActMarkLinePlayer *act)
{
  /* First, check the type of action */
  if(actionType != ActionVisitor::MARKLINEPLAYER) {
    
    result = false;
    return;
  }

  /* Now, check if they share the same UNums. We can check this by
     using the inline subtraction function defined in the UNumSet
     A - B and B - A will be both empty if they are equal.
     
     However, just computing A - B is necessary since both UNumSets
     are guaranteed to have the same size after reduction */


  /* First retrieve the UNum from the ActPassUNum act */
  UNum::unum_t B = (act->getPlayers()).getIndexAt(0).getUNum();

  /* Then, retrieve the UNum from the player set */
  UNum::unum_t A = players.getIndexAt(0).getUNum();

  /* Check if the two UNums are equal */
  result = (A == B);

}

/* This action will be reduced to ActMarkLinePlayer, so ignore */
void CompareActionVisitor::visit(ActMarkLineReg *)
{
  /* Undefined equality since reduction will reduce all of these
     actions to ActMarkLinePlayer actions */
  result = false;
}

/* Compare function for ActOffsidesLine actions */
void CompareActionVisitor::visit(ActOffsidesLine *)
{
  /* Equality on offsidesline???? */
  result = false;
}

/* This is an obsolete action, so ignore */
void CompareActionVisitor::visit(ActHetType *)
{

}

/* Ignore named action */
void CompareActionVisitor::visit(ActNamed *)
{

}

/* Same thing with ActMarkLineReg actions, this will be reduced to
   its UNUMSet equivalent actions, so ignore */
void CompareActionVisitor::visit(ActPassReg *)
{
  result = false;
}

/* Compare function for ActPassUNum actions */
void CompareActionVisitor::visit(ActPassUNum *act)
{
  /* First, check the type of action */
  if(actionType != ActionVisitor::PASSUNUM) {
    
    result = false;
    return;

  }

  /* Now, check if they share the same UNums. We can check this by
     using the inline subtraction function defined in the UNumSet
     A - B and B - A will be both empty if they are equal.

     However, just retrieving the UNum in each UNumSet and comparing them
     would be sufficient since both UNumSets are guaranteed to have only one
     UNum after reduction */

  /* First retrieve the UNum from the ActPassUNum act */
  UNum::unum_t B = (act->getPlayers()).getIndexAt(0).getUNum();

  /* Then, retrieve the UNum from the player set */
  UNum::unum_t A = players.getIndexAt(0).getUNum();

  /* Check if the two UNums are equal */
  result = (A == B);

}

/* Compare function for ActDribble actions */
void CompareActionVisitor::visit(ActDribble *)
{
  /* Maybe this isn't necessary since all the directions will be considered
     in the Coachable player's dribbling code */
  result = false;
}

/* Compare function for ActClear actions */
void CompareActionVisitor::visit(ActClear *)
{
  /* Same thing with Dribbling actions, so may not be necessary. Not sure
     How to compare two different clearing actions since ActClear is on
     regions anyways */
  result = false;
}

/* Compare function for ActShoot actions */
void CompareActionVisitor::visit(ActShoot *)
{
  /* Check the type of the shoot action */
  result = (actionType == ActionVisitor::SHOOT);
}

/* Compare function for ActHold actions */
void CompareActionVisitor::visit(ActHold *)
{
  /* Check the type of the hold action */
  result = (actionType == ActionVisitor::HOLD);
}

/* Compare function for ActIntercept actions */
void CompareActionVisitor::visit(ActIntercept *)
{
  /* Check the type of the action */
  result = (actionType == ActionVisitor::INTERCEPT);
}

/* Compare function for ActTackle actions */
void CompareActionVisitor::visit(ActTackle *act)
{
  /* First, check the type of action */
  if(actionType != ActionVisitor::TACKLE) {
    
    result = false;
    return;
  }

  /* Now, check if they share the same UNums. We can check this by
     using the inline subtraction function defined in the UNumSet
     A - B and B - A will be both empty if they are equal.
     
     However, just computing A - B is necessary since both UNumSets
     are guaranteed to have the same size after reduction */


  /* First retrieve the UNum from the ActPassUNum act */
  UNum::unum_t B = (act->getPlayers()).getIndexAt(0).getUNum();

  /* Then, retrieve the UNum from the player set */
  UNum::unum_t A = players.getIndexAt(0).getUNum();

  /* Check if the two UNums are equal */
  result = (A == B);

}
