/*
 * A Cond Visitor class which allows us to evaluate the truth value of the condition
 * (and store in member variable truthValue) and store variable bindings. 
 * Only conditions with UNUMs in them (and therefore Regions also) need to change variable bindings. 
 * 
 * John Davin
 */


#ifndef COND_VISITOR_C
#define COND_VISITOR_C


#include "CondVisitor.h"
#include "WorldModelAccessor.h"
#include "msgstore.h"

WorldModelAccessor MyWorldModel;
extern MsgStore<Cond*> defcondstorage;




bool CondVisitor::getEval() 
{ return truthValue; }

CondVisitor::~CondVisitor(){
}



/*
 * Helper function for CondPlayerPos handler
 * Return true if player is in the region, false otherwise. 
 */
bool CondVisitor::playerInRegion(const UNum& player, bool ourside, 
				 const RegionVisitor& regionvisitor) const {
  VecPosition pos;
  //VecPosition myballpos;

  // if we don't know where the player is for sure, be skeptical and 
  // assume it's not in the region
  if( ! MyWorldModel.playerConfidenceGood(player.getUNum(), ourside)){
    return false; 
  }

  /*
  if(Log.isInLogLevel(8)) 
    myballpos = MyWorldModel.getBallPos(); // just for debugging
  */

  pos = MyWorldModel.getGlobalPosition(player.getUNum(), ourside);

  /*
  if(Log.isInLogLevel(8)) 
    *logger << "[CondPlayerPos]: "<<player<<" got position "<< pos.getX()
	    <<" "<< pos.getY()<<endl;
  */

  if(regionvisitor.isPointInsideRegion(pos)){

    /*
    PR(8, 
      "[CondPlayerPos:playerInRegion]: Ball pos = %f %f Player pos = %f %f\n", 
       myballpos.getX(), myballpos.getY(),  pos.getX(), pos.getY() );
    */

    return true;
  }
  
  return false;
}


/*
 * Helper function to CondPlayerPos. 
 * Return the number of players from playerset that are in the region defined 
 * by regionvisitor
 * Input Requirement: regionvisitor should have *no* unbound variables in it. 
 */
int CondVisitor::checkPlayersInRegion(const UNumSet& playerset, 
				      bool ourside, 
				      const RegionVisitor& regionvisitor, 
				      VarBindings *bindings,
				      int min, int max) const
{
  int count = 0;

  // check if each player in the set is in the desired region
  for(UNumSet::const_iterator player = playerset.begin(); 
      player != playerset.end(); player++)       {

    if((*player).getVar() == "")  {
 
      // then it's *not* a variable (just a normal UNum)
      
      /*
      PR(8, "[CondPlayerPos]: handling player unum %d\n", (*player).getUNum());
      */

      if( playerInRegion(*player, ourside, regionvisitor))
	{
	  count++;
	}
      
    }
    else  {
 
      /* then we have a clang variable
	 bind this variable to all the players that are within the region, 
	 except players that are already explicitly specified in the unum set. 
	 eg, if 1,2,3 and 4 are in the region, then {X 1 2} will bind X to 3,4.
      */

      UNumSet varbindingset;

      for(UNum u = UNum::u1; u.getUNum() <= UNum::u11; u++)   {
	
	if(playerset.contains(u)) continue;
 
	// skip this player since its explicitly specified
	
	// if this player is in the region, add it to the variable's binding
	if(playerInRegion(u, ourside, regionvisitor))  {
	  varbindingset.add(u);
	  count++;
	}
	
      }
      
      bindings->add((*player).getVar(), varbindingset);


      /* NOTE FOR WHEN WE SWITCH TO VarBindingSet's: this binding needs to go 
	 into the same VarBinding
	 as any VarBinding the regionvisitor is using ( eg, ppos X (region Y)
	 ... need to make X,Y pairs)
      */
    }

    if (count >= min && max >= 11)
      return true;
    if (count > max)
      return false;

  }
  return (min <= count) && (count <= max);
}


/*
 * Determine if the players specified are within the region specified.
 * Return true if the number of players is between the min and max count. 
 */
void CondVisitor::visit( CondPlayerPos* cond ) 
{
  bool ourside = cond->isOurSide(); // true if our side, false if opp

  bool regionhasvars = false;

  Region *theregion = (Region*)cond->getRegion();
  RegionVisitor regionvisitor;
  theregion->accept(regionvisitor);

  regionhasvars = regionvisitor.hasVariables();
  VarBindings regionvars = regionvisitor.getBindings();

  // the minimum specified players required to be in the region: 
  int min = cond->getMinMatch();

  // the max players that can be in the region
  int max = cond->getMaxMatch();

  UNumSet playerset = cond->getPlayerSet();
  

  /* if the player set has 0 (uAll), then just map that case into 
     {1 2 3 ... 11}
  */
  if(playerset.contains(UNum::uAll)){

    UNumSet tempset;
    
    for(UNum u = UNum::u1; u.getUNum() <= UNum::u11; u++)

      tempset.add(u);

    // also add any variables the original player set has
    for(UNumSet::iterator player = playerset.begin(); 
	player != playerset.end(); player++)   {

      if((*player).getVar() != "") // variable
	tempset.add(*player);
    }

    playerset = tempset;
  }

  if(regionhasvars){

    // iterate over the region variables and bind them to each player number
    for(VarBindings::BindingIterator var = regionvars.Begin(); 
	var != regionvars.End(); var++)   {

      UNumSet thisvarbindingset;

      /*
      PR(8, 
	 "[CondPlayerPos]: testing region with var %s\n", 
	 (*var).variable.c_str());
      */

      for(UNum u = UNum::u1; u.getUNum() <= UNum::u11; u++)   {

	VarBindings bindthisplayer; // a single binding of { X -> u }
	UNumSet thisplayer;
	thisplayer.add(u);
	bindthisplayer.add((*var).variable, thisplayer);

	// now visit the region with this player binding
	RegionVisitor regionvisitor(bindthisplayer);
	theregion->accept(regionvisitor);

	if (checkPlayersInRegion(playerset, ourside, regionvisitor, 
				 &bindthisplayer, min, max))
	  {

	    //PR(8, "[CondVisitor:CondPlayerPos]: cond evaluated to true\n");

	    truthValue = true;
	  
	    myBindings->set_mergeWithUnion(bindthisplayer); // TODO: should this be add?
	  }

      }

      // we have now tested the region with the current variable as each player
      // add the region variable binding to our bindings
      //myBindings->add((*var).variable, thisvarbindingset);

      break; // this actually can't handle more than one region variable yet. 
      // if there are two or more region variables, we need to evaluate the 
      // region 
      // with bindings for all the variables at the same time. 
      // (number of permutations = 11^(# of variables)
    }

    // done evaluating ppos with bound region variables

    /*
    if(Log.isInLogLevel(8))  { 
      *logger << "[CondPlayerPos]: final bindings: " << endl;
      myBindings->PrintAll();
    }
    */


  }
  else{
    VarBindings playerbindings;

    if (checkPlayersInRegion(playerset, ourside, regionvisitor, 
			     &playerbindings, min, max))
      {
	//PR(8, "[CondVisitor:CondPlayerPos]: cond evaluated to true\n");
	truthValue = true;

	//myBindings->add(playerbindings);

	// NO! bug: we can not add() the bindings because that just appends the bindings 
	// to the end of the VarBindingSet. And there may already be a matching variable
	// bound to something. So we need to merge the bindings. -jdavin

	myBindings->set_mergeWithIntersect(playerbindings);

      }
    else
      {
      /*
	PR(8, 
	"[CondVisitor:CondPlayerPos]: players not within min and max count\n");
      */

      truthValue = false;
    }
  }


}


/*
 * Return: true if the current ball position is in the cond region
 *         false otherwise. 
 */
void CondVisitor::visit( CondBallPos* cond ) {

  truthValue = false; // set to false until we determine otherwise
  VecPosition ballpos = MyWorldModel.getBallPos();


  /*
  PR(8, "[CondVisitor:CondBallPos]: Ball is at %f, %f\n", ballpos.getX(), 
     ballpos.getY());
  */


  Region *condregion = (Region*)cond->getRegion();

  RegionVisitor regionvisitor;
  condregion->accept(regionvisitor);

  // then our region has variables in it, which we need to bind. 
  if(regionvisitor.hasVariables())   {
   
    /*
    PR(8, 
    "[CondBallPos]: Region had variables: (Iterating over regions with all possible players now)\n");
    */


    VarBindings regionvars = regionvisitor.getBindings();

    regionvars.PrintAll();

    for(VarBindings::BindingIterator var = regionvars.Begin(); 
	var != regionvars.End(); var++){

      /*
	PR(8, "[CondBallPos]: Binding variable %s\n\n\n\n", 
	(*var).variable.c_str());
      */


      UNumSet thisvarBindings; // the bindings for our current variable

      // iterate over the players: 
      for(UNum u = UNum::u1; u.getUNum() <= UNum::u11; u++)  {

	VarBindings bindthisplayer; // a single binding of { X -> u }
	UNumSet thisplayer;
	thisplayer.add(u);
	bindthisplayer.add((*var).variable, thisplayer);

	// now visit the region 
	RegionVisitor rvisitor(bindthisplayer);
	condregion->accept(rvisitor);

	/* if the ball is inside this region, then bind this player number to 
	   the variable */
	if(rvisitor.isPointInsideRegion(ballpos)){
	  truthValue = true; // the BallPos condition can be satisfied

	  thisvarBindings.add(u);

	  /*
	  if(Log.isInLogLevel(8)) 
	    *logger <<"[CondBallPos]: binding player "<< u
		    <<" to "<<(*var).variable<<"\n";
	  */
	}

      } // done iterating thru all players

      VarBindings bindings;
      bindings.add((*var).variable, thisvarBindings);

      // add the bindings to the CondVisitor's stored bindings:
      myBindings->set_mergeWithIntersect(bindings);

      //PR(8, "[CondBallPos]: myBindings are now: \n");

      //myBindings->PrintAll();

    } // done with bindings collected in the RegionVisitor

    return; // all possible player regions have been evaluated
  }
  // done with variables


  if(regionvisitor.isPointInsideRegion(ballpos))
    truthValue = true;
  else truthValue = false;

}

/*
 * Return true if the last player to touch the ball is listed in the UNumSet. 
 *
 */
void CondVisitor::visit( CondBallOwner* cond ) {

  bool ourside = cond->isOurSide(); // true if our side, false if opp

  const UNumSet players = cond->getPlayerSet();

  UNum::unum_t ballowner;
  bool team = MyWorldModel.getBallOwner(&ballowner);

  if(team != ourside){  // then ball owner team does not match the cond.
    truthValue = false;
    return;
  }

  truthValue = false; // stays false unless we find the ballowner 
                      // in the players set

  // check if the ballowner UNum is in the player UNumSet
  if( players.contains(ballowner) ){
      truthValue = true;
  }

  // now check if any of the UNum's are variables
  for(UNumSet::const_iterator i = players.begin(); i != players.end(); i++){

    if((*i).getVar() != ""){ // then the unum is a variable

      // first check to see if the variable is already bound
      UNumSet varset = myBindings->lookup((*i).getVar());

      // if the returned set is nonempty, check if the binding contains the 
      // current ballowner
      if (varset.begin() != varset.end() ) {
	if( ! varset.contains(UNum(ballowner))){
	  // variable binding does not contain ballowner, so this cond is false
	  truthValue = false;

	}
      }
      else 
	// then variable is not bound, so it could possibly be bound to 
	// ballowner, 
	// and cond would be true
	truthValue = true;


      // bind the variable to whoever is the current ball owner
      UNumSet ballowner_set;
      VarBindings bindings;
      ballowner_set.add(UNum(ballowner));
      bindings.add(*i, ballowner_set);
      myBindings->set_mergeWithIntersect(bindings);

    } // end of variable case

  }

}

/*
 * Return: true if current playmode matches the cond playmode. 
 */
void CondVisitor::visit( CondPlayMode* cond ) {

  // The WorldModelAccessor function getPlayMode queries the WorldModel
  // and returns a PlayMode.
  truthValue = (cond->getPlayMode() == MyWorldModel.getPlayMode());

}

/*
 * Look up up the named condition and return its truth value. 
 */
void CondVisitor::visit( CondNamed* cond ) {

  /* Get the Cond object referred by the CondNamed */
  Cond *namedcond = (Cond* ) cond->getCond();

  /* Visit this Cond object */
  namedcond->accept(*this);

}

/*
 * A CondTime is a list of times. It's a subclass of CondComp, 
 * which is a composition of conds.
 */
void CondVisitor::visit( CondTime* cond ) {

  const int curtime = MyWorldModel.getCurrentTime();

  const int time = cond->getValue();
  const util::CompOp *compop = &cond->getComp();

  // now compare the current time to the cond time, using whatever comparison 
  // the CompOp specifies
  truthValue =  compop->compare(curtime, time);

}

/*
 * Evaluate comparison relations on our goal score.
 */
void CondVisitor::visit( CondOurGoal* cond ) {
  
  // get opponent's goal score: 
  const int ourgoal = MyWorldModel.getOurGoal();

  const int condgoal = cond->getValue();
  const util::CompOp *compop = &cond->getComp();

  // compare the condition goal number to the actual goal #
  truthValue =  compop->compare(condgoal, ourgoal);

}

/*
 * Evaluate comparison relations on opponent's goal score.
 */
void CondVisitor::visit( CondOppGoal* cond ) {

  // get opponent's goal score: 
  const int oppgoal = MyWorldModel.getOppGoal();

  const int condgoal = cond->getValue();
  const util::CompOp *compop = &cond->getComp();

  // compare the condition goal number to the actual goal #
  truthValue =  compop->compare(condgoal, oppgoal);

}

/*
 * Evaluate comparison relations on the goal difference (ourgoal-oppgoal).
 */
void CondVisitor::visit( CondGoalDiff* cond ) {

  // get opponent's goal score: 
  const int actualdiff = MyWorldModel.getGoalDiff();

  const int conddiff = cond->getValue();
  const util::CompOp *compop = &cond->getComp();

  // compare the condition goal diff to the actual goal diff
  truthValue =  compop->compare(conddiff, actualdiff);

}

/*
 * If the unum is in the unum set, return true. 
 */
void CondVisitor::visit( CondUNum* cond ) {

  UNum unum = cond->getUNum();
  UNumSet set = cond->getUNumSet();
  const std::string Varname = unum.getVar();

  if(Varname == "")   { // then we have a nonvariable UNum

    truthValue = false; // stays false unless we find the unum in the set

    if(set.contains(unum)){
      truthValue =  true; 
    }

  }
  else{ // we have a variable unum

    // add the (unum variable, player set) pair to this condition's bindings
    VarBindings bindings;
    bindings.add(unum, set);
    myBindings->set_mergeWithIntersect(bindings);
    truthValue = true;

  }

}




/**** Basic boolean operations: ****/
void CondVisitor::visit( CondBool* cond ){
  
  truthValue = cond->getState();
  
}

void CondVisitor::visit( CondAnd* cond ){
  
  Cond *membercond;
  truthValue = true; 
  
  for( CondAnd::HasManyConds::Storage::const_iterator iter = 
	 (cond->getConds()).begin(); 
       iter != (cond->getConds()).end(); ++iter ) {
    
    if( *iter == NULL )
      throw util::NullErr( __FILE__, __LINE__, "Null condition in CondAnd\n" );
    
    membercond = *iter;
    CondVisitor condvis(myBindings);
    
    /*
      if(Log.isInLogLevel(8))
      { 
      *logger << "[CondAnd]: Evaluating cond: "<<endl;
      
      membercond->print(*logger); *logger << endl;
      }
    */
    
    membercond->accept(condvis);
    
    //PR(8, "[CondAnd]: Got bindings: \n");
    //condvis.getBindingSet().PrintAll();
    //myBindings->set_mergeWithIntersect(condvis.getBindingSet());
    
    //PR(8, "[CondAnd]: My bindings now: \n");
    //myBindings->PrintAll();
    
    truthValue &= condvis.getEval();
    
    /*
      if(condvis.getEval())
      PR(8, "[CondAnd]: true\n");
      else 
      PR(8, "[CondAnd]: false\n"); 
    */
    
    
    if(truthValue == false) break; // we're done - the And is false. 
    
  }

}

/*
 * An Or with variable bindings in it will produce a VarBindingSet, 
 *  ie. an Or'd sequence of each conditions variable bindings. 
 * but we may need to compact the VarBindingSet... 
 * Example: { x -> 1 }, { x -> 2} should be compacted to { x -> 1,2 }
 * We can *not* compact something like this: { x -> 1, y -> 6 }, { x -> 2, y -> 7 }
 * But we could compact something like this: { x -> 1, y -> 6 }, { x -> 1, y -> 7 } can be { x->1, y->6,7 }
 */
void CondVisitor::visit( CondOr* cond ){
  Cond *membercond;
  
  for( CondOr::HasManyConds::Storage::const_iterator iter = 
	 (cond->getConds()).begin(); 
       iter != (cond->getConds()).end(); ++iter ) {
    
    if( *iter == NULL )
      throw util::NullErr( __FILE__, __LINE__, "Null condition in CondOr\n" );
    
    membercond = *iter;
    
    VarBindingSet copyset = *myBindings;
    CondVisitor condvis(&copyset);
    
    /*
    if(Log.isInLogLevel(8)){ *logger << "[CondOr]: Evaluating cond: "<<endl;
    membercond->print(*logger); *logger << endl;}
    */


    membercond->accept(condvis);
    myBindings->add(copyset);

    truthValue |= condvis.getEval();
    //if(truthValue == true) break; // we're done - the Or is true. 
    
  }
  
  /*
  if(Log.isInLogLevel(8)){ *logger << "[CondOr]: final bindings: " << endl;
  myBindings->PrintAll();}
  */

}

void CondVisitor::visit( CondNot* cond ){

  // get the condition inside the Not and negate its truthvalue. 
  Cond *subcond =  (Cond*) cond->getCond();
  
  //PR(8, "[CondNot]: entered\n");
  
  VarBindingSet copyset = *myBindings;
  CondVisitor condvis(&copyset);
  subcond->accept(condvis);
  // add the inverse of the bindings we got in the condition (because we're 
  // taking the not
  // of the condition)

  *myBindings = *myBindings - copyset;

  truthValue = ! condvis.getEval();

  // but if the condition had variables, we want to count it as true if 
  // the final bindings have players that they can bind to. 
  if( myBindings->hasNonemptySets() )
    truthValue = true;
  
  /*
  if(Log.isInLogLevel(8)){ *logger << "[CondNot]: final bindings: " << endl;
  myBindings->PrintAll();}
  */
}





#endif // COND_VISITOR_C
