/*
 * Dir Visitor - used to handle the two Dir (directive) types. 
 *
 * John Davin
 */


#include "DirVisitor.h"
#include "msgstore.h"
#include "ActionPicker.h"
#include "WorldModelAccessor.h"
#include "Filter.h"
#include "Logger.h"

extern MsgStore<Dir*> defdirstorage; 
extern ActionPicker action_picker; // from msgrulestore.C
extern WorldModelAccessor MyWorldModel; // defined in CondVisitor.C
extern Filter filter; /* Defined in  msgrulestore.C */


/*
 * Return: true if any elements in the set are variables
 *                   false otherwise (if all elements are plain unum_t  types)
 */
bool ContainsVariables(const UNumSet& set)
{

  for(UNumSet::const_iterator i = set.begin(); i != set.end(); i++){

    if( (*i).getVar() != "")  // found a variable
      return true;
  }

  return false;
}


/*
 * Given a source UNumSet and a set of variable bindings, expand any variable bindings in the UNumSet
 * and return the new flattened set. 
 */
UNumSet FlattenSet(const UNumSet& srcSet, 
		   const VarBindings& bindings, int count)
{

  UNumSet newset;

  for(UNumSet::const_iterator i = srcSet.begin(); i != srcSet.end(); i++)  {

    if((*i).getVar() != ""){ // then we have a variable

      UNumSet varset = bindings.lookup((*i).getVar());

      // check if the returned set is empty: 
      if (varset.begin() == varset.end() ) {
	
	/*
        PR(9, "[FlattenSet]: Variable %s is not bound in the variable bindings!\n", (*i).getVar().c_str() ); 
	*/

        return newset;
      }

      // iterate over this UNumSet, adding the bindings to the newset
      for(UNumSet::const_iterator i = varset.begin(); i != varset.end(); i++){
        newset.add( *i);
      }

    }
    else{ // just a regular UNum
      newset.add( *i );

    }
  } // end of iterating over srcSet
  
  // we need to bound the number of recursive calls to prevent infinite loops
  // for example, if someone did (unum X {X}), that would cause infinite 
  // lookups
  // something like (and (unum X {0}) (unum Y {X})) is fine though. 
  if(count > MAX_VARIABLE_EXPANSIONS)  {

    /*
    PR(9, "[FlattenSet]: WARNING!: Variable expansion exceeded max recursion. Possible infinite loop in a UNumSet?\n");
    */

    return newset;
  }

  // if the newset still has unexpanded variables (they were nested), then 
  // recursively flatten again
  if( ContainsVariables(newset))
    return FlattenSet(newset, bindings, ++count);
  
  return newset;

}




void DirVisitor::visit( DirComm* dir )
{ 
  Action *curact;
  bool found = false;
  
  /* Determines whether the directive is negative or not */
  bool isPositiveDir = dir->isPositive();

  
  //PR(4, "[DirVisitor]: Got DirComm\n");


  /* Throw out all directives pertaining to the opponent team */
  if(!dir->isOurSide()){
    actionlog(30) << "\n\nERROR\n\n\n\n\n\n"<< spades::ende;
    return;
  }
  
  
  // then it's our side
  // check if this player is in the set... 
  UNumSet playerlist = dir->getPlayerSet();
  UNum thisplayer = MyWorldModel.getMyNum();


  // thisplayer will be uMax if the player has not yet received a 
  // server initialization message designating our player number
  if(!thisplayer.isValid())   { 
    
    //PR(9, "[DirVisitor:ERROR]: our player number is not set\n");
    actionlog(30) << "\n\nERROR Playernum\n\n\n\n\n\n"<< spades::ende;
    return; 
  }


  // if the VarBindingSet is empty, add an empty VarBindings to it so that 
  // we iterate through the for loop once
  if(myBindings.Begin() == myBindings.End()){

    VarBindings empty;
    myBindings.add(empty);
  }

  // for each VarBindings in the VarBindingSet myBindings, add an action with 
  // that binding
  for(VarBindingSet::BindingSetIter bindings = myBindings.Begin(); 
      bindings != myBindings.End(); bindings++){

    // now create a new UNumSet that has any variables expanded to their 
    // bindings in myBindings
    // for now, select a VarBindings from the VarBindingSet using the 
    // algorithm specified in VarBindingSet (probably random)
    UNumSet expandedset = FlattenSet(playerlist, *bindings, 0);


    /*
    if(Log.isInLogLevel(7)) 
      *logger <<"[" << thisplayer <<"]: " << "expanded direc set: " 
	      <<expandedset << " orig:" << playerlist << endl;
    */


    // check if the ball owner is in our set of players
    if( expandedset.contains(thisplayer)){
      
      /*
	if(Log.isInLogLevel(7)) 
	*logger << "found " << thisplayer << " in direc set\n";
      */


      found = true; 
    }

    if(!found){ 
	actionlog(30) << "Playernum not in set on dir:  "<< *dir<< spades::ende;

	continue;} // then the player wasn't in this set

    // add all of the actions to our action picker: 

    DirComm::HasManyActions::Storage actions = dir->getActions();

    
    /* If the directive is positive, */
    if(isPositiveDir) {
      
      for( DirComm::HasManyActions::Storage::iterator i = actions.begin();
	   i != actions.end(); i++){
	
	curact = *i;

	actionlog(30) << "Added action to actionpicker: "<< *curact<< spades::ende;

	if(Log.isInLogLevel(4)){ 
	  *logger <<"Added to actionpicker: "; curact->print(*logger); *logger <<endl;}

	/* Adding to action picker */
	action_picker.addAction(curact, *bindings, sourceType);
      }
    }
    else {
      
      for( DirComm::HasManyActions::Storage::iterator i = actions.begin();
	   i != actions.end(); i++){
	
	curact = *i;
	
	/* Otherwise, add to our fliter list */
	filter.addDoNotAdvice(curact, *bindings, sourceType);

      }
    }
    
    
  }

}



void DirVisitor::visit( DirNamed* dir ){ 
  PR(4, "[DirVisitor]: Got DirNamed\n");

  // look up the corresponding dir in defdirstorage and call visit on it. 
  Dir *thedir = (Dir *) dir->getDir();


  /* Visit the directive */
  thedir->accept(*this);

}


