/* Dongryeol Lee
 *
 * Level.h
 *
 * This is the header file for the implementation of the Level class. Each
 * level class object represents a different type of action, with different
 * predicates to satisfy: world model state/action-specific predicates
 *
 */



#ifndef LEVEL
#define LEVEL


#include <vector>
#include "clangaction.h"
#include "WorldStatePredicate.h"
#include "ActionPredicate.h"
#include "ActionBinding.h"
#include "ActionVisitor.h"
#include "VarBindings.h"


class Level
{
 public:

  /* Default constructor for the level object: Sets the source to ANY */
  Level() {
    
    sourceType = ActionBinding::ANY;
  }


  /* Constructor for the level object */
  Level(ActionVisitor::ActionType atype, ActionBinding::SourceType stype):
    actionType(atype), sourceType(stype)
    {}


  /* Boolean method for testing whether the action matches the current world
     state */
  bool matchWorldState() const;


  /* Boolean method for testing whether the action is of the right type */
  bool matchActionType(rcss::clang::Action *act) const;


  /* Boolean method for testing whether the action is of the right source
     type */
  bool matchSourceType(ActionBinding::SourceType s) const;


  /* Boolean method for testing whether the action matches the action-specific
     properties of the level */
  bool matchAction(rcss::clang::Action *act, const VarBindings &v) const;


  /* Method to add world-state predicates to the level */
  Level addWorldStatePredicate(WorldStatePredicate *p);


  /* Method to add action predicates to the level */
  Level addActionPredicate(ActionPredicate *p);



  
  /* Debugging method to print out the representation of the level */
  std::ostream& print( std::ostream& out) const
    {
      /* First we need to print out the action type */
      switch(actionType) {
	
      case ActionVisitor::POS:
	out << "ActPos            ";
	break;
      case ActionVisitor::HOME:
	out << "ActHome           ";
	break;
      case ActionVisitor::MARK:
	out << "ActMark           ";
	break;
      case ActionVisitor::MARKLINEPLAYER:
	out << "ActMarkLinePlayer ";
	break;
      case ActionVisitor::MARKLINEREG:
	out << "ActMarkLineReg    ";
	break;
      case ActionVisitor::OFFSIDESLINE:
	out << "ActOffsidesLine   ";
	break;
      case ActionVisitor::PASSREG:
	out << "ActPassReg        ";
	break;
      case ActionVisitor::PASSUNUM:
	out << "ActPassUNum       ";
	break;
      case ActionVisitor::DRIBBLE:
	out << "ActDribble        ";
	break;
      case ActionVisitor::CLEAR:
	out << "ActClear          ";
	break;
      case ActionVisitor::SHOOT:
	out << "ActShoot          ";
	break;
      case ActionVisitor::HOLD:
	out << "ActHold           ";
	break;
      case ActionVisitor::INTERCEPT:
	out << "ActIntercept      ";
	break;
      case ActionVisitor::TACKLE:
	out << "ActTackle         ";
	break;
      default:
	out << "Illegal Action    ";
	break;
      }
      
      /* Next, print out the source of the action */
      switch(sourceType) {
	
      case ActionBinding::DEFAULT:
	out << "Default Action   ";
	break;
      case ActionBinding::RULEFILE:
	out << "Rule File Action ";
	break;
      case ActionBinding::COACH:
	out << "Coach Action     ";
	break;
      case ActionBinding::ANY:
	out << "Any Action       ";
	break;
      case ActionBinding::COMMUNICATION:
	out << "Communication Action ";
	break;
      }
      
      /* Return the stream */
      return out << "\n";
    }

  
  /* Some useful accessor methods */
  ActionVisitor::ActionType getActionType() const { return actionType; }
  
  ActionBinding::SourceType getSourceType() const { return sourceType; }


 private:

  /* Tells what type of an action the level represents */
  ActionVisitor::ActionType actionType;

  /* Tells the source of the action */
  ActionBinding::SourceType sourceType;
  
  /* Lists of world state predicates an action has to satisfy to be in this
     level */
  std::vector <WorldStatePredicate *> worldStatePredicates;
  
  /* Lists of action-specific predicates an action has to satisfy to be in
     this level */
  std::vector <ActionPredicate *> actionPredicates;

};


std::ostream& operator<<( std::ostream & os, const Level& l );



#endif
