/*
Copyright (c) 2000-2002, Jelle Kok, University of Amsterdam
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
list of conditions and the following disclaimer. 

2. Redistributions in binary form must reproduce the above copyright notice, 
this list of conditions and the following disclaimer in the documentation 
and/or other materials provided with the distribution. 

3. Neither the name of the University of Amsterdam nor the names of its 
contributors may be used to endorse or promote products derived from this 
software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE 
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
/*! \file Player.C

<pre>
<b>File:</b>          Player.C
<b>Project:</b>       Robocup Soccer Simulation Team: UvA Trilearn
<b>Authors:</b>       Jelle Kok
<b>Created:</b>       03/03/2001
<b>Last Revision:</b> $ID$
<b>Contents:</b>      This file contains the definitions for the Player class,
               which is a superclass from BasicPlayer and contains the
               decision procedure to select the skills from the
               BasicPlayer. 
<hr size=2>
<h2><b>Changes</b></h2>
<b>Date</b>             <b>Author</b>          <b>Comment</b>
03/03/2001       Jelle Kok       Initial version created
</pre>
*/
#include "Player.h"
#include "Parse.h"
#include "region.h"

#include <sys/poll.h> // needed for 'poll'


/*!This is the constructor the Player class and calls the constructor of the
   superclass BasicPlayer.
   \param act ActHandler to which the actions can be sent
   \param wm WorldModel which information is used to determine action
   \param ss ServerSettings that contain parameters used by the server
   \param ps PlayerSettings that contain parameters important for the client
   \param strTeamName team name of this player
   \param dVersion version this basicplayer corresponds to
   \param iReconnect integer that defines player number (-1 when new player) 
   \param useGUI boolean that determines whether Coachable player talks to actions gui*/
Player::Player( ActHandler* act, WorldModel *wm, ServerSettings *ss,
		PlayerSettings *ps, Formations *fs, PlayerInterface *pi, 
		SenseHandler *sense, CoachableAgentComm *comm, ofstream *positionlog,
		char* strTeamName, double dVersion, int iReconnect, bool useGUI,
		ofstream *WMstatslog )

{
  char str[MAX_MSG];

  ACT           = act;
  WM            = wm;
  SS            = ss;
  PS            = ps;
  SENSE         = sense;
  formations    = fs;
  AgentComm     = comm;
  logpos        = positionlog;
  bContLoop     = true;
  WM->setTeamName( strTeamName );
  m_timeLastSay = -5;
  Player_interface = pi;
  statslog      = WMstatslog;


  InitCoachablePlayer(useGUI);

  // wait longer as role number increases, to make sure players appear at the
  // field in the correct order
  poll( 0, 0, formations->getPlayerInFormation()*100 );

  // create initialisation string
  if( iReconnect != -1 )
    sprintf( str, "(reconnect %s %d)", strTeamName, iReconnect );
  else if( formations->getPlayerType() == PT_GOALKEEPER )
    sprintf( str, "(init %s (version %f) (goalie))", strTeamName, dVersion );
  else
    sprintf( str, "(init %s (version %f))", strTeamName, dVersion );
  ACT->sendMessage( str );

}

/*! This is the main loop of the agent. This method calls the update methods
    of the world model after it is indicated that new information has arrived.
    After this, the correct main loop of the player type is called, which
    puts the best soccer command in the queue of the ActHandler. */
void Player::mainLoop( )
{
  // send our clang support command
  ACT->putCommandInQueue( SoccerCommand( CMD_CLANGVER, 8, 9 ) );

  while( bContLoop )                                 // as long as server alive
  {
    Log.logWithTime( 3, "  start update_all" );
    Log.setHeader( WM->getCurrentCycle(), WM->getPlayerNumber() );
    
    if( WM->updateAll( ) == true )
    {
 
      Player_interface->EvaluateRules();

      // if necessary, change the player we're listening to for coachable agent communication
      changeAttentionto();

      Log.logWithTime( 3, "  checking shallISaySomething()" );
      if( shallISaySomething() == true ){             // shall I communicate
	Log.logWithTime( 3, "  sending coachable agent say message" );
	ACT->sendCommandDirect( sayCoachableAgentMsg() );
	m_timeLastSay = WM->getCurrentTime();
      }

      logPositions(); // log player positions (if applicable)
      logWorldModelStats(); // compute world model stats (if applicable)

      Log.logWithTime( 3, "  determined action; waiting for new info" );

      if(WM->getAgentTiredNess() == TIREDNESS_VERY_BAD)
	Log.log(5, "[%d]: I'm very tired\n", WM->getPlayerNumber()); 

      // directly after see message, will not get better info, so send commands
      if( WM->getTimeLastSeeMessage() == WM->getCurrentTime() ){
	Log.logWithTime( 3, "  sending commands" );
        ACT->sendCommands( );
      }
    }
    else
      Log.logWithTime( 3, "  HOLE no action determined; waiting for new info" );

    // wait for new information from the server
    // cannot say bContLoop=WM->wait... since bContLoop can be changed elsewhere
    if(  WM->waitForNewInformation() == false )
      {
	cout << "Failed waiting for new information from the server, exiting" << endl;
        bContLoop =  false;
      }
  }

  // shutdown, print hole and number of players seen statistics

  /* dongryel: I took this out since the output is annoying and offers nothing

  printf("Shutting down player %d\n", WM->getPlayerNumber() );
  printf("   Number of holes: %d (%f)\n", WM->iNrHoles,
                         ((double)WM->iNrHoles/WM->getCurrentCycle())*100 );
  printf("   Teammates seen: %d (%f)\n", WM->iNrTeammatesSeen,
                         ((double)WM->iNrTeammatesSeen/WM->getCurrentCycle()) );
  printf("   Opponents seen: %d (%f)\n", WM->iNrOpponentsSeen,
                         ((double)WM->iNrOpponentsSeen/WM->getCurrentCycle()) );

  */
}


/*!This method logs our position and teammate positions to the log file, 
 * if a logpos file is opened. Only logs once per cycle (after see update).
 */
void Player::logPositions( )
{
  VecPosition pos;
  if( logpos != NULL &&  
      (!BasicPlayer::seeMsgArrivingThisCycle() || 
       WM->getTimeLastSeeMessage() == WM->getCurrentTime())){

    *logpos << WM->getCurrentTime() << ": "; 

    for(int i=0; i< 11; i++){
      pos = WM->getGlobalPosition(SoccerTypes::getTeammateObjectFromIndex(i));
      *logpos << pos.getX() << " " << pos.getY() << " ";

    }
    *logpos <<endl;

#if 0
#warning Remove FullState logging
    pos = WM->getGlobalPosition(OBJECT_BALL);
    *logpos << "ball: "<<pos.getX() << " " << pos.getY() << " ";
    *logpos << endl;

    *logpos << "Fullst: ";
    for(int i=0; i< 11; i++){
      pos = WM->getFullstatePosition(SoccerTypes::getTeammateObjectFromIndex(i));
      *logpos << pos.getX() << " " << pos.getY() << " ";
    }
    *logpos <<endl;
    *logpos << "Oppon:  ";
    for(int i=0; i< 11; i++){
      pos = WM->getFullstatePosition(SoccerTypes::getOpponentObjectFromIndex(i));
      *logpos << pos.getX() << " " << pos.getY() << " ";
    }
    *logpos <<endl;

    pos = WM->getFullstatePosition(OBJECT_BALL);
    *logpos << "Full ball: "<<pos.getX() << " " << pos.getY() << " ";
    *logpos << endl;

#endif

  }
}


/*!This method logs statistics about the world model
 * and prints them out at the end of the game. 
 */
void Player::logWorldModelStats()
{
  static bool logwritten = false;

  // only log if statslog is turned on, and only log once per cycle
  if( !( statslog &&  
      (!BasicPlayer::seeMsgArrivingThisCycle() || 
       WM->getTimeLastSeeMessage() == WM->getCurrentTime()))){
    return;
  }

  int curcycle = WM->getCurrentCycle();
  // don't start logging until 10th cycle because first 10 cycles 
  // will be unreliable since players may still be getting oriented
  if( curcycle < 10){
    totalcycles = 0;
    ballconfcycles = 0;
    for(int i=0; i< 11; i++){
      iTeammatecycles[i] = 0;
      iOpponentcycles[i] = 0;

      fTeammateError[i] = 0.0;
      fOpponentError[i] = 0.0;
    }

    fBallposError = 0.0;

    return;
  }
  ///////////////////////////////////////////////////

  totalcycles++;
  if(WM->isConfidenceGood(OBJECT_BALL)){
    ballconfcycles++;
	 // ballposError is the error between where we think the ball is, and where
	 // it actually is (based on the fullstate data)
	 fBallposError += fabs( WM->getGlobalPosition(OBJECT_BALL).getDistanceTo
				(WM->getFullstatePosition(OBJECT_BALL))  );
  }

  // add to count of cycles that we're confident of teammate locations.
  ObjectT player;
  for(int i=0; i<11; i++){
    if(WM->isConfidenceGood(SoccerTypes::getTeammateObjectFromIndex(i))){

      iTeammatecycles[i]++;

      player = SoccerTypes::getTeammateObjectFromIndex(i);
      fTeammateError[i] += fabs( WM->getGlobalPosition(player).getDistanceTo
			       (WM->getFullstatePosition(player)) );
    }
    if(WM->isConfidenceGood(SoccerTypes::getOpponentObjectFromIndex(i))){ 
      iOpponentcycles[i]++;

      player = SoccerTypes::getOpponentObjectFromIndex(i);
      fOpponentError[i] += fabs( WM->getGlobalPosition(player).getDistanceTo
			       (WM->getFullstatePosition(player)) );
    }
  }


  //cout << WM->getCurrentTime()<<" ball error: "<<
  //  fabs( WM->getGlobalPosition(OBJECT_BALL).getDistanceTo
  //	  (WM->getFullstatePosition(OBJECT_BALL)) ) <<endl;


  /*   */
  /* if we're at end of game, compute and print statistics */
  if( curcycle >= 2997 && !logwritten ){ //5999 ){

    *statslog << "\nPLAYER "<<WM->getPlayerNumber()<<" WORLD MODEL STATISTICS: \n"<<endl;
    *statslog << "Confident of ball location: "<<ballconfcycles<<"/"<<totalcycles
	      <<" ("<<(float)ballconfcycles/totalcycles*100. <<"%).\n"<<endl;
      
    for(int i=0; i<11; i++){
      *statslog << "Confident of teammate "<<i+1<<": "<<iTeammatecycles[i]
		<< " ("<<(float)iTeammatecycles[i]/totalcycles*100. <<"%)."<<endl;
    }

    for(int i=0; i<11; i++){
      *statslog << "Confident of opponent "<<i+1<<": "<<iOpponentcycles[i]
		<< " ("<<(float)iOpponentcycles[i]/totalcycles*100. <<"%)."<<endl;
    }
    // compute average number of teammates we were confident of each cycle
      
    int totalTeammateCycles = 0;
    for(int i=0; i<11; i++)
      totalTeammateCycles += iTeammatecycles[i];

    *statslog << "\nAverage number of teammates we were confident of per cycle: "
	      << (float)totalTeammateCycles/totalcycles <<" (max is 10)"<<endl;


    *statslog << "\nERROR values (need server:fullstate_l and fullstate_r turned on):"<<endl;
    *statslog << "Average ball position error: "<< fBallposError/ballconfcycles <<endl;

    *statslog << "Average teammate errors: "<<endl;
    for(int i=0; i<11; i++){
      if(iTeammatecycles[i]==0) *statslog << ""<<i+1<<": "<<0.0 <<"   ";
      else *statslog << ""<<i+1<<": "<<fTeammateError[i]/iTeammatecycles[i] <<"   ";
    }
    *statslog <<endl;
    *statslog << "Average opponent errors: "<<endl;
    for(int i=0; i<11; i++){
	if(iOpponentcycles[i]==0) *statslog << ""<<i+1<<": "<<0.0 <<"   ";
	else *statslog << ""<<i+1<<": "<<fOpponentError[i]/iOpponentcycles[i] <<"   ";
    }

    *statslog << endl;
    *statslog << "NumPasses: "<<WM->getPassCount()<<endl;

    *statslog <<endl<<endl;

    logwritten = true; // so we know we already wrote to log.
  }


}


/*!This method determines whether a player should say something. 
   \return bool indicating whether the agent should say a message */
bool Player::shallISaySomething( )
{
  if (!PS->getUseCommunication())
    return false;

  bool        bReturn;

  bReturn  = ((WM->getCurrentTime() - m_timeLastSay) >= SS->getHearDecay());
  bReturn  &= (WM->getCurrentCycle() > 0 );

  /* jdavin: also make sure that if we are getting a see message this cycle, 
   * we should send our say *after* the see update.(so if we're passing, we can 
   * send a PassTo message asap). 
   * so we return true if it is not time to turn (meaning we're not getting a see update)
   *  *or* we have already received a see message this cycle. 
   */
  bReturn &= !BasicPlayer::seeMsgArrivingThisCycle() || 
    WM->getTimeLastSeeMessage() == WM->getCurrentTime();

  return bReturn;
}

/*! This method returns a boolean indicating whether I should communicate my
    world model to the other agents. 
    \return boolean indicating whether I should communicate my world model. */
bool Player::amIAgentToSaySomething()
{
  // if got good information of the ball, communicate. 
	 //  if( WM->getTimeChangeInformation(OBJECT_BALL) == WM->getCurrentTime() ||
	 //   WM->getRelativeDistance( OBJECT_BALL ) < SS->getVisibleDistance()    ) 
  return true;
}

/*! This method changes our attentionto (which player we're listening to)
    if necessary. (it is only necessary to change focus when the AgentComm class
    wants us to). 
*/
void Player::changeAttentionto()
{
  if (!PS->getUseCommunication())
    return;

  if(WM->getPlayerNumber() == 0)
    return; // playernum not set yet, so don't do anything

  unsigned short focuson = AgentComm->getPlayerToFocusOn();
  if(focuson == 100)
    return; // don't need to change focus. 
  else if(focuson == 0)
    ACT->sendCommandDirect( setAttentionto(1, -1) );
    // -1 as the playernum indicates to issue (attentionto off).

  else // setAttentionto(1, focuson). (1 indicates our team. we never listen to opponents)
    ACT->sendCommandDirect( setAttentionto(1, focuson) );

}

/*! This method is calls the CoachableAgentComm class 
    to get a message to send. */
SoccerCommand Player::sayCoachableAgentMsg( ){

  if(m_timeLastSay > 0 && m_timeLastSay < WM->getCurrentTime()-1){
    cerr << "[WARNING]: player "<<WM->getPlayerNumber()<<" missed issuing a say (lastsay="
	 << m_timeLastSay <<", curtime="<< WM->getCurrentTime()<<")\n"<<endl;
  }

  string msg = AgentComm->getSayMsg();

  if( msg == "" ){ // AgentComm couldn't give us a message
    return SoccerCommand(CMD_ILLEGAL);
  }

  char saymsg[msg.length()];
  sprintf(saymsg, msg.c_str());
  
  return communicate( saymsg );
}

/*! This method is used to communicate the status of the ball, that is its
    position and velocity.*/
SoccerCommand Player::sayBallStatus( )
{
  char    strMsg[MAX_SAY_MSG];

  // First add al values to a positive interval, since then we don't need
  // another byte for the minus sign. then take one digit at a time
  VecPosition posBall = WM->getGlobalPosition( OBJECT_BALL );
  VecPosition velBall = WM->getGlobalVelocity( OBJECT_BALL );
  double x = min( posBall.getX() + 52.5, 99.9);
  sprintf( strMsg, "%c%c%c%c%c%c%c%c",
     '0' + ((int)( x                    ) % 100 ) / 10 ,
     '0' + ((int)( x                    ) % 100 ) % 10 ,      
     '0' + ((int)( posBall.getY() + 34.0) % 100 ) / 10 ,
     '0' + ((int)( posBall.getY() + 34.0) % 100 ) % 10 ,
     '0' + ((int)(( velBall.getX() + 2.7) * 10 )) / 10 ,
     '0' + ((int)(( velBall.getX() + 2.7) * 10 )) % 10 ,     
     '0' + ((int)(( velBall.getY() + 2.7) * 10 )) / 10 ,
     '0' + ((int)(( velBall.getY() + 2.7) * 10 )) % 10 );                    

  return communicate( strMsg );
}


/*!This method does nothing except updating the worldmodel after each arrival of
   a see or sense_body message. Using this test program it is possible to
   control the commands of the agent using the standard input and see the
   resulting behavior of the agent. */
void Player::test_only_update()
{
  while( bContLoop )
  {
    WM->updateAll();
    if(  WM->waitForNewInformation() == false )
    {
        printf("Server dead; exiting" );
        bContLoop =  false;
    }
  }
}

/*!This method can be called in a separate thread (see pthread_create) since
   it returns a void pointer. It is assumed that this function receives a
   BasicPlayer class as argument. The only thing this function does
   is starting the method handleStdin() from the corresponding BasicPlayer
   class that listens to user input from the keyboard. This function is
   necessary since a method from a class cannot be an argument of
   pthread_create.
   \param v void pointer to a BasicPlayer class.
   \return should never return since function handleStdin has infinite loop*/
void* stdin_callback( void * v )
{
  Log.log( 1, "Starting to listen for user input" );
  Player* p = (Player*)v;
  p->handleStdin();
  return NULL;
}

/*!This method listens for input from the keyboard and when it receives this
   input it converts this input to the associated action. See
   showStringCommands for the possible options. This method is used together
   with the SenseHandler class that sends an alarm to indicate that a new
   command can be sent. This conflicts with the method in this method that
   listens for the user input (fgets) on Linux systems (on Solaris this isn't a
   problem). The only known method is to use the flag SA_RESTART with this
   alarm function, but that does not seem to work under Linux. If each time
   the alarm is sent, this gets function unblocks, it will cause major
   performance problems. This function should not be called when a whole match
   is played! */
void Player::handleStdin( )
{
  char buf[MAX_MSG];

  while( bContLoop )
  {
    fgets( buf, MAX_MSG, stdin ); // does unblock with signal !!!!!
    printf( "after fgets: %s\n", buf );
    executeStringCommand( buf );
  }
}

/*!This method prints the possible commands that can be entered by the user.
   The whole name can be entered to perform the corresponding command, but
   normally only the first character is sufficient. This is indicated by
   putting brackets around the part of the command that is not needed.
   \param out output stream to which the possible commands are printed */
void Player::showStringCommands( ostream& out )
{
  out << "Basic commands:"                                << endl <<
         " a(ctions)"                                     << endl <<
         " c(atch) direction"                             << endl <<
         " cs(lientsettings"                              << endl <<
         " d(ash) power [ times ]"                        << endl <<
         " de(bug) nr_cycles"                             << endl <<
         " g(oto) x y"                                    << endl <<
         " h(elp)"                                        << endl <<
         " i(ntercept) x y"                               << endl <<
         " k(ick) power angle"                            << endl <<
         " ka x y endspeed "                              << endl <<
         " m(ove) x y"                                    << endl <<
         " n(eck) angle"                                  << endl <<
         " o(pponents in cone) width dist"                << endl <<
         " p(redict cycles to) x y"                       << endl <<
         " q(uit)"                                        << endl <<
         " s(ay) message"                                 << endl <<
         " ss(erversettings)"                             << endl <<
         " t(urn) angle"                                  << endl <<
         " v(iewmode) narrow | normal | wide low | high"  << endl <<
         " w(orldmodel)"                                  << endl;
}

/*!This method executes the command that is entered by the user. For the
   possible command look at the method showStringCommands.
   \param str string that is entered by the user
   \return true when command could be executed, false otherwise */
bool Player::executeStringCommand( char *str)
{
  SoccerCommand socCommand;
  int           i;
  double        x, y;

  switch( str[0] )
  {
    case 'a':                                 // actions
      WM->showQueuedCommands();
      break;
    case 'c':                                 // catch dir or cs
      if( strlen(str) > 1 && str[1] == 's' )
      {
        PS->show( cout, ":" );
        break;
      }
      socCommand.makeCommand( CMD_CATCH, Parse::parseFirstInt( &str ) );
      break;
    case 'd':                                 // dash
      socCommand.commandType = CMD_DASH;
      socCommand.dPower      = Parse::parseFirstDouble( &str );
      socCommand.iTimes      = Parse::parseFirstInt   ( &str );
      if( socCommand.iTimes == 0 ) socCommand.iTimes = 1;
      break;
    case 'h':                                // help
      showStringCommands( cout );
      return true;
    case 'k':                                // kick or ka (kick advanced)
      socCommand.commandType = CMD_KICK;
      if( str[1] == 'a' ) // advanced kick
      {
        double x = Parse::parseFirstDouble( &str );
        double y = Parse::parseFirstDouble( &str );
        double e = Parse::parseFirstDouble( &str );
        socCommand = kickTo( VecPosition( x, y), e );
      }
      else
      {
        socCommand.dPower = Parse::parseFirstDouble( &str );
        socCommand.dAngle = Parse::parseFirstDouble( &str );
      }
      break;
    case 'm':                               // move
      socCommand.commandType = CMD_MOVE;
      socCommand.dX          = Parse::parseFirstDouble( &str );
      socCommand.dY          = Parse::parseFirstDouble( &str );
      socCommand.dAngle      = Parse::parseFirstDouble( &str );
      break;
    case 'n':                              // turn_neck
      socCommand.commandType = CMD_TURNNECK;
      socCommand.dAngle      = Parse::parseFirstDouble( &str );
      break;
    case 'o':                              // count nr opp in cone
      x = Parse::parseFirstDouble( &str );
      y = Parse::parseFirstDouble( &str );
      i = WM->getNrInSetInCone( OBJECT_SET_OPPONENTS, x, WM->getAgentGlobalPosition(),
              WM->getAgentGlobalPosition()+VecPosition( y,
          WM->getAgentGlobalNeckAngle(), POLAR ) );
      printf( "%d opponents\n", i );
      return true;
    case 'p':                              // predict cycles to point
      x = Parse::parseFirstDouble( &str );
      y = Parse::parseFirstDouble( &str );
      i = WM->predictNrCyclesToPoint( WM->getAgentObjectType(),
                     VecPosition( x, y ), PS->getPlayerWhenToTurnAngle() );
      printf( "%d cycles\n", i );
      return true;
    case 'q':                             // quit
      bContLoop = false;
      return true;
    case 's':                             // ss (serversettings) or say
      if( strlen(str) > 1 && str[1] == 's' )
      {
        SS->show( cout, ":" );
        break;
      }
      socCommand.commandType = CMD_SAY;
      Parse::gotoFirstOccurenceOf( ' ', &str );
      Parse::gotoFirstNonSpace( &str );
      strcpy( socCommand.str, str);
      break;
    case 't':                             // turn
      socCommand.commandType = CMD_TURN;
      socCommand.dAngle      = Parse::parseFirstDouble( &str );
      break;
    case 'v':                             // change_view
      socCommand.commandType = CMD_CHANGEVIEW;
      Parse::gotoFirstOccurenceOf(' ', &str );
      Parse::gotoFirstNonSpace( &str );
      socCommand.va          = SoccerTypes::getViewAngleFromStr( str );
      Parse::gotoFirstOccurenceOf(' ', &str );
      Parse::gotoFirstNonSpace( &str );
      socCommand.vq          = SoccerTypes::getViewQualityFromStr( str );
      break;
    case 'w':                            // worldmodel
      WM->show();
      return true;
    default:                             // default: send entered string
      ACT->sendMessage( str );
      return true;
  }
  if( socCommand.commandType != CMD_ILLEGAL ) // when socCommand is set
    ACT->putCommandInQueue( socCommand );     // send it.

  return true;
}




/**************************************************************************************/
/*** THESE FUNCTIONS ARE DEFUNCT. Wyverns use rule evaluation to determine actions. ***/

#if 0
/*! This is the main decision loop for the goalkeeper. */
void Player::goalieMainLoop( )
{
  deMeer5_goalie();
}    

/*! This is the main decision loop for a defender. */
void Player::defenderMainLoop( )
{
  deMeer5() ;
}

/*! This is the main decision loop for a midfielder. */
void Player::midfielderMainLoop( )
{
  deMeer5() ;
}

/*! This is the main decision loop for an agent. */
void Player::attackerMainLoop( )
{
  deMeer5();
}
#endif

/*!This method is the first complete simple team and defines the actions taken
   by all the players on the field (excluding the goalie). It is based on the
   high-level actions taken by the simple team FC Portugal that it released in
   2000. The players do the following:
   - if ball is kickable
       kick ball to goal (random corner of goal)
   - else if i am fastest player to ball and no opponent can intercept ball
       intercept the ball
   - else
       move to strategic position based on your home position and pos ball */

/* 
void Player::deMeer5(  )
{

  SoccerCommand soc(CMD_ILLEGAL);
  VecPosition   posAgent = WM->getAgentGlobalPosition();
  VecPosition   posBall  = WM->getBallPos();
  int           iTmp;
  
  if( WM->isKickOffThem(  ) )
    ; // do nothing
  else if( WM->isBeforeKickOff( ) )
  {
    if( formations->getFormation() != FT_INITIAL )
      formations->setFormation( FT_INITIAL );       // go to kick_off formation
    VecPosition posStrat = WM->getStrategicPosition();
    if( WM->isKickOffThem( ) && posStrat.getDistanceTo( VecPosition(0,0) ) < 9.0 )
      posStrat.setX( -10.0 );

    if( posAgent.getDistanceTo( posStrat ) > 2.0  )
      ACT->putCommandInQueue( teleportToPos( posStrat ) );
     else                                            // else turn to center
    {
      ACT->putCommandInQueue( turnBodyToPoint( VecPosition( 0, 0 ), 0 ) );
      ACT->putCommandInQueue( alignNeckWithBody( ) );
    }
  }
  else
  {
    formations->setFormation( FT_433_OFFENSIVE );
    soc.commandType = CMD_ILLEGAL;

    if( WM->getConfidence( OBJECT_BALL ) < PS->getBallConfThr() )
    {
      ACT->putCommandInQueue( searchBall() );         // if ball pos unknown
      ACT->putCommandInQueue( alignNeckWithBody( ) ); // search for it
    }
    else if( WM->isBallKickable())                    // if kickable
    {
      VecPosition posGoal( PITCH_LENGTH/2.0,
              (-1 + 2*(WM->getCurrentCycle()%2)) * 0.4 * SS->getGoalWidth() );
      soc = kickTo( posGoal, SS->getBallSpeedMax() ); // kick maximal

      ACT->putCommandInQueue( soc );
      ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      Log.log( 100, "kick ball" );      
    }
    else if( WM->getFastestInSetTo( OBJECT_SET_TEAMMATES, OBJECT_BALL, &iTmp )
              == WM->getAgentObjectType()  && !WM->isDeadBallThem() )
    {                                                // if fastest to ball
      Log.log( 100, "I am fastest to ball; can get there in %d cycles", iTmp );
      soc = intercept( false );                      // intercept the ball

      if( soc.commandType == CMD_DASH &&             // if stamina low
          WM->getAgentStamina().getStamina() <
             SS->getRecoverDecThr()*SS->getStaminaMax()+200 )
      {
        soc.dPower = 30.0 * WM->getAgentStamina().getRecovery(); // dash slow
        ACT->putCommandInQueue( soc );
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      }
      else                                           // if stamina high
      {
        ACT->putCommandInQueue( soc );               // dash as intended
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      }
     }
     else if( posAgent.getDistanceTo(WM->getStrategicPosition()) >
                  1.5 + fabs(posAgent.getX()-posBall.getX())/10.0)
                                                  // if not near strategic pos
     {
       if( WM->getAgentStamina().getStamina() >     // if stamina high
                            SS->getRecoverDecThr()*SS->getStaminaMax()+200 )
       {
         soc = moveToPos(WM->getStrategicPosition(),PS->getPlayerWhenToTurnAngle());
         ACT->putCommandInQueue( soc );            // move to strategic pos
         ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
       }
       else                                        // else watch ball
       {
         ACT->putCommandInQueue( soc = turnBodyToObject( OBJECT_BALL ) );
         ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
       }
     }
     else if( fabs( WM->getRelativeAngle( OBJECT_BALL ) ) > 1.0 ) // watch ball
     {
       ACT->putCommandInQueue( soc = turnBodyToObject( OBJECT_BALL ) );
       ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
     }
     else                                         // nothing to do
       ACT->putCommandInQueue( SoccerCommand(CMD_TURNNECK,0.0) );
   }
}
*/



/*!This method is a simple goalie based on the goalie of the simple Team of
   FC Portugal. It defines a rectangle in its penalty area and moves to the
   position on this rectangle where the ball intersects if you make a line
   between the ball position and the center of the goal. If the ball can
   be intercepted in the own penalty area the ball is intercepted and catched.
*/
/*
void Player::deMeer5_goalie(  )
{
  int i;
  SoccerCommand soc;
  VecPosition   posAgent = WM->getAgentGlobalPosition();
  AngDeg        angBody  = WM->getAgentGlobalBodyAngle();

  // define the top and bottom position of a rectangle in which keeper moves
  static const VecPosition posLeftTop( -PITCH_LENGTH/2.0 +
               0.7*PENALTY_AREA_LENGTH, -PENALTY_AREA_WIDTH/4.0 );
  static const VecPosition posRightTop( -PITCH_LENGTH/2.0 +
               0.7*PENALTY_AREA_LENGTH, +PENALTY_AREA_WIDTH/4.0 );

  // define the borders of this rectangle using the two points.
  static Line  lineFront = Line::makeLineFromTwoPoints(posLeftTop, posRightTop);
  static Line  lineLeft  = Line::makeLineFromTwoPoints(
                         VecPosition( -50.0, posLeftTop.getY()), posLeftTop );
  static Line  lineRight = Line::makeLineFromTwoPoints(
                         VecPosition( -50.0, posRightTop.getY()),posRightTop );


  if( WM->isBeforeKickOff( ) )
  {
    if( formations->getFormation() != FT_INITIAL || // not in kick_off formation
        posAgent.getDistanceTo( WM->getStrategicPosition() ) > 2.0 )  
    {
      formations->setFormation( FT_INITIAL );       // go to kick_off formation
      ACT->putCommandInQueue( teleportToPos( WM->getStrategicPosition() ) );
    }
    else                                            // else turn to center
    {
      ACT->putCommandInQueue( turnBodyToPoint( VecPosition( 0, 0 ), 0 ) );
      ACT->putCommandInQueue( alignNeckWithBody( ) );
    }
    return;
  }

  if( WM->getConfidence( OBJECT_BALL ) < PS->getBallConfThr() )
  {                                                // confidence ball too  low
    ACT->putCommandInQueue( searchBall() );        // search ball
    ACT->putCommandInQueue( alignNeckWithBody( ) );
  }
  else if( WM->getPlayMode() == PM_PLAY_ON || WM->isFreeKickThem() ||
           WM->isCornerKickThem() )               
  {
    if( WM->isBallCatchable() )
    {
      ACT->putCommandInQueue( soc = catchBall() );
      ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
    }
     else if( WM->isBallKickable() )
    {
       soc = kickTo( VecPosition(0,posAgent.getY()*2.0), 2.0 );    
       ACT->putCommandInQueue( soc );
       ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
    }
    else if( WM->isInOwnPenaltyArea( getInterceptionPointBall( &i, true ) ) &&
             WM->getFastestInSetTo( OBJECT_SET_PLAYERS, OBJECT_BALL, &i ) == 
                                               WM->getAgentObjectType() )
    {
      ACT->putCommandInQueue( soc = intercept( true ) );
      ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
    }
    else
    {
      // make line between own goal and the ball
      VecPosition posMyGoal = ( WM->getSide() == SIDE_LEFT )
             ? SoccerTypes::getGlobalPositionFlag(OBJECT_GOAL_L, SIDE_LEFT )
             : SoccerTypes::getGlobalPositionFlag(OBJECT_GOAL_R, SIDE_RIGHT);
      Line lineBall = Line::makeLineFromTwoPoints( WM->getBallPos(), posMyGoal);

      // determine where your front line intersects with the line from ball
      VecPosition posIntersect = lineFront.getIntersection( lineBall );

      // outside rectangle, use line at side to get intersection
      if (posIntersect.isRightOf( posRightTop ) )
        posIntersect = lineRight.getIntersection( lineBall );
      else if (posIntersect.isLeftOf( posLeftTop )  )
        posIntersect = lineLeft.getIntersection( lineBall );

      if( posIntersect.getX() < -49.0 )
        posIntersect.setX( -49.0 );
        
      // and move to this position
      if( posIntersect.getDistanceTo( WM->getAgentGlobalPosition() ) > 0.5 )
      {
        soc = moveToPos( posIntersect, PS->getPlayerWhenToTurnAngle() );
        ACT->putCommandInQueue( soc );
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      }
      else
      {
        ACT->putCommandInQueue( soc = turnBodyToObject( OBJECT_BALL ) );
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      }
    }
  }
  else if( WM->isFreeKickUs() == true || WM->isGoalKickUs() == true )
  {
    if( WM->isBallKickable() )
    {
      if( WM->getTimeSinceLastCatch() == 25 && WM->isFreeKickUs() )
      {
        // move to position with lesser opponents.
        if( WM->getNrInSetInCircle( OBJECT_SET_OPPONENTS, 
                                          Circle(posRightTop, 15.0 )) <
            WM->getNrInSetInCircle( OBJECT_SET_OPPONENTS, 
                                           Circle(posLeftTop,  15.0 )) )
          soc.makeCommand( CMD_MOVE, posRightTop.getX(),posRightTop.getY(),0.0);
        else
          soc.makeCommand( CMD_MOVE, posLeftTop.getX(), posLeftTop.getY(), 0.0);
        ACT->putCommandInQueue( soc );
      }
      else if( WM->getTimeSinceLastCatch() > 28 )
      {
        soc = kickTo( VecPosition(0,posAgent.getY()*2.0), 2.0 );    
        ACT->putCommandInQueue( soc );
      }
      else if( WM->getTimeSinceLastCatch() < 25 )
      {
        VecPosition posSide( 0.0, posAgent.getY() ); 
        if( fabs( (posSide - posAgent).getDirection() - angBody) > 10 )
        {
          soc = turnBodyToPoint( posSide );
          ACT->putCommandInQueue( soc );
        }
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      }
    }
    else if( WM->isGoalKickUs()  )
    {
      ACT->putCommandInQueue( soc = intercept( true ) );
      ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
    }
    else
      ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
  }
  else
  {
     ACT->putCommandInQueue( soc = turnBodyToObject( OBJECT_BALL ) );
     ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
  }

}
*/






