/*
 * A PlayerInterface is instantiated from main() and is used to interface between 
 * the player and the coach commands. At each round, the player interface will 
 * evaluate rules and select commands for the agent to carry out. 
 * John Davin
 */


#include "PlayerInterface.h"

#include <stdio.h>
#include <fstream.h>
#include <iostream.h>
#include <unistd.h>

#include "ActionBinding.h"
#include "SoccerTypes.h"
#include "clangparser.h"
#include "clangmsgbuilder.h"
#include "clangmsg.h"
#include "config.h"
#include "mymsgsorter.h"
#include "msgstore.h"
#include "msgrulestore.h"
#include "CondVisitor.h"
#include "Filter.h"
#include "LevelStorage.h"



WorldModel *trilearn_wm; // our pointer to the WorldModel



// max length of coach messages? 
//#define MSG_LEN 10000 // this is what it's defined as in the clangparser lib

extern MsgRuleStore defrulestorage; // declared in msgsorterdefine.h

extern MsgStore< Cond* > defcondstorage;
extern MsgStore< Dir* > defdirstorage; 
extern MsgStore< Region* > defregstorage; 
extern MsgStore< Action* > defactstorage; 


ostream *logger; // a pointer to the Logger output stream


/*
 * wm is a pointer to the trilearn WorldModel
 * cmdline is true if we should read coach commands from stdin.
 */
PlayerInterface::PlayerInterface(WorldModel *wm, bool cmdline, char *coachfile, int rulelogging)
{
  int rtn;
  char line[MAX_MSG];
  //char tmp;

  using namespace rcss;
  using namespace clang;
  setbuf(stdout, 0);

  trilearn_wm = wm; // keep a pointer to the WorldModel

  logger = &Log.getOutputStream();
  // turn on the error log number (9 currently) so user sees errors/warnings
  Log.addLogLevel(ERROR_LOGLEVEL);

  MsgBuilder bld;
  rcss::clang::Parser p(bld);
  Msg *mymsg = NULL;

  // MyMsgSorter is a Visitor that allows us to determine message types. 
  MyMsgSorter mysorter(ActionBinding::RULEFILE);

  defrulestorage.setRuleLogging(rulelogging);

  istream *infile; // open stdin
  bool delete_infile;
  if(cmdline){ // if we're not using the cmdline, then
    infile = &cin;
    delete_infile = false;
    printf("[PlayerInterface]: Taking input from stdin.\n");
  }
  else{
    delete_infile = true;
    if(coachfile != NULL)
      infile = new ifstream(coachfile);
    else // else use the default
      infile = new ifstream(RULES_FILE);
  }


  if(!infile->good()){
    cerr<<"Could not open rules file!"<<endl;
    Log.log(5, "Could not open rules file!\n"); exit(1);
  }

  int linecount = 0; // current line number
  while(!infile->eof()){
    bzero(line, MAX_MSG);
    infile->getline(line, MAX_MSG);
    linecount++;

    if(line[0] == 0 || line[0] == '\n') continue; // don't add blank lines
    if(line[0] == '#') continue; // also ignore lines starting with # (our comment character)
    if(line[MAX_MSG-1] != 0 && line[MAX_MSG-1] != '\n'){
      Log.log(5, "Error: Message too long: <%c>\n", line[MAX_MSG-1]); return;}
    
    // get pointer to current message 
    try{
      rtn = p.parse(line);
      if(rtn == 1){ cerr << "ERROR parsing line #" << linecount << ": "
			 << line <<"\nPress enter to continue with next line.\n"; 
      getchar();  // pause
      continue;}

      mymsg = bld.getMsg();
    } catch( BuilderErr e){ cerr << e.what() << endl;}
    catch( rcss::util::NullErr e){ cerr << e.what() << endl;}

    if(mymsg == NULL){ PR(5, "Failed to get message!\n"); continue;}

    //cout <<"visiting message: "<<*mymsg<<endl<<endl;
    // handle the specific msg type:
    mymsg->accept(mysorter);

  }

  // this will also close the file
  if (delete_infile)
    delete infile;

  //defrulestorage.PrintAll();

  PR(5, "Successfully parsed rules file.");
  PR(5, "\n============================\n");
}


void PlayerInterface::EvaluateRules(){

  defrulestorage.EvaluateRules();

}

/*
 * Used to add a say message from a coach to our rules storage. 
 * Parses the message and adds it to our storage using a visitor. 
 * Returns false on error, true on success
 * The buffer that is passed in *must* have MAX_MSG bytes of memory allocated
 */
bool PlayerInterface::handleMessage(char *msg){
  int rtn;
  MsgBuilder bld;
  rcss::clang::Parser p(bld);
  Msg *mymsg = NULL;

  //PR(5, "[PlayerInterface:handleMessage]: Received coach msg %s\n", msg);


  // MyMsgSorter is a Visitor that allows us to determine message types. 
  MyMsgSorter mysorter(ActionBinding::COACH);

  //if(msg[MAX_MSG-1] != 0 && msg[MAX_MSG-1] != '\n'){
  //  Log.log(5, "Error: Message too long: <%c>\n", msg[MAX_MSG-1]); return false;}
    
    // get pointer to current message 
    try{
      rtn = p.parse(msg);
      if(rtn == 1){ cerr << "ERROR parsing msg: "<< msg << endl;
      return false;}

      mymsg = bld.getMsg();
    } catch( BuilderErr e){ cerr << e.what() << endl;}
    catch( rcss::util::NullErr e){ cerr << e.what() << endl;}

    if(mymsg == NULL){ PR(5, "Failed to get message!\n"); return false;}
    //mymsg->print(cout);
    
    // handle the specific msg type:
    mymsg->accept(mysorter);
    
    
    //PR(5, "[PlayerInterface:handleMessage]: Successfully handled coach msg %s\n", msg);


    return true;
}

