/* Dongryeol Lee
 *
 * PointVisitor.C
 * Visitor Class for Point Objects
 *
 * Implementation of visitor class for Point Objects
 * Basically reduces every Point Object to PointSimple,
 * the global position on the field
 */

#include "visitor.h"
#include "clangutil.h"
#include "vector.h"
#include "SoccerTypes.h"
#include "Geometry.h"
#include "PointVisitor.h"
#include "WorldModelAccessor.h"
#include <stdio.h>
#include <typeinfo>

using namespace rcss::clang;
using namespace rcss::geom;
using namespace rcss::util;
using namespace std;

/* Access to the world model */
extern WorldModelAccessor MyWorldModel;

/* Constructor function */
void PointVisitor::init()
{
  globalX = 0.0;
  globalY = 0.0;
}

/* Return the global position of any point object referred
   by its PointVisitor */
VecPosition PointVisitor::returnGlobalPosition()
{ 
  return VecPosition(globalX, globalY);
}

/* Visit a simple point that is representative of the global
   position */
void PointVisitor::visit(PointSimple *p)
{
  /* Get the 2D vector referred by the simple point */
  Vector2D pos = p->getVec();
  
  //Log.log(8, "[PointSimple]: got point %f %f\n", pos.getX(), pos.getY());

  /* Simply add the x and y-coordinates the global position */
  globalX += pos.getX();
  globalY += pos.getY();
}

/* Visit a point relative to some other point */
void PointVisitor::visit(PointRel *p)
{
  /* Retrieve the "origin" of the point p */
  Point *origin = (Point *) p->getOrigin();
  
  /* Retrieve the offset from the "origin" */
  PointSimple offset = p->getOffset();
  
  /* Again, simply add the x and y-coordinates of the offset to
     the running tally */
  globalX += offset.getVec().getX();
  globalY += offset.getVec().getY();
  
  /* Recursively search for the final global position of
     the origin by visiting it */
  origin->accept(*this);
}

/* Visit a point defined by the position of the ball on the field */
void PointVisitor::visit(PointBall *)
{
  /* Retrieve the global position of the ball from the world model */
  VecPosition ballPos = MyWorldModel.getBallPos();

  /* Add to the global X and Y position of the point visitor */
  globalX += ballPos.getX();
  globalY += ballPos.getY();
}

/* Visit a point defined by a player on the field */
void PointVisitor::visit(PointPlayer *p)
{
  /* First, retrieve the UNum of the player and determine whether
     it is on our side or not */
  UNum::unum_t playerUNum = p->getUNum().getUNum();
  bool isPlayerOurSide = p->isOurSide();
  VecPosition playerPos;

  string varname = p->getUNum().getVar();

  if(varname != ""){ // then we have a clang variable

    UNumSet varset = myBindings.lookup(varname);
    if(varset.begin() == varset.end()){ // empty set
      // if the variable is not bound, then 
      // add an empty binding for this variable to myBindings. 
      myBindings.add(varname, UNumSet());


      /*
      Log.log(11, "[PointPlayer]: Unbound unum %s added to myBindings.\n", 
      varname.c_str());
      */


      return;
    }
    else{ // nonempty set
      // then the variable is bound, so evaluate this point with its binding

      if(varset.size() != 1){
	
	/*
	Log.log(9, "[PointPlayer]: WARNING: player point has UNum %s bound to more than 1 player\n", varname.c_str());
	*/


      }

      playerUNum = (*varset.begin()).getUNum();
      
      /*
      Log.log(11, "[PointPlayer]: Got bound variable %s. Evaluating point with unum %d\n",
      varname.c_str(), playerUNum);
      */

    }

  }

  /* Determine the position of the field given by the world model */
  playerPos = MyWorldModel.getGlobalPosition(playerUNum, isPlayerOurSide);
  
  /*
  Log.log(11, "[PointPlayer]: got player pos %f %f\n", playerPos.getX(), playerPos.getY());
  */

  /* Add to the global X and Y position */
  globalX += playerPos.getX();
  globalY += playerPos.getY();			       
}

/* Visit the point defined by PointArith. Basically retrieve the left
   and the right Point of PointArith and recursively get the positions of
   left and right and apply operation on them at the end */
void PointVisitor::visit(PointArith *p)
{
  /* Get the left and right points of the PointArith on its arithmetic
     tree. Also retrieve the operation to be performed on two points */
  PointVisitor *pvl = new PointVisitor(myBindings);
  PointVisitor *pvr = new PointVisitor(myBindings);
  //PointVisitor pvl, pvr;
  Point *left = p->getPoint(0);
  Point *right = p->getPoint(1);
  const ArithOp *operation = p->getOp();
  VecPosition leftPos, rightPos;

  /* Recursively visit the left and right points */
  left->accept(*pvl);
  right->accept(*pvr);

  myBindings.mergeWithIntersect(pvl->getBindings());
  myBindings.mergeWithIntersect(pvr->getBindings());

  /* Retrieve the result of the visiting for left and right */
  leftPos = pvl->returnGlobalPosition();
  rightPos = pvr->returnGlobalPosition();

  //Log.log(9, "[PointArith]: got left: %g %g  and right: %f %f\n", leftPos.getX(), leftPos.getY(), (float)rightPos.getX(), (float)rightPos.getY());
  //  cout << leftPos.getX() << " " << leftPos.getY() <<" "<< rightPos.getX()<<" "<< rightPos.getY()<<endl;

  /* perform the operation on two points */
  switch(*(operation->getStr()))
    {
    case '+':
      globalX += leftPos.getX() + rightPos.getX();
      globalY += leftPos.getY() + rightPos.getY();
      break;
    case '-':
      globalX += (leftPos.getX() - rightPos.getX());
      globalY += (leftPos.getY() - rightPos.getY());
      break;   
    case '*':
      globalX += leftPos.getX() * rightPos.getX();
      globalY += leftPos.getY() * rightPos.getY();
      break;   
    case '/':
      globalX += leftPos.getX() / rightPos.getX();
      globalY += leftPos.getY() / rightPos.getY();
      break;   
    }

  delete pvl;
  delete pvr;
}
