/* Dongryeol Lee
 *
 * ReductionActionVisitor.C
 *
 * This file implements the Reduction class, as outlined in the header file,
 * ReductionActionVisitor.h
 *
 */


#include <string>
#include "msgstore.h"
#include "ReductionActionVisitor.h"
#include "WorldModelAccessor.h"


/* Our access to the Player's World Model */
extern WorldModelAccessor MyWorldModel;


using namespace rcss::clang;
using namespace std;


/* ReductionActionVisitor constructor */
ReductionActionVisitor::ReductionActionVisitor(const ActionBinding &abinding)
{
  /* Get source of the action advice */
  sourceType = abinding.sourceType;
  
  /* Get variable bindings for the action */
  bindings = abinding.myBindings;
}

/* Reduction function for ActPos actions */
void ReductionActionVisitor::visit(ActPos *act)
{
  /* This isn't implemented yet */
  atomizedActions.push_back(ActionBinding(act, bindings, sourceType));
}

/* Reduction function for ActHome actions */
void ReductionActionVisitor::visit(ActHome *act)
{
  /* This isn't implemented yet as well. */
  atomizedActions.push_back(ActionBinding(act, bindings, sourceType));
}

/* This action is obsolete, so ignore */
void ReductionActionVisitor::visit(ActBallToReg *)
{

}

/* Same with this action */
void ReductionActionVisitor::visit(ActBallToPlayer *)
{

}

/* Reduction function for ActMark actions */
void ReductionActionVisitor::visit(ActMark *act)
{
  /* Boolean flags for which players are in the UNumSet */
  bool playerList[12] = {false, false, false, false, false, false,
			 false, false, false, false, false, false};


  /* Apply the 2nd reduction: Multi-UNumSet -> SingleUNumSet */
  UNumSet multiu = act->getPlayers();


  /* Size of the UNum Set */
  int size = multiu.size();


  /* For each UNum in the UNumSet, create ActMark acts, and
     push them into the atomized list */
  for(int i = 0; i < size; i++) {
    

    /* Retrieve the current UNum */
    UNum curUNum = multiu.getIndexAt(i);


    /* Variable name if it exists */
    std::string varName = curUNum.getVar();


    /* If the current UNum is a variable, */
    if(varName != "")    {

      /* Look up the UNumSet for the bindings */
      UNumSet varu = bindings.lookup(varName);

      /* Retrieve its size */
      int varSize = varu.size();


      /* Go through its list and atomize */
      for(int j = 0; j < varSize; j++) {
	
	/* Turn on the boolean flag */
	playerList[(int) (varu.getIndexAt(j).getUNum())] = true;

	/* If we have encountered 0 somewhere, terminate */
	if(playerList[0])
	  break;
      }

    }

    /* If it's just a UNum, */
    else
      playerList[(int) curUNum.getUNum()] = true;


    /* If we have encountered the UNum = 0 somewhere, terminate */
    if(playerList[0])
      break;

  }

  /* Add our atomized actions */
  for(int i = 1; i <= 11; i++)     {
      
    /* Add it to our atomized list */
    if(playerList[0] || playerList[i]) {
      
      UNumSet singleu;

      singleu.add((UNum::unum_t) i);

      atomizedActions.push_back(ActionBinding(new ActMark(singleu),
					      bindings,
					      sourceType) );
    }

  }
}

/* Reduction function for ActMarkLinePlayer actions */
void ReductionActionVisitor::visit(ActMarkLinePlayer *act)
{
  /* Boolean flags for which players are in the UNumSet */
  bool playerList[12] = {false, false, false, false, false, false,
			 false, false, false, false, false, false};


  /* Apply the 2nd reduction: Multi-UNumSet -> SingleUNumSet */
  UNumSet multiu = act->getPlayers();


  /* Size of the UNum Set */
  int size = multiu.size();


  /* For each UNum in the UNumSet, create ActMark acts, and
     push them into the atomized list */
  for(int i = 0; i < size; i++) {
    

    /* Retrieve the current UNum */
    UNum curUNum = multiu.getIndexAt(i);


    /* Variable name if it exists */
    std::string varName = curUNum.getVar();


    /* If the current UNum is a variable, */
    if(varName != "")    {

      /* Look up the UNumSet for the bindings */
      UNumSet varu = bindings.lookup(varName);

      /* Retrieve its size */
      int varSize = varu.size();


      /* Go through its list and atomize */
      for(int j = 0; j < varSize; j++) {
	
	/* Turn on the boolean flag */
	playerList[(int) (varu.getIndexAt(j).getUNum())] = true;

	/* If we have encountered 0 somewhere, terminate */
	if(playerList[0])
	  break;
      }

    }

    /* If it's just a UNum, */
    else
      playerList[(int) curUNum.getUNum()] = true;


    /* If we have encountered the UNum = 0 somewhere, terminate */
    if(playerList[0])
      break;

  }

  /* Add our atomized actions */
  for(int i = 1; i <= 11; i++)     {
      
    /* Add it to our atomized list */
    if(playerList[0] || playerList[i]) {
      
      UNumSet singleu;

      singleu.add((UNum::unum_t) i);

      atomizedActions.push_back(ActionBinding(new ActMarkLinePlayer(singleu),
					      bindings,
					      sourceType) );
    }

  }
}

/* Reduce this to ActMarkLinePlayer action */
void ReductionActionVisitor::visit(ActMarkLineReg *act)
{
  /* Retrieve the region */
  Region *reg = (Region *) act->getRegion();

  /* UNumSet needed for converting ActMarkLineReg to ActMarkLinePlayer act */
  UNumSet newUNumSet;

  /* Now, go through all the opponent players and determine who is in the
     region */
  for(UNum u = UNum::u1; u.getUNum() <= UNum::u11; u++){
    
    /* Retrieve the UNum of the current teammate */
    UNum::unum_t currUNum = u.getUNum();
    
    
    /* If the location of the opponent players is in high confidence and
       the opponent is in the region, then add it in */
    if( MyWorldModel.playerConfidenceGood(currUNum, false) ) {

      /* First retrieve the location of the opponent player */
      VecPosition pos = MyWorldModel.getGlobalPosition(currUNum, false);

      /* Add to our UNumSet if the opponent player is in the region */
      if( reg->isPtIn(pos, &MyWorldModel, bindings) )
	newUNumSet.add(u);

    }
	
  }   /* End of for-loop */

  /* Now construct our converted ActMarkLinePlayer action */
  ActMarkLinePlayer *newact = new ActMarkLinePlayer(newUNumSet);

  /* Call visitor on the new action for 2nd reduction */
  newact->accept(*this);
}

/* Reduction function for ActOffsidesLine actions */
void ReductionActionVisitor::visit(ActOffsidesLine *act)
{
  /* For now, ActOffsidesLine actions are not reducible */
  atomizedActions.push_back(act);
}

/* This is an obsolete action, so ignore */
void ReductionActionVisitor::visit(ActHetType *)
{

}

/* Ignore named action */
void ReductionActionVisitor::visit(ActNamed *)
{
  /* Namedy actions are invalid, so don't do
     anything */
}


void ReductionActionVisitor::visit(ActPassReg *act)
{
  /* Retrieve the region */
  Region *reg = (Region *) act->getRegion();

  /* UNumSet needed for converting ActMarkLineReg to ActMarkLinePlayer act */
  UNumSet newUNumSet;

  /* Now, go through all the players and determine who is in the region */
  for(UNum u = UNum::u1; u.getUNum() <= UNum::u11; u++){
    
    /* Retrieve the UNum of the current teammate */
    UNum::unum_t currUNum = u.getUNum();
    
    /* If the location of the teammate player is in high confidence and
       the teammate is in the region, then add it in */
    if( MyWorldModel.playerConfidenceGood(currUNum, true) ) {

      /* First retrieve the location of the teammate */
      VecPosition pos = MyWorldModel.getGlobalPosition(currUNum, true);

      /* Add to our UNumSet if the player is in the region */
      if( reg->isPtIn(pos, &MyWorldModel, bindings) )
	newUNumSet.add(u);

    }
	
  }   /* End of for-loop */

  /* Now construct our converted ActPassUNum action */
  ActPassUNum *newact = new ActPassUNum(newUNumSet);

  /* Call visitor on the new action for 2nd reduction */
  newact->accept(*this);
}

/* Reduction function for ActPassUNum actions */
void ReductionActionVisitor::visit(ActPassUNum *act)
{
  /* Boolean flags for which players are in the UNumSet */
  bool playerList[12] = {false, false, false, false, false, false,
			 false, false, false, false, false, false};


  /* Apply the 2nd reduction: Multi-UNumSet -> SingleUNumSet */
  UNumSet multiu = act->getPlayers();


  /* Size of the UNum Set */
  int size = multiu.size();


  /* For each UNum in the UNumSet, create ActMark acts, and
     push them into the atomized list */
  for(int i = 0; i < size; i++) {
    

    /* Retrieve the current UNum */
    UNum curUNum = multiu.getIndexAt(i);


    /* Variable name if it exists */
    std::string varName = curUNum.getVar();


    /* If the current UNum is a variable, */
    if(varName != "")    {

      /* Look up the UNumSet for the bindings */
      UNumSet varu = bindings.lookup(varName);

      /* Retrieve its size */
      int varSize = varu.size();


      /* Go through its list and atomize */
      for(int j = 0; j < varSize; j++) {
	
	/* Turn on the boolean flag */
	playerList[(int) (varu.getIndexAt(j).getUNum())] = true;

	/* If we have encountered 0 somewhere, terminate */
	if(playerList[0])
	  break;
      }

    }

    /* If it's just a UNum, */
    else
      playerList[(int) curUNum.getUNum()] = true;


    /* If we have encountered the UNum = 0 somewhere, terminate */
    if(playerList[0])
      break;

  }
  
  /* Check the agent's UNum number */
  int agentUNum = (int) MyWorldModel.getMyNum().getUNum();


  /* Add our atomized actions */
  for(int i = 1; i <= 11; i++)     {
      
    /* Add it to our atomized list */
    if( (playerList[0] || playerList[i]) && i != agentUNum) {
      
      UNumSet singleu;

      singleu.add((UNum::unum_t) i);

      atomizedActions.push_back(ActionBinding(new ActPassUNum(singleu),
					      bindings,
					      sourceType) );
    }

  }

}

/* Reduction function for ActDribble actions */
void ReductionActionVisitor::visit(ActDribble *act)
{
  /* For now, ActDribble can't be reduced */
  atomizedActions.push_back( ActionBinding(act, bindings, sourceType));
}

/* Reduction function for ActClear actions */
void ReductionActionVisitor::visit(ActClear *act)
{
  /* For now, ActClear can't be reduced */
  atomizedActions.push_back(ActionBinding(act, bindings, sourceType));
}

/* Reduction function for ActShoot actions */
void ReductionActionVisitor::visit(ActShoot *act)
{
  /* ActShoot can't be reduced */
  atomizedActions.push_back(ActionBinding(act, bindings, sourceType));
}

/* Reduction function for ActHold actions */
void ReductionActionVisitor::visit(ActHold *act)
{
  /* ActHold actions can't be reduced */
  atomizedActions.push_back(ActionBinding(act, bindings, sourceType));
}

/* Reduction function for ActIntercept actions */
void ReductionActionVisitor::visit(ActIntercept *act)
{
  /* ActIntercept actions can't be reduced */
  atomizedActions.push_back(ActionBinding(act, bindings, sourceType));
}

/* Reduction function for ActTackle actions */
void ReductionActionVisitor::visit(ActTackle *act)
{
  /* Boolean flags for which players are in the UNumSet */
  bool playerList[12] = {false, false, false, false, false, false,
			 false, false, false, false, false, false};


  /* Apply the 2nd reduction: Multi-UNumSet -> SingleUNumSet */
  UNumSet multiu = act->getPlayers();


  /* Size of the UNum Set */
  int size = multiu.size();


  /* For each UNum in the UNumSet, create ActMark acts, and
     push them into the atomized list */
  for(int i = 0; i < size; i++) {
    

    /* Retrieve the current UNum */
    UNum curUNum = multiu.getIndexAt(i);


    /* Variable name if it exists */
    std::string varName = curUNum.getVar();


    /* If the current UNum is a variable, */
    if(varName != "")    {

      /* Look up the UNumSet for the bindings */
      UNumSet varu = bindings.lookup(varName);

      /* Retrieve its size */
      int varSize = varu.size();


      /* Go through its list and atomize */
      for(int j = 0; j < varSize; j++) {
	
	/* Turn on the boolean flag */
	playerList[(int) (varu.getIndexAt(j).getUNum())] = true;

	/* If we have encountered 0 somewhere, terminate */
	if(playerList[0])
	  break;
      }

    }

    /* If it's just a UNum, */
    else
      playerList[(int) curUNum.getUNum()] = true;


    /* If we have encountered the UNum = 0 somewhere, terminate */
    if(playerList[0])
      break;

  }

  /* Add our atomized actions */
  for(int i = 1; i <= 11; i++)     {
      
    /* Add it to our atomized list */
    if(playerList[0] || playerList[i]) {
      
      UNumSet singleu;

      singleu.add((UNum::unum_t) i);

      atomizedActions.push_back(ActionBinding(new ActTackle(singleu),
					      bindings,
					      sourceType) );
    }

  }

}



/* Testing purposes: Uncomment this part to test reduction/atomization  */
/*
void main()
{
  VarBindings b;
  ReductionActionVisitor r;
  vector <ActionBinding> v;



  ActPos *act = new ActPos();

  ReductionActionVisitor r(ActionBinding(act, b, ActionBinding::UNDEFINED));

  act->accept(r);

  v = r.getAtomizedActions();

  printf("Atomizing ActPos action gave %d actions\n", v.size());
}
*/
