/* Dongryeol Lee
 * 
 * RegionVisitor.C
 *
 * Visitor Class for Region
 * Extracts information about the region type, region vertices, etc.
 * Implemented functionality to calculate region area, return
 * point inside a region, return point outside a region, test if 
 * a point lies inside a region, etc.
 */

#include "Geometry.h"
#include "RegionVisitor.h"
#include "PointVisitor.h"
#include "visitor.h"
#include "hasa.h"
#include "UvaLogger.h"
#include <stdio.h>
#include <math.h>
#include <typeinfo>
#include <stdlib.h>
#include <time.h>
#include <vector>
#include "msgstore.h"
#include "config.h"

using namespace rcss::clang;
using namespace rcss::util;
using namespace std;

typedef vector <RegionVisitor::myRegionType>::const_iterator region_iterator;
typedef vector <VecPosition>::const_iterator pos_iterator;

extern MsgStore<Region*> defregstorage;

#ifndef EPSILON
#define EPSILON 0.01
#endif


/* Constructor function for RegionVisitor: initialize the region area to 0 */
void RegionVisitor::init()
{
  regionArea = 0;
  region_xmax = -10000;
  region_xmin = 10000;
  region_ymax = -10000;
  region_ymin = 10000;

  found_variables = false;
}

/* Simply return the accumulated area for the region */
double RegionVisitor::getRegionArea() const
{
  return regionArea;
}

/* 4 accessor methods to return the xmax, xmin, ymax, ymin of the region */
double RegionVisitor::getRegionXMax() const
{
  return region_xmax;
}

double RegionVisitor::getRegionXMin() const
{
  return region_xmin;
}

double RegionVisitor::getRegionYMax() const
{
  return region_ymax;
}

double RegionVisitor::getRegionYMin() const
{
  return region_ymin;
}


/* This method takes three points on the cartesian plane and returns the
   area formed by the trianlge of three points (without taking the sign
   into account). Could return negative or negative. The user needs to
   take the absolute value if he/she wants the area */
double RegionVisitor::calculateSignedTriangleArea(const VecPosition &v0, 
						  const VecPosition &v1, 
						  const VecPosition &v2) const
{
  return (0.5 * (v0.getX() * v1.getY() + v1.getX() * v2.getY() +
		 v2.getX() * v0.getY() - v1.getX() * v0.getY() -
		 v2.getX() * v1.getY() - v0.getX() * v2.getY()));
}

/* The idea of this function is to see how the current region can be bounded by a rectangular region,
   defined by region_xmax, region_xmin, region_ymin, region_ymax. This will be helpful in the 
   implementation of the RegionIterator in extracting potential points */
void RegionVisitor::updateBoundary(vector <VecPosition> v)
{
  for(unsigned int i = 0; i < v.size(); i++) {
    
    /* Update max x value of the boundary */
    if(region_xmax < v[i].getX())
      region_xmax = v[i].getX();

    /* min x value */
    if(region_xmin > v[i].getX())
      region_xmin = v[i].getX();

    /* Now, max y value of the boundary */
    if(region_ymax < v[i].getY())
      region_ymax = v[i].getY();
    
    /* min y value */
    if(region_ymin > v[i].getY())
      region_ymin = v[i].getY();
  }
}

/*
 * Added for variable handling
 * Returns true if the region visitor encountered any clang variables. 
 * Return false otherwise. 
 */
bool RegionVisitor::hasVariables() const
{

  return found_variables;
}



/* If a given region is generic Region, use Dynamic casting to 
   determine the appropriate region type and call its visitor */
void RegionVisitor::visit(RegNull *)
{
  // jdavin: 
  // I changed this function to RegNull (instead of Region*) because 
  // RegNull wasn't handled, and if even you're not going to use it, you should 
  // still define the visit function to be safe. (and print out a message if we get a RegNull?)

  //Log.log(15, "We've got a Null Region\n");
}

/* Unimplemented: just display that it is invalid */
void RegionVisitor::visit(RegQuad *)
{
  //Log.log(15, "Use of RegQuad is invalid\n");
}

/* Visit the area defined by arc and calculate its area */
void RegionVisitor::visit(RegArc *reg)
{
  /* Pass this vector to update the boundary of the region if necessary */
  vector <VecPosition> v;

  /* Retrieve the center of the arc */
  Point *center = (Point *) reg->getCenter();
  PointVisitor pv(myBindings);   /* Declare point visitor */

  /* Retrieve the start radius, end radius, start angle, span angle */
  double startRad = reg->getStartRad();
  double endRad = reg->getEndRad();
  double startAng = reg->getStartAng();
  double spanAng = reg->getSpanAng();

  /* Visit the point to calculate the global position of the center */
  center->accept(pv);
  found_variables |= pv.hasVariables();
  myBindings.mergeWithIntersect(pv.getBindings());

  /* Use the formula: Area = 0.5 * (r1^2 - r2^2) * theta */
  regionArea += 0.5 * (startRad * startRad - endRad * endRad) 
    * (spanAng - startAng);

  /* Record that an Arc region has been visited */
  regionsVisited.push_back(ARC);

  /* Transform the 4 scalars into 4 VecPositions and just simply
     push back into the vector */
  VecPosition ct = pv.returnGlobalPosition();

  points.push_back(ct);
  points.push_back(VecPosition(startRad, 0));
  points.push_back(VecPosition(endRad, 0));
  points.push_back(VecPosition(startAng, 0));
  points.push_back(VecPosition(spanAng, 0));

  /* Now push back the 4 points defining the arc region and update boundary */
  /* Calculate the points forming the region arc */

  VecPosition p0 = ct + VecPosition::getVecPositionFromPolar(startRad, startAng);
  VecPosition p1 = ct + VecPosition::getVecPositionFromPolar(endRad, startAng);
  VecPosition p2 = ct + VecPosition::getVecPositionFromPolar(startRad, spanAng);
  VecPosition p3 = ct + VecPosition::getVecPositionFromPolar(endRad, spanAng);

  /* Push back the 4 points */
  v.push_back(p0);
  v.push_back(p1);
  v.push_back(p2);
  v.push_back(p3);

  /* Update boundary if necessary */
  updateBoundary(v);
}

/* Method to visit a union of regions: just retrieve its children
   and visit all of them and calculate the union of their area */
void RegionVisitor::visit(RegUnion *reg)
{
  /* Get the list of all regions */
  RegUnion::HasManyRegions::Storage allRegions = reg->getRegions();

  /* Iterate through the children and visit them in order */
  for(RegUnion::HasManyRegions::Storage::iterator it = allRegions.begin(); 
      it != allRegions.end(); it++) {
    Region *child = *it;
    //    RegionVisitor::visit(child);

    // jdavin: 
    // we need to create a new visitor to visit the child Region. 
    RegionVisitor visitor2(myBindings);
    child->accept(visitor2);
    // now call functions on visitor2 to determine whatever info you need, if any. 

    // get the child's bindings
    found_variables |= visitor2.hasVariables();
    myBindings.mergeWithIntersect(visitor2.getBindings());

  }
}

/* Method to visit a region defined by a point. Not much to do here
   since area = 0 */
void RegionVisitor::visit(RegPoint *reg)
{
  /* temporary vector holder for updating the boundary */
  vector <VecPosition> v;

  PointVisitor pv(myBindings);   /* PointVisitor for the point region */
  Point *p = reg->getPoint();   /* retrieve the point from the region */

  p->accept(pv);   /* visit the point */

  found_variables |= pv.hasVariables();
  myBindings.mergeWithIntersect(pv.getBindings());

  //if(found_variables)
  // PR(9, "[RegionVisitor]: found variables in point\n");
  //else
  // PR(9, "[RegionVisitor]: did not find variables in point\n");

  /* Record that point region has been visited */
  regionsVisited.push_back(POINT);
  
  /* Push back the position of the region */
  points.push_back(pv.returnGlobalPosition());

  /* Push back the position of the region again and update boundary */
  v.push_back(pv.returnGlobalPosition());
  updateBoundary(v);
}

/* Named region is just a region, so visit it */
void RegionVisitor::visit(RegNamed *reg)
{
  // look up reg->getName() in my defregstorage container. 
  // then visit the region that is returned from storage. 

  Region *theregion = (Region *) reg->getRegion();

  theregion->accept(*this);
}

/* Visit a triangular region and retrieve its area and vertices */
void RegionVisitor::visit(RegTri *reg)
{
  /* Vector holder for updating boundary */
  vector <VecPosition> v;

  /* Declare point visitor each point of the triangle */
  PointVisitor pv0(myBindings), pv1(myBindings), pv2(myBindings);

  /* Retrieve the three points of the triangle */
  Point *p0 = (Point *) reg->getPt(0);
  Point *p1 = (Point *) reg->getPt(1);
  Point *p2 = (Point *) reg->getPt(2);

  /* Will store the reduced global position of the point from
     "PointVisitor" */
  VecPosition v0, v1, v2;

  /* Visit p0, p1, p2 to retrieve their global positions */
  p0->accept(pv0);
  p1->accept(pv1);
  p2->accept(pv2);

  found_variables |= pv0.hasVariables() || pv1.hasVariables() || pv2.hasVariables();
  myBindings.mergeWithIntersect(pv0.getBindings());
  myBindings.mergeWithIntersect(pv1.getBindings());
  myBindings.mergeWithIntersect(pv2.getBindings());

  /*Retrieve the global positions of the points from the PointVisitor */
  v0 = pv0.returnGlobalPosition();
  v1 = pv1.returnGlobalPosition();
  v2 = pv2.returnGlobalPosition();

  /* Calculate the area of the triangle */
  regionArea += fabs(calculateSignedTriangleArea(v0, v1, v2));

  /* Record that a triangle region has been visited */
  regionsVisited.push_back(TRI);

  /* Push back the three points of the triangle */
  points.push_back(v0);
  points.push_back(v1);
  points.push_back(v2);

  /* Push back the vertices into holder and update boundary */
  v.push_back(v0);
  v.push_back(v1);
  v.push_back(v2);
  updateBoundary(v);
}

/* Method to visit a rectangular region defined by two opposite
   corner points and calculate its area */
void RegionVisitor::visit(RegRec *reg)
{
  /* Vector holder for updating boundary */
  vector <VecPosition> v;

  /* Declare point visitor for each point on the corner of the
     rectangle */
  PointVisitor pv0(myBindings), pv1(myBindings);
  
  /* Retrieve the corners of the rectangle */
  Point *p0 = (Point *) reg->getPt(0);
  Point *p1 = (Point *) reg->getPt(1);

  VecPosition v0, v1;  /* will store the reduced global position of
			  the point from "PointVisitor" */


  /* visit p0 and p1 to get the global position of
     the corners of the rectangle */
  p0->accept(pv0);
  p1->accept(pv1);

  found_variables |= pv0.hasVariables() || pv1.hasVariables();
  myBindings.mergeWithIntersect(pv0.getBindings());
  myBindings.mergeWithIntersect(pv1.getBindings());
  
  /*
    if(found_variables)
    PR(9, "[RegionVisitor]: found variables in RegRec\n");
    else
    PR(9, "[RegionVisitor]: did not find variables in RegRec\n");
  */

  //  PR(9, "[RegionVisitor:RegRec]: myBindings are now: \n");
  //myBindings.PrintAll();

  /* Retrieve the global positions of the points from the
     PointVisitor */
  v0 = pv0.returnGlobalPosition();
  v1 = pv1.returnGlobalPosition();
  
  /* Calculate the area of the rectangle */
  regionArea += fabs(v0.getX() - v1.getX()) *
    fabs(v0.getY() - v1.getY());

  /* Record that a rectanglular region has been visited */
  regionsVisited.push_back(REC);

  /* Push back the two corner points */
  points.push_back(v0);
  points.push_back(v1);

  /* Push back the two corner points for updating boundary */
  v.push_back(v0);
  v.push_back(v1);
  updateBoundary(v);
}

/* Given the center, starting radius, ending radius, starting angle, span angle, 
   determine whether point p is in the arc region */
int RegionVisitor::isPointInsideRegArc(const VecPosition &center, double startRad, double endRad,
			double startAng, double spanAng, const VecPosition &p) const
{
  /* Calculate the directional vector connecting the center and p, and its 
     direction and its magnitude  */
  VecPosition centerTop = p - center;
  double directionAngle = centerTop.getDirection();
  double magnitude = centerTop.getMagnitude();

  /* If the angle is negative, add 360 degrees to it compensate for it. */
  if(directionAngle < 0)
    directionAngle += 360.0;

  return (directionAngle >= startAng && directionAngle <= spanAng &&
	  magnitude >= startRad && magnitude <= endRad);
}

/* Given the three points of a triangle, determine whether point p is in
   the region */
int RegionVisitor::isPointInsideRegTri(const VecPosition& p0, 
				       const VecPosition& p1, 
				       const VecPosition& p2, 
				       const VecPosition& p) const
{
  /* calculate the area formed by triangle p0, p1, p, triangle p1, p2, p,
     and triangle p2, p0, p */
  double area1 = calculateSignedTriangleArea(p, p0, p1);
  double area2 = calculateSignedTriangleArea(p, p1, p2);
  double area3 = calculateSignedTriangleArea(p, p2, p0);

  return((area1 >= 0 && area2 >= 0 && area3 >= 0) ||
	 (area1 <= 0 && area2 <= 0 && area3 <= 0));
}

/* Determine if the point p is in the rectangle, given two corner points */
int RegionVisitor::isPointInsideRegRec(const VecPosition& p0, 
				       const VecPosition& p1, 
				       const VecPosition& p) const
{
  double minx, miny, maxx, maxy;
  double px = p.getX();
  double py = p.getY();

  //Log.log(8, "[isPointInsideRegRec]: chking if point %f %f is in (%f %f)x(%f %f)\n", px, py, p0.getX(), p0.getY(), p1.getX(), p1.getY());

  if(p0.getX() > p1.getX() - EPSILON) {
    minx = p1.getX() - EPSILON;
    maxx = p0.getX() + EPSILON;
  }
  else {
    minx = p0.getX() - EPSILON;
    maxx = p1.getX() + EPSILON;
  }

  if(p0.getY() > p1.getY() - EPSILON) {
    miny = p1.getY() - EPSILON;
    maxy = p0.getY() + EPSILON;
  }
  else {
    miny = p0.getY() - EPSILON;
    maxy = p1.getY() + EPSILON;
  }
  
  //Log.log(8, "[isPointInsideRegRec]: is point %f %f between min (%f %f) x max (%f %f)\n", px, py, minx, miny, maxx, maxy);

  return (px >= minx && px <= maxx && py >= miny && py <= maxy);

}

/* This method tests whether a global position point lies in or on the given
   region. This is the heart of all of the region methods defined in 
   CoachablePlayer */
int RegionVisitor::isPointInsideRegion(const VecPosition& p) const
{
  int result = 0;
  VecPosition v0, v1, v2, v3, v4;


  /* Point to the beginning of the vertices vector for iteration */
  pos_iterator vertices = points.begin();

  /* Iterate through each region, and break out if a point p is in any of
     the regions */
  for(region_iterator regions = regionsVisited.begin();
      regions != regionsVisited.end() && !result; regions++) {

    /* Fall into cases to test for point falling into a region */
    switch(*regions) {
    case QUAD:
      //Log.log(15, "RegQuad unsupported\n");
      break;

      /* Peel off 5 "vertices" off from the vertices list and call the 
	 function */
    case ARC:
      v0 = *(vertices++);
      v1 = *(vertices++);
      v2 = *(vertices++);
      v3 = *(vertices++);
      v4 = *(vertices++);

      result = isPointInsideRegArc(v0, v1.getX(), v2.getX(), v3.getX(), 
				   v4.getX(), p);
      break;
      /* Simply determine whether two points are equal */
    case POINT:
      result = (*(vertices++) == p);
      break;
      /* Peel off 3 "vertices" and call testing for RegTri */
    case TRI:
      v0 = *(vertices++);
      v1 = *(vertices++);
      v2 = *(vertices++);

      result = isPointInsideRegTri(v0, v1, v2, p);
      break;
      /* Peel off 2 corner "vertices" and call testing for RegRec */
    case REC:
      v0 = *(vertices++);
      v1 = *(vertices++);
      
      /*
      PR(11, "[isPointInsideRegion]: is pos %f %f inside (%f %f) x (%f %f)\n", p.getX(), p.getY(), v0.getX(), v0.getY(), v1.getX(), v1.getY());
      */
      
      result = isPointInsideRegRec(v0, v1, p);
      break;
    default:
      //Log.log(15, "Unsupported region type\n");
      break;
    }
  }
  return result;
}

VecPosition RegionVisitor::getPointClosestToObjectWithinSet(vector <VecPosition> solutions, 
							    const VecPosition &objectPos) const
{
  /* Initialize the closest point position to bogus location on the field really
     far away from the object */
  VecPosition closest(-10000, -10000);
  
  for(pos_iterator p = solutions.begin(); p != solutions.end(); p++) {

    /* Retrieve the current point */
    VecPosition result = (*p);

    /* If the current point is closer to the object than before, update */
    if(result.getDistanceTo(objectPos) < closest.getDistanceTo(objectPos))
      closest = result;
  }
  
  /* Return the closest point of all */
  return closest;
}
  
VecPosition RegionVisitor::getPointClosestFromRegArcToObject(const VecPosition &center,
  double startRad, double endRad, double startAng, double spanAng, const VecPosition &objectPos) const
{
  /* closest point of all */
  VecPosition closest, result;

  /* First, we need to make a line segment connecting the object to the center of the arc
     region, to see if any of them intersect */
  Line l1 = Line::makeLineFromTwoPoints(objectPos, center);

  /* Calculate the points forming the region arc */
  VecPosition p0 = center + VecPosition::getVecPositionFromPolar(startRad, startAng);
  VecPosition p1 = center + VecPosition::getVecPositionFromPolar(endRad, startAng);
  VecPosition p2 = center + VecPosition::getVecPositionFromPolar(startRad, spanAng);
  VecPosition p3 = center + VecPosition::getVecPositionFromPolar(endRad, spanAng);

  /* Make the lines connecting the 4 vertices of the arc region from the object position */
  Line l2 = Line::makeLineFromTwoPoints(objectPos, p0);
  Line l3 = Line::makeLineFromTwoPoints(objectPos, p1);
  Line l4 = Line::makeLineFromTwoPoints(objectPos, p2);
  Line l5 = Line::makeLineFromTwoPoints(objectPos, p3);

  /* Intersect all three lines with the arc region and retrieve the closest intersection */
  
  /* Line 1 */
  closest = getPointClosestToObjectWithinSet(intersectArc(center, startRad, endRad, startAng, spanAng, l1), objectPos);

  /* Line 2 */
  result = getPointClosestToObjectWithinSet(intersectArc(center, startRad, endRad, startAng, spanAng, l2), objectPos);

  /* If the 2nd intersection was closer than the 1st one, update */
  if(result.getDistanceTo(objectPos) < closest.getDistanceTo(objectPos))
    closest = result;

  /* Line 3 */
  result = getPointClosestToObjectWithinSet(intersectArc(center, startRad, endRad, startAng, spanAng, l3), objectPos);

  /* If the 3rd one is better than the first two, update */
  if(result.getDistanceTo(objectPos) < closest.getDistanceTo(objectPos))
    closest = result;

  /* Line 4 */
  result = getPointClosestToObjectWithinSet(intersectArc(center, startRad, endRad, startAng, spanAng, l4), objectPos);

  /* If the 4th one is better than the first three, update */
  if(result.getDistanceTo(objectPos) < closest.getDistanceTo(objectPos))
    closest = result;

  /* Line 5 */
  result = getPointClosestToObjectWithinSet(intersectArc(center, startRad, endRad, startAng, spanAng, l5), objectPos);

  /* If the 5th one is better than the first four, update */
  if(result.getDistanceTo(objectPos) < closest.getDistanceTo(objectPos))
    closest = result;

  /* Return the closest point */
  return closest;
}

VecPosition 
RegionVisitor::getPointClosestFromRegTriToObject(const VecPosition &v0,
						 const VecPosition &v1,
						 const VecPosition &v2,
						 const VecPosition &objectPos) const
{
  /* Solution */
  VecPosition closest, result;

  /* First, make the lines of three sides of the triangle and get their "tangent lines"
     going through the position of the object */
  Line l1 = Line::makeLineFromTwoPoints(objectPos, v0);
  Line l2 = Line::makeLineFromTwoPoints(objectPos, v1);
  Line l3 = Line::makeLineFromTwoPoints(objectPos, v2);

  /* Intersect all three lines with the triangular region and retrieve intersecitons */
  /* Line 1*/
  closest = getPointClosestToObjectWithinSet(intersectTri(v0, v1, v2, l1), objectPos);

  /* Line 2 */
  result = getPointClosestToObjectWithinSet(intersectTri(v0, v1, v2, l2), objectPos);

  /* If the 2nd one is better than the first one, update */
  if(result.getDistanceTo(objectPos) < closest.getDistanceTo(objectPos))
    closest = result;

  /* Line 3 */
  result = getPointClosestToObjectWithinSet(intersectTri(v0, v1, v2, l3), objectPos);

  /* If the 3rd one is better than the first two, update */
  if(result.getDistanceTo(objectPos) < closest.getDistanceTo(objectPos))
    closest = result;
  
  /* Return the closest of all to the object within the gathered solution */
  return closest;
}

VecPosition 
RegionVisitor::getPointClosestFromRegRecToObject(const VecPosition& v0,
						 const VecPosition& v1,
						 const VecPosition& objectPos) const
{
  /* Sub-Solution list */
  VecPosition s0, s1, s2, s3;

  /* Vector of the sub-solutions */
  vector <VecPosition> sl;
  
  /* Now retrieve the x coordinates and y-cooredinates of the 4 corner points */
  double x1 = v0.getX();
  double x2 = v1.getX();
  double y1 = v0.getY();
  double y2 = v1.getY();

  /* First, make the 4 lines connecting the object and each of the 4 vertices */
  Line l0 = Line::makeLineFromTwoPoints(objectPos, VecPosition(x1, y1));
  Line l1 = Line::makeLineFromTwoPoints(objectPos, VecPosition(x1, y2));
  Line l2 = Line::makeLineFromTwoPoints(objectPos, VecPosition(x2, y1));
  Line l3 = Line::makeLineFromTwoPoints(objectPos, VecPosition(x2, y2));
  /* Intersect all lines with the rectangular region and retrieve intersecitons */
  s0 = getPointClosestToObjectWithinSet(intersectRec(v0, v1, l0), objectPos);
  s1 = getPointClosestToObjectWithinSet(intersectRec(v0, v1, l1), objectPos);
  s2 = getPointClosestToObjectWithinSet(intersectRec(v0, v1, l2), objectPos);
  s3 = getPointClosestToObjectWithinSet(intersectRec(v0, v1, l3), objectPos);

  /* First, push back the 4 sub-solutions and return the most optimal one */
  sl.push_back(s0);
  sl.push_back(s1);
  sl.push_back(s2);
  sl.push_back(s3);

  /* return the most optimal of the 4 */
  return getPointClosestToObjectWithinSet(sl, objectPos);
} 



/* Retrieve a point closest from Region to Object. Basically it will iterate through
   all the stored region, and return the closest found point to a given object.
   This method is different from selectRandomPointInsideRegion in that it will choose
   a point by using a deterministic method.
*/
VecPosition 
RegionVisitor::getPointClosestFromRegionToObject(const VecPosition& objectPos) const
{
  /* Iterator for the vertices */
  pos_iterator vertices = points.begin();

  /* Some necessary VecPosition stuffs: closestPoint will represent the closest
     point in the region to the object */
  VecPosition closestPoint(-10000, -10000), result, v0, v1, v2, v3, v4;

  /* If the ball is already inside the region, obviously it will be point we want */
  if(isPointInsideRegion(objectPos))
    return objectPos;

  /* Iterate through all subregions and choose the closest point of all */
  for(region_iterator r = regionsVisited.begin(); r != regionsVisited.end(); r++) {

    /* Fall through case statement */
    switch(*r) {

    case ARC:
      v0 = *(vertices++);
      v1 = *(vertices++);
      v2 = *(vertices++);
      v3 = *(vertices++);
      v4 = *(vertices++);
      
      /* get the closest point from the object to the arc region */
      result =  getPointClosestFromRegArcToObject(v0, v1.getX(), v2.getX(), v3.getX(), v4.getX(), objectPos);
      
      break;

      /* Simply return the point itself */
    case POINT:
      result = *vertices;
      break;
      /* Peel off 3 "vertices" and call selection for RegTri */
    case TRI:
      v0 = *(vertices++);
      v1 = *(vertices++);
      v2 = *(vertices++);

      /* Calculate the point closest to the given object in the triangular region */
      result = getPointClosestFromRegTriToObject(v0, v1, v2, objectPos);

      break;
      /* Peel off 2 corner "vertices" and call selection for RegRec */
    case REC:
      v0 = *(vertices++);
      v1 = *(vertices++);
      
      /* If the object of our interest is already in the rectangular region, return its
	 position */
      if(isPointInsideRegRec(v0, v1, objectPos))
	return objectPos;
      else
	return getPointClosestFromRegRecToObject(v0, v1, objectPos);

      break;
    default:
      //Log.log(15, "Unsupported region type\n");
      break;
    }

    /* If the closest point found just now is better than what we have before, update */
    if(isPointInsideRegion(result) && 
       result.getDistanceTo(objectPos) < closestPoint.getDistanceTo(objectPos))
      closestPoint = result;
  }

  /* Return the closest point */
  if(isPointInsideRegion(closestPoint))
    return closestPoint;
  else
    return objectPos;
}


/* Basically randomly generate a random radius between starting radius and the ending
   radius, and the angle between starting angle and spanning one. Then, generate 
   a VecPosition that is offset from the center. The result is simply adding this vector
   with the center vector position */
VecPosition 
RegionVisitor::selectRandomPointInsideArc(const VecPosition &center, 
					  double startRad, 
					  double endRad, 
					  double startAng, 
					  double spanAng) const
{
  double newRad = (startRad + endRad) / 2.0;
  double newAng = (startAng + spanAng * 0.5);
  
  VecPosition p(newRad * cosDeg(newAng), newRad * sinDeg(newAng));

  return center + p;
}

/* This method selects a random point inside a triangular region */
VecPosition 
RegionVisitor::selectRandomPointInsideTri(const VecPosition& v0, 
					  const VecPosition& v1, 
					  const VecPosition& v2) const
{
  /* Return the point inside the triangle */
  return ((v0 + v1 + v2) / 3.0);
}

/* This method selects a random point inside a rectangular region */
VecPosition 
RegionVisitor::selectRandomPointInsideRec(const VecPosition &v0, 
					  const VecPosition &v1) const
{
  /* Return the middle point of the region */
  return VecPosition(0.5 * (v0.getX() + v1.getX()), 0.5 * (v0.getY() + v1.getY()));
}

/* This method basically selects a random region out of the set of
   regions and selects a random point in the selected region */
VecPosition RegionVisitor::selectRandomPointInsideRegion() const
{
  /* Get the number of regions visited */
  int numRegions = regionsVisited.size();
  assert(numRegions != 0);
  
  /* And use it to generate a random number to select the region */
  int regionNumber = rand() % numRegions;


  /* Get the iterator for the regions visited and the points */
  region_iterator r = regionsVisited.begin();
  pos_iterator vertices = points.begin();
  VecPosition result, v0, v1, v2, v3, v4;
  
  /* Get the appropriate region by iterating to it */
  while(regionNumber-- > 0) {
    switch(*r) {
    case ARC:
      vertices += 5;
      break;
    case POINT:
      vertices++;
      break;
    case TRI:
      vertices += 3;
      break;
    case REC:
      vertices += 2;
	break;
    default:
      //Log.log(15, "Unsupported region type\n");
      break;
    }
    r++;
  }
  
  /* Fall through yet another case statement */
  switch(*r) {
    /* Retrieve the 5 information and call selectRandomPointInsideArc */
  case ARC:
    v0 = *(vertices++);
    v1 = *(vertices++);
    v2 = *(vertices++);
    v3 = *(vertices++);
    v4 = *(vertices++);
    
    result = selectRandomPointInsideArc(v0, v1.getX(), v2.getX(), v3.getX(), 
					v4.getX());
    break;
    /* Simply return the point itself */
  case POINT:
    result = *vertices;
    break;
    /* Peel off 3 "vertices" and call selection for RegTri */
  case TRI:
    v0 = *(vertices++);
    v1 = *(vertices++);
    v2 = *(vertices++);
    
    result = selectRandomPointInsideTri(v0, v1, v2);
    break;
    /* Peel off 2 corner "vertices" and call selection for RegRec */
  case REC:
    v0 = *(vertices++);
    v1 = *(vertices++);
    
    result = selectRandomPointInsideRec(v0, v1);
    break;
  default:
    //Log.log(15, "Unsupported region type\n");
    break;
  }

  /* Return the randomly selected point */
  return result;
}

/* Calculate the intersection with the line and the arc region and return
   intersections. Technically, what this function does is calculating the intersection
   of the line with the two concentric circles forming the arc. */
vector <VecPosition> RegionVisitor::intersectArc(const VecPosition &center, 
						 double startRad, 
						 double endRad, 
						 double startAng, 
						 double spanAng, 
						 const Line &l) const
{
  /* Solution vector for the intersection */
  VecPosition i1, i2;

  /* List of solutions to return */
  vector <VecPosition> solution;

  /* Number of solutions returned from the intersecting */
  int numSolutions;

  /* 1: Solve the intersection with the first inner circle */
  numSolutions = l.getCircleIntersectionPoints(Circle(center, startRad), &i1, &i2);
  
  /* If there is at least one solution, take the first solution and generate random points */
  if(numSolutions > 0 && isPointInsideRegArc(center, startRad, endRad, startAng, spanAng, i1))
    solution.push_back(i1);
  
  /* If there are two solutions, take the next one also. */
  if(numSolutions > 1 && isPointInsideRegArc(center, startRad, endRad, startAng, spanAng, i2))
    solution.push_back(i2);

  /* 2: Solve the intersection with the outer circle */
  numSolutions = l.getCircleIntersectionPoints(Circle(center, endRad), &i1, &i2);
  
  /* If there is at least one solution, take the first solution and generate random points */
  if(numSolutions > 0 && isPointInsideRegArc(center, startRad, endRad, startAng, spanAng, i1))
    solution.push_back(i1);
  
  /* If there are two solutions, take the next one also. */
  if(numSolutions > 1 && isPointInsideRegArc(center, startRad, endRad, startAng, spanAng, i2))
    solution.push_back(i2);
 
  return solution;
}


/* This method selects a random point inside a triangular region */
vector <VecPosition> RegionVisitor::intersectTri(const VecPosition &v0, 
						 const VecPosition &v1, 
						 const VecPosition &v2, 
						 const Line &l) const
{
  /* Generate the line for three sides, and calculate intersections */
  Line l1 = Line::makeLineFromTwoPoints(v0, v1);
  Line l2 = Line::makeLineFromTwoPoints(v1, v2);
  Line l3 = Line::makeLineFromTwoPoints(v2, v0);

  /* Calculate the intersections with the three sides of the triangle with the given line */
  VecPosition s1 = l.getIntersection(l1);
  VecPosition s2 = l.getIntersection(l2);
  VecPosition s3 = l.getIntersection(l3);

  vector <VecPosition> solution;   /* solution list */


  /* Push back three intersections, if they are at the boundary of the region */
  if(isPointInsideRegTri(v0, v1, v2, s1))
    solution.push_back(l.getIntersection(l1));
  if(isPointInsideRegTri(v0, v1, v2, s2))
    solution.push_back(l.getIntersection(l2));
  if(isPointInsideRegTri(v0, v1, v2, s3))
    solution.push_back(l.getIntersection(l3));

  /* Return the solution */
  return solution;
}


/* This method selects a random point inside a rectangular region */
vector <VecPosition> RegionVisitor::intersectRec(const VecPosition &v0, 
						 const VecPosition &v1, 
						 const Line &l) const
{
  vector <VecPosition> solution;   /* solution list */
  
  VecPosition s1(v0.getX(), l.getYGivenX(v0.getX()));
  VecPosition s2(v1.getX(), l.getYGivenX(v1.getX()));
  VecPosition s3(l.getXGivenY(v0.getY()), v0.getY());
  VecPosition s4(l.getXGivenY(v1.getY()), v1.getY());

  /* Push back the intersections, if they are valid */
  if(isPointInsideRegRec(v0, v1, s1))
    solution.push_back(s1);
  if(isPointInsideRegRec(v0, v1, s2))
    solution.push_back(s2);
  if(isPointInsideRegRec(v0, v1, s3))
    solution.push_back(s3);
  if(isPointInsideRegRec(v0, v1, s4))
    solution.push_back(s4);

  return solution;  /* return solution vector */
}

/* This method basically selects a random region outside a region.
   Basic algorithm is to select two random points inside a region and
   construct a line out of them. Take this line, and intersect with 
   the boundaries of every sub-regions. Take one of the boundary
   intersection points and return */
VecPosition RegionVisitor::selectRandomPointOutsideRegion() const
{  
  /* Get the iterator for the regions visited and the points */
  //region_iterator r = regionsVisited.begin();
  pos_iterator vertices = points.begin();
  VecPosition v0, v1, v2, v3, v4;

  /* Select two random points inside the region */
  VecPosition rp0 = selectRandomPointInsideRegion();
  VecPosition rp1 = selectRandomPointInsideRegion();

  /* Construct the line connecting two random points */
  Line l = Line::makeLineFromTwoPoints(rp0, rp1);
    
  /* store potential candidates here for random points outside a region */
  vector <VecPosition> result;

  for(region_iterator r = regionsVisited.begin(); r != regionsVisited.end(); r++) {
    switch(*r) {
      /* Retrieve the 5 information and call selectRandomPointInsideArc */
    case ARC:
      v0 = *(vertices++);
      v1 = *(vertices++);
      v2 = *(vertices++);
      v3 = *(vertices++);
      v4 = *(vertices++);
    
    result = intersectArc(v0, v1.getX(), v2.getX(), v3.getX(), v4.getX(), l);
    break;
    /* Simply return the point itself */
    case POINT:
      result.push_back(*(vertices++));
      break;
      /* Peel off 3 "vertices" and call selection for RegTri */
    case TRI:
      v0 = *(vertices++);
      v1 = *(vertices++);
      v2 = *(vertices++);
      
      result = intersectTri(v0, v1, v2, l);
      break;
      /* Peel off 2 corner "vertices" and call selection for RegRec */
    case REC:
      v0 = *(vertices++);
      v1 = *(vertices++);
      
      result = intersectRec(v0, v1, l);
      break;
    default:
      //Log.log(15, "Unsupported region type\n");
      break;
    }

    /* if any one of the points generated lie outside the region, immediately 
       return the first one */
    for(pos_iterator p = result.begin(); p != result.end(); p++) {
      VecPosition tempPoint = *p;
      VecPosition upperRight(tempPoint.getX() + 0.01, tempPoint.getY() + 0.01);
      VecPosition lowerRight(tempPoint.getX() + 0.01, tempPoint.getY() - 0.01);
      VecPosition upperLeft(tempPoint.getX() - 0.01, tempPoint.getY() + 0.01);
      VecPosition lowerLeft(tempPoint.getX() - 0.01, tempPoint.getY() - 0.01);
      VecPosition north(tempPoint.getX(), tempPoint.getY() + 0.01);
      VecPosition east(tempPoint.getX() + 0.01, tempPoint.getY());
      VecPosition west(tempPoint.getX() - 0.01, tempPoint.getY());
      VecPosition south(tempPoint.getX(), tempPoint.getY() - 0.01);

      if(!isPointInsideRegion(*p))
	return (*p);
      else if(!isPointInsideRegion(upperRight))
	return upperRight;
      else if(!isPointInsideRegion(lowerRight))
	return lowerRight;
      else if(!isPointInsideRegion(upperLeft))
	return upperLeft;
      else if(!isPointInsideRegion(lowerLeft))
	return lowerLeft;
      else if(!isPointInsideRegion(north))
	return north;
      else if(!isPointInsideRegion(east))
	return east;
      else if(!isPointInsideRegion(west))
	return west;
      else if(!isPointInsideRegion(south))
	return south;
    }
  }

  /* Return whatever here. This return statement deals with the case when no point
     can be found outside a region, but this shouldn't be happening all the time. 
     This just returns the last failed point as the candidate: but shouldn't have 
     to deal with this case. */
  return *(result.begin());
}
