/* Dongryeol Lee
 *
 * ScoringActionVisitor.C
 *
 * This file implements the ScoringActionVisitor class described in the header
 * file, ScoringActionVisitor.h
 *
 */


#include "Geometry.h"
#include "WorldModelAccessor.h"
#include "ScoringActionVisitor.h"


/* Our access to the player's world model */
extern WorldModelAccessor MyWorldModel;

extern WorldModel *trilearn_wm;


/* Denormalizing helper function */
double denormalize(const double & d)
{
  if( d < 0.0 )
    return d + 360.0;
  else if( d > 360.0 )
    return d - 360.0;
  else
    return d;

}

/* Scoring function for ActPos actions */
void ScoringActionVisitor::visit(ActPos *)
{
  /* This isn't implemented yet

     Future consideration: if we consider region as a set of discretized
     points, maybe SPAR and other positioning algorithm can be integrated
     into this scoring function */
  score = 0.0;
}

/* Scoring function for ActHome actions */
void ScoringActionVisitor::visit(ActHome *)
{
  /* This isn't implemented yet as well. May not make sense at all */
}

/* This action is obsolete, so ignore */
void ScoringActionVisitor::visit(ActBallToReg *)
{
}

/* Same with this action */
void ScoringActionVisitor::visit(ActBallToPlayer *)
{
}

/* Scoring function for ActMark actions */
void ScoringActionVisitor::visit(ActMark *act)
{
  /* Very simple heuristic: the score is the inverse function of the distance
     from this player to the opponent player to be marked */
  /* jdavin: modified to also account for the opponent's distance from our goal. 
   * opponents closer to our goal will get higher scores since they're more important 
   * to mark 
   * Score is now = 5./goalDist +  1./dist, because I want to make distance to goal more 
   * significant than distance to player. 
   */ 

  /* Retrieve the UNum of the opponent player */
  UNum::unum_t opp = (act->getPlayers()).getIndexAt(0).getUNum();

  // but if we're not confident about the opponent position, give it a low score
  if( ! MyWorldModel.playerConfidenceGood(opp, false)){
    score = 0;
    return;
  }

  /* Retrieve the position of the agent */
  VecPosition agentPos = MyWorldModel.getAgentGlobalPosition();

  /* Retrieve the position of the opponent player */
  VecPosition oppPos = MyWorldModel.getGlobalPosition( opp, false );


  /* Calculate the distance from the agent to the opponent */
  double dist = agentPos.getDistanceTo(oppPos);
  /* Calculate the x distance from the opponent to our goal */
  VecPosition posMyGoal = ( MyWorldModel.TeamIsLeftSide() )
    ? SoccerTypes::getGlobalPositionFlag(OBJECT_GOAL_L, SIDE_LEFT )
    : SoccerTypes::getGlobalPositionFlag(OBJECT_GOAL_R, SIDE_RIGHT);
  double goalDist = oppPos.getX() - posMyGoal.getX();

  //cout <<"Opponent "<<opp<< " dist to goal is "<<goalDist<<" and dist to us is "<<dist<<endl;

  /* If distance is very close, then arbitrarily set it to a large value, also
     preventing a division by 0 */
  if(dist < 0.000000001 || goalDist < 0.000000001)
    score = 99999.0;
  
  /* Else, the score is the inverse of the distance and goalDist */
  else
    score = 5./goalDist +  1./dist;
}

/* Scoring function for ActMarkLinePlayer actions */
void ScoringActionVisitor::visit(ActMarkLinePlayer *act)
{
  /* Heuristic function:

     The inverse of the distance from the agent to the projected agent point
     on the line connecting the ball to the opponent player
  */

  /* Construct the line connecting the ball to the opponent player */

  UNum::unum_t oppUNum = (act->getPlayers()).getIndexAt(0).getUNum();
  VecPosition oppPos = MyWorldModel.getGlobalPosition( oppUNum, false );
  VecPosition ballPos = MyWorldModel.getBallPos();

  Line l = Line::makeLineFromTwoPoints(oppPos, ballPos);

  /* Calculate the distance from the agent to this line */
  double dist = l.getDistanceWithPoint(MyWorldModel.getAgentGlobalPosition());

  
  /* If distance is very close, then arbitrarily set it to a large value, also
     preventing a division by 0 */
  if(dist < 0.000000001)
    score = 99999.0;
  
  /* Else, the score is the inverse of the distance */
  else
    score = 1.0 / dist; 
}

/* This action will be reduced to ActMarkLinePlayer, so ignore */
void ScoringActionVisitor::visit(ActMarkLineReg *)
{
}

/* Scoring function for ActOffsidesLine actions */
void ScoringActionVisitor::visit(ActOffsidesLine *)
{
  /* Not sure if one offside line is better than the other */
}

/* This is an obsolete action, so ignore */
void ScoringActionVisitor::visit(ActHetType *)
{
}

/* Ignore named action */
void ScoringActionVisitor::visit(ActNamed *)
{
}

/* Same thing with ActMarkLineReg actions, this will be reduced to
   its UNUMSet equivalent actions, so ignore */
void ScoringActionVisitor::visit(ActPassReg *)
{
}

/* Scoring function for ActPassUNum actions */
void ScoringActionVisitor::visit(ActPassUNum *act)
{
  /* The heuristic used:

     The expected degree of success is calculated by looking at the line
     connecting the current position of the ball to the location of the
     passee. Then, the opponent players around that line will be considered
     in terms of the distance from the ball and the distance from the line
     connecting the ball to the passee
  */
  

  /* Retrieve the sole UNum in the UNumSet */
  UNum::unum_t u = (act->getPlayers()).getIndexAt(0).getUNum();
   

  /* Get the current position of the passee */
  VecPosition passee = MyWorldModel.getGlobalPosition(u, true);


  /* next, avoid passing to a passee who is near our goal and also closer
     to our goal to prevent dangerous situation (within 20 m), if the advice 
     is not from the coach. */
  if( sourceType != ActionBinding::COACH &&
      passee.getX() < MyWorldModel.getAgentGlobalPosition().getX() &&
      passee.getDistanceTo( MyWorldModel.getPosOwnGoal() ) < 20.0 ) {

    score = -1.0;

    return;
  }


  /* Retrieve the current ball position */
  VecPosition ball = MyWorldModel.getBallPos();

  /* Get the vector from the ball to the passee */
  VecPosition ballToPassee = passee - ball;

  /* Distance from the ball to the passee */
  double distFromBallToPassee = ballToPassee.getMagnitude();


  /* Check for division by 0 in case of really short distance */
  /* jdavin
if(distFromBallToPassee < 0.0001)
    score = 10000.0;
  else
    score = 1.0 / distFromBallToPassee;
  */

  // very short passes are of little value, so don't waste time on them
  if(distFromBallToPassee < 5.) 
    score = 0.0; // -1. 
  else
    score = 1.0 / distFromBallToPassee;


  // but if this player has told us he wants a pass, give him a higher initial score
  if(trilearn_wm->getWantPassPlayer() == MyWorldModel.getObjectTFromUNum(u, true)){
    // only use if the player requested a pass within last 10 cycles
    if(trilearn_wm->getWantPassTime() >= MyWorldModel.getCurrentTime() - 10){
      score = 10.0;
    }
  }



  /* Calculate the direction from the ball to the passee */
  double angleDir = ballToPassee.getDirection();
  
  
  /* If this is a forward pass, */
  if( angleDir > -90.0 && angleDir < 90.0 ) {

    /* Calculate the lower and upper angles */
    double lowerAngle = angleDir - 90.0;
    double upperAngle = angleDir + 90.0;


    /* Consider all opponents in the interval 
       [angleDir - 90.0,angleDir + 90.0] */
    for(UNum u = UNum::u1; u.getUNum() <= UNum::u11; u++)   {
      
      /* Retrieve the UNum of the current teammate */
      UNum::unum_t currUNum = u.getUNum();
    
    
      /* Consider opponents with high confidence value at all times */
      if( !MyWorldModel.playerConfidenceGood(currUNum, false) )
	continue;

    
      /* Get the current position of the opponent */
      VecPosition opp = MyWorldModel.getGlobalPosition(currUNum, false);

    
      /* Now, if the angle of the direction of the vector from the ball to
	 the opponent player is not within the range, don't do more comptuation
      */
      double ballToOppDirection = (opp - ball).getDirection();


      if(ballToOppDirection <= lowerAngle || ballToOppDirection >= upperAngle)
	continue;


      /* Calculate the value of d, the distance from the ball to the projected
	 position of the opponent on the ballToPassee line */
      Line ballToPasseeLine = Line::makeLineFromTwoPoints(ball, passee);
      double d = ball.getDistanceTo( ballToPasseeLine.getPointOnLineClosestTo
				     (opp) );
    
      /* Calculate the value of w, the distance from the opponent to the
	 its projected position on the line */
      double w = ballToPasseeLine.getDistanceWithPoint(opp);
      
    
      /* If either d or w is close to 0, then we have a bad situation, so give
	 a score of 0. */
      if( d < 0.1 || w < 0.1 ) {
	
	score = 0.0;
	return;
      }
      else
	score *= (w / d);
      
    }
  }

  /* Otherwise, this is a backward pass, we work with denormalized angles */
  else {
    
    // don't do very long backpasses
    if(distFromBallToPassee > 10. || distFromBallToPassee < 1.) 
      score = 0.; //-1.;   
    // jdavin: don't make it -1 b/c then coach will not be able to do
    // backpasses in setplays. 

    // also decrease score for backpasses (so score is now 1/distFromBallToPassee^2)             
    else
      score = score / distFromBallToPassee;


    double lowerAngle = denormalize( angleDir ) - 90.0;
    double upperAngle = denormalize( angleDir ) + 90.0;


    /* Consider all opponents in the interval 
       [angleDir - 90.0,angleDir + 90.0] */
    for(UNum u = UNum::u1; u.getUNum() <= UNum::u11; u++)   {
      
      /* Retrieve the UNum of the current teammate */
      UNum::unum_t currUNum = u.getUNum();
    
    
      /* Consider opponents with high confidence value at all times */
      if( !MyWorldModel.playerConfidenceGood(currUNum, false) )
	continue;

    
      /* Retrieve the position of the opponent */
      VecPosition opp = MyWorldModel.getGlobalPosition(currUNum, false);

    
      /* Now, if the angle of the direction of the vector from the ball to
	 the opponent player is not within the range, don't do more comptuation
      */
      double ballToOppDirection = denormalize( (opp - ball).getDirection() );


      if(ballToOppDirection <= lowerAngle || ballToOppDirection >= upperAngle)
	continue;


      /* Calculate the value of d, the distance from the ball to the projected
	 position of the opponent on the ballToPassee line */
      Line ballToPasseeLine = Line::makeLineFromTwoPoints(ball, passee);
      double d = ball.getDistanceTo( ballToPasseeLine.getPointOnLineClosestTo
				     (opp) );
    
      /* Calculate the value of w, the distance from the opponent to the
	 its projected position on the line */
      double w = ballToPasseeLine.getDistanceWithPoint(opp);
      
    
      /* If either d or w is close to 0, then we have a bad situation, so give
	 a score of 0. */
      if( d < 0.1 || w < 0.1 ) {
	
	score = 0.0;
	return;
      }
      else
	score *= (w / d);
      
    }

  }

}

/* Scoring function for ActDribble actions */
void ScoringActionVisitor::visit(ActDribble *)
{
  /* Maybe this isn't necessary since all the directions will be considered
     in the Coachable player's dribbling code */
}

/* Scoring function for ActClear actions */
void ScoringActionVisitor::visit(ActClear *)
{
  /* Same thing with Dribbling actions, so may not be necessary. Not sure
     How to compare two different clearing actions since ActClear is on
     regions anyways */
}

/* Scoring function for ActShoot actions */
void ScoringActionVisitor::visit(ActShoot *)
{
  /* Not necessary at all, since the player code will decide which direction
     to shoot */
}

/* Scoring function for ActHold actions */
void ScoringActionVisitor::visit(ActHold *)
{
  /* All holding actions are equal, so not necessary at all */
}

/* Scoring function for ActIntercept actions */
void ScoringActionVisitor::visit(ActIntercept *)
{
  /* Not necessary at all */
}

/* Scoring function for ActTackle actions */
void ScoringActionVisitor::visit(ActTackle *act)
{
  /* Very simple heuristic: the score is the inverse function of the distance
     from the ball to the opponent player to be marked */
  
  /* Retrieve the UNum of the opponent player */
  UNum::unum_t opp = (act->getPlayers()).getIndexAt(0).getUNum();

  /* Retrieve the position of the agent */
  VecPosition agentPos = MyWorldModel.getAgentGlobalPosition();

  /* Retrieve the position of the opponent player */
  VecPosition oppPos = MyWorldModel.getGlobalPosition( opp, false );

  /* Calculate the distance from the agent to the opponent */
  double dist = agentPos.getDistanceTo(oppPos);


  /* If distance is very close, then arbitrarily set it to a large value, also
     preventing a division by 0 */
  if(dist < 0.000000001)
    score = 99999.0;
  
  /* Else, the score is the inverse of the distance */
  else
    score = 1.0 / dist;

}





