/* Dongryeol Lee
 *
 * Spar.C
 * 
 * The implementation of the SPAR dynamic positioning described in the
 * Header file Spar.h
 *
 */

#include "Spar.h"
#include "WorldModelAccessor.h"


using namespace rcss::clang;
using namespace std;


/* Reference to the player's view of the world */
extern WorldModelAccessor MyWorldModel;


/* This function will implement the evaluation function based on dynamic
 * positioning concept implemented in the paper, "Anticipation as a Key
 * for Collaboration in a Team of Agents: A Case Study in Robotic Soccer,"
 * SPAR positioning.
 * Larger values are better.
 */
double Spar::evaluatePosition(const VecPosition& pos, const UNum::unum_t& num)
{
  #ifndef WEIGHT_BALLDIST
  #define WEIGHT_BALLDIST 1.0
  #endif


  #ifndef WEIGHT_PLAYERDIST
  #define WEIGHT_PLAYERDIST 75.0
  #endif


  /* Evaluated result from the player positions */
  double playerEval = 0.0;


  /* Evaluated result from the current ball position */
  double ballEval = 0.0;


  /*
  if(pos.getX() >= MyWorldModel.getOffsideX() - 0.5)
    return -9999.0;
  */


  /* Go through our players and accumulate eval result */
  for(int i = 1; i <= 11; i++) {

    /* Check the positions of the other teammates, and calculate evaluation */
    if((UNum::unum_t) i != num &&
       MyWorldModel.playerConfidenceGood( (UNum::unum_t) i, true )) {
      
      /* Get the teammate position */
      VecPosition teammatePos = 
	MyWorldModel.getGlobalPosition((UNum::unum_t) i, true);


      /* Calculate the distance from the position to the teammate */
      double posToTeammate = pos.getDistanceTo( teammatePos );

      
      /* Add onto our evaluated result */
      playerEval += 1.0 / (posToTeammate * posToTeammate);

    }
  }

  /* Go through the opponent players and accumulate eval result */
  for(int i = 1; i <= 11; i++) {
    
    if( MyWorldModel.playerConfidenceGood( (UNum::unum_t) i, false ) ) {

      /* Get the opponent's position */
      VecPosition oppmatePos = 
	MyWorldModel.getGlobalPosition((UNum::unum_t) i, false);
      
      
      /* Calculate the distance from the position to the opponent */
      double posToOpponent = pos.getDistanceTo( oppmatePos );

      
      /* Add onto our evaluated result */
      playerEval += 1.0 / (posToOpponent * posToOpponent);
    }
  }
  

  /* Get distance from the position to the ball and subtract */
  ballEval = (MyWorldModel.predictBallPosAfterNrCycles(1))
    .getDistanceTo(pos);


  /* Return our evaluated resullt */
  return 1.0 / (WEIGHT_BALLDIST * ballEval + WEIGHT_PLAYERDIST * playerEval);
}

/* This function will simply return the best point by using eval function */
VecPosition Spar::getBestPt(RegionIterator *ri, 
			    const UNum::unum_t& num, 
			    const VarBindings& bindings)
{
  VecPosition bestPt(0, 0);
  double bestPtEval = -99999;

  
  /* Iterate all the points in the region, and get the best one */
  while(ri->valid()) {
    
    double temp;
    VecPosition currPt = ri->getPt(&MyWorldModel, bindings);

    if(bestPtEval < (temp = evaluatePosition(currPt, num))) {
      bestPt = currPt;
      bestPtEval = temp;
    }

    /* Go to the next point */
    ri->moveToNextPt();
  }


  /*
  cout << "Best: " << bestPt.getX() << " " << bestPt.getY() <<
    " " << bestPtEval << "\n";
  */


  return bestPt;
}
