/*
 * Defines WorldModel functions that are used by the coachable agent communication class
 *  (CoachableAgentComm) to update information about the ball,players, etc. 
 *
 * Creation Date: 12/14/03
 * Author: John Davin
 */



#include "WorldModel.h"
#include "config.h"


/*
 * Convenience functions that call updatePlayerPos
 *
 */
bool WorldModel::updateTeammatePos(int playernum, float X, float Y, unsigned short cycles){

  return updatePlayerPos(playernum, X, Y, cycles, OUR_TEAM);
}

bool WorldModel::updateOpponentPos(int playernum, float X, float Y, unsigned short cycles){

  return updatePlayerPos(playernum, X, Y, cycles, OPPONENT_TEAM);
}


/*
 * Update the world model absolute position of player indicated by playernum. 
 * only update if the cycles is lower (more current) than the number of cycles 
 * since our current info on this player was updated. 
 * Returns true if the update succeeds, false o/w. 
 */
bool WorldModel::updatePlayerPos(int playernum, float X, float Y, 
				 unsigned short cycles, TeamT team){

  if( playernum == getPlayerNumber() ){ // don't update our own position
    return false; 
  }

  ObjectT playertype;
  if(team == OUR_TEAM)
    playertype = SoccerTypes::getTeammateObjectFromIndex(playernum-1); 
  else // team == OPPONENT_TEAM
    playertype = SoccerTypes::getOpponentObjectFromIndex(playernum-1);
  
  PlayerObject *object = (PlayerObject *)getObjectPtrFromType( playertype );
  if( object == NULL ){
    cerr <<"[updatePlayerPos]: error: "<<playernum<<" was not a known player"<<endl;
    return false;
  }

  Time time = getCurrentTime() - cycles;

  // update the position if the player is unknown (because this pos will provide better info)
  // or if it was last updated before the time of this X,Y data. 
  if( object->getTimeLastSeen() < time ||
      object->getIsKnownPlayer() == false  ){

    object->setGlobalPosition     ( VecPosition(X, Y) , time );
    object->setTimeLastSeen       ( time );
    object->setIsKnownPlayer      ( true );
    
    updateObjectRelativeFromGlobal( playertype );

    LOGOUT(51) << "[updatePlayerPos]: "<<getPlayerNumber()<<": updated player "<<playernum
	 <<" to pos "<<X<<","<<Y<<endl;

    return true;
  }
  return false;

}


/*
 * Update our ball position only if the information received is more current
 * than the last time we saw the ball. 
 *
 * Returns true if the update succeeds, false o/w. 
 */
bool WorldModel::updateBallPos(float X, float Y, unsigned short cycles){

  Time updatetime = getCurrentTime() - cycles;

  // if the ball was last seen before 'cycles' cycles ago. 
  if( Ball.getTimeLastSeen() < updatetime )
    {
      Ball.setGlobalPosition( VecPosition(X,Y), updatetime );
      Ball.setTimeLastSeen  ( updatetime );
      updateObjectRelativeFromGlobal( OBJECT_BALL );

      LOGOUT(51) << "[updateBallPos]: "<<getPlayerNumber()<<": updated ball"
	 <<" to pos "<<X<<","<<Y<<endl;

      return true;
    }
  return false;

}


/*
 * Update our ball velocity only if the information received is more current
 * than the last time we saw the ball. 
 *
 * Returns true if the update succeeds, false o/w. 
 */
bool WorldModel::updateBallVel(float dX, float dY, unsigned short cycles){

  Time updatetime = getCurrentTime() - cycles;

  // if the ball was last seen before 'cycles' cycles ago. 
  if( Ball.getTimeLastSeen() < updatetime )
    {
      Ball.setGlobalVelocity( VecPosition(dX,dY), updatetime );
      Ball.setTimeLastSeen  ( updatetime );
      updateObjectRelativeFromGlobal( OBJECT_BALL );
      return true;
    }
  return false;

}

/*
 * Update the ball owner player and the team variable. 
 *
 * Returns true if the update succeeds, false o/w. 
 */
bool WorldModel::updateBallOwner(unsigned short playernum, TeamT team){


  // only update BallOwner if playernum != 0 (if playernum = 0, that means 
  // that the sending player only knew which team owns the ball, not the player#).
  if(playernum != 0){

    ObjectT playertype;
    if(team == OUR_TEAM)
      playertype = SoccerTypes::getTeammateObjectFromIndex(playernum-1); 
    else
      playertype = SoccerTypes::getOpponentObjectFromIndex(playernum-1);
    
    BallOwner = playertype;
  }
  else{ // playernum == 0
    if(team == OUR_TEAM)
      BallOwner = OBJECT_TEAMMATE_UNKNOWN;
    else // team == OPPONENT_TEAM
      BallOwner = OBJECT_OPPONENT_UNKNOWN;
  }

  return true;
}



/*
 * Parameter 'playernum' is a number from 1 to 11 indicating the player
 *  that wants a pass. 
 * Returns true if update succeeds, false o/w
 */
bool WorldModel::updateWantPassPlayer(unsigned short playernum){

  assert(playernum >= 1 && playernum <= 11);
  assert(playernum != getPlayerNumber()); // it cannot be ourself

  wantPassPlayer = SoccerTypes::getTeammateObjectFromIndex(playernum-1);
  wantPassTime = getCurrentTime();
  return true;
}

/*
 * Return object of the teammate who wants a pass
 */
ObjectT WorldModel::getWantPassPlayer(){
  return wantPassPlayer;
}

/*
 * Returns the wantPassTime, which tells us when wantPassPlayer was 
 * last updated. (wantPassTime == -1 if it's not set yet)
 */
Time WorldModel::getWantPassTime(){
  return wantPassTime;
}


/*****
 * methods for setting/accessing info on who is passing to us
 *****/

/*
 * Parameter 'passernum' is the number of the player who is passing
 * to us. Range is [1,11]. 
 * Return true if update succeeded, false o/w.
 */
bool WorldModel::updatePassToPlayer(int passernum){
  assert(passernum >= 1 && passernum <= 11);

  PassToPlayerSender = SoccerTypes::getTeammateObjectFromIndex(passernum-1);
  PassToPlayerTime   = getCurrentTime();

  return true;
}

/*
 * Return the object type of player who is passing to us. 
 */
ObjectT WorldModel::getPassToPlayerSender() const {
  return PassToPlayerSender;
}

/*
 * Return time PassToPlayer info was last updated
 */
Time  WorldModel::getPassToPlayerTime() const {
  return PassToPlayerTime;
}

/*
 * Update the PassToCoord info. 
 * Parameter 'passernum' is the number of the teammate (range [1,11]) 
 * that is passing to coordinate 'X','Y'.
 * Return true if update succeeded, false o/w.
 */
bool WorldModel::updatePassToCoord(int passernum, float X, float Y) {
  assert(passernum >= 1 && passernum <= 11);

  PassToCoordSender = SoccerTypes::getTeammateObjectFromIndex(passernum-1);
  PassToCoord = VecPosition(X,Y);
  PassToCoordTime = getCurrentTime();

  return true;
}

/*
 * Return the coordinate our teammate is passing to
 */
VecPosition WorldModel::getPassToCoord() const {
  return PassToCoord;
}

/*
 * Return object of player who is passing to the coordinate
 */
ObjectT WorldModel::getPassToCoordSender() const {
  return PassToCoordSender;

}

/*
 * Return the time that PassToCoord info was updated
 */
Time  WorldModel::getPassToCoordTime() const {
  return PassToCoordTime;
}



