/* Dongryeol Lee
 *
 * WorldStatePredicate.h
 *
 * This is the header file for the WorldStatePredicate class. Each object
 * represents a predicate that pertains to the current world state. This class
 * is provided as a virtual base class for other world-state predicates to
 * extend from.
 *
 */


#ifndef WORLD_STATE_PREDICATE
#define WORLD_STATE_PREDICATE


#include "WorldStatePredicate.h"
#include "WorldModelAccessor.h"
#include "Geometry.h"


#include "Logger.h"


/* Our access to the player's world model */
extern WorldModelAccessor MyWorldModel;



/* Our virtual base class for WorldStatePredicate objects */
class WorldStatePredicate
{
 public:

  /* Constructor */
  WorldStatePredicate(bool positive) {  positiveFlag = positive;  }
  
  /* Virtual destructor */
  virtual ~WorldStatePredicate() {}


  /* Virtual boolean function for other classes to implement */
  virtual bool isSatisfiable() const = 0;
  

  /* Return whether the current predicate is positive or negative */
  bool getPositiveFlag() const { return positiveFlag; }


 private:
  
  /* Tells whether the Predicate P represents P or ~P. 
     True if the Predicate represents P
     False if the Predicate represents ~P
  */
  bool positiveFlag;
};





/* Some world-state properties extending from the virtual base class... */


/* IsClosestToBall: True if the agent is the closest to the ball among 
   its team */
class IsClosestToBall: public WorldStatePredicate
{
 public:

  /* Constructor */
  IsClosestToBall(bool positive): WorldStatePredicate(positive) {}

  
  bool isSatisfiable() const
    {
      using namespace rcss::clang;
      
      /* First, save the expensive computation by checking the ball
	 confidence. If it is low, then return false (Since the player would
         have to have the good confidence in the ball location, for closeness 
      */
      if(!MyWorldModel.ballConfidenceGood())
	return false;


      /* Retrieve the current ball position */
      VecPosition ballPos = MyWorldModel.getBallPos();

      /* Closest distance found so far */
      double closestDist = 99999;
      
      /* UNum of the closest teammate */
      UNum::unum_t closestUNum;

      /* Position of the current teammate in consideration */
      VecPosition pos;

      /* Distance from the current teammate to the ball */
      double dist;


      /* Go through each teammate in our team */
      for(UNum u = UNum::u1; u.getUNum() <= UNum::u11; u++){

	/* Retrieve the UNum of the current teammate */
	UNum::unum_t currUNum = u.getUNum();


	/* Check the confidence of the location of the teammates we want to
	   check first.... If it is low, then don't bother with computation */
	if( !MyWorldModel.playerConfidenceGood(currUNum, true) )
	  continue;


	/* Retrieve the position of the teammate */
	pos = MyWorldModel.getGlobalPosition(currUNum, true);

	/* Calculate the distance from the ball to the player */
	dist = pos.getDistanceTo(ballPos);

	
	/* Replace the result if better than previous */
	if(dist < closestDist) {

	  closestDist = dist;
	  closestUNum = currUNum;
	}

      }   /* End of for-loop */

      
      //actionlog(50) << "IsClosestToBall: closest is " << closestUNum << " at dist " << closestDist << spades::ende;
      /* Get the boolean result of the positive
	 Just check the UNum of the closest player equals the agent UNum */
      return MyWorldModel.getMyNum() == closestUNum;

    }
};


/* IsNearOpponent: True if the opponent player is within the distance
   threshold, false otherwise */
class IsNearOpponent: public WorldStatePredicate
{
 public:


  IsNearOpponent(bool positive): WorldStatePredicate(positive) {}


  bool isSatisfiable() const
    {

      /* Define the distance as a macro */
      #ifndef NEAR_OPPONENT_DIST
      #define NEAR_OPPONENT_DIST    3.0
      #endif


      using namespace rcss::clang;


      /* Current position of the opponent in consideration */
      VecPosition pos;
      
      /* distance to the opponent in consideration */
      double dist;

      /* First, retrieve the position of the agent */
      VecPosition agentPos = MyWorldModel.getAgentGlobalPosition();

      
      /* Go through each opponent in the opponent team */
      for(UNum u = UNum::u1; u.getUNum() <= UNum::u11; u++){
	

	/* Retrieve the UNum of the current opponent */
	UNum::unum_t currUNum = u.getUNum();


	/* Check the confidence of the location of the opponent we want to
	   check first.... If it is low, then don't bother with computation */
	if( !MyWorldModel.playerConfidenceGood(currUNum, false) )
	  continue;


	/* Retrieve the position of the opponent */
	pos = MyWorldModel.getGlobalPosition(currUNum, false);

	/* Calculate the distance from the opponent to the agent */
	dist = pos.getDistanceTo(agentPos);


	/* Return true if any opponent is observed with in the threshold
	   distance */
	if(dist < NEAR_OPPONENT_DIST)
	  return true;

      }   /* End of for-loop */      
      

      /* If you are here, no opponents have been found with "confidence" in
	 near area, so return false */
      return false;
    }

};


/* IsCloseToOpponentGoal: True if the agent is witin the threshold to the
   opponent goal */
class IsCloseToOpponentGoal: public WorldStatePredicate
{
  
 public:

  
  IsCloseToOpponentGoal(bool positive): WorldStatePredicate(positive) {}


  bool isSatisfiable() const
    {
      
      /* Also, define the threshold distance as a macro */
      #ifndef NEAR_OPPGOAL
      #define NEAR_OPPGOAL   15.0
      #endif
      
      /* Retrieve the opponent goal position */
      VecPosition oppgoalPos = MyWorldModel.getPosOpponentGoal();

      /* Retrieve the agent position */
      VecPosition agentPos = MyWorldModel.getAgentGlobalPosition();


      /* Calculate the distance from the agent to the opponent goal
	 position */
      double dist = agentPos.getDistanceTo(oppgoalPos);

      
      /* Check if the distance is within the threshold */
      return dist < NEAR_OPPGOAL;
    }
};




/* IsCloseToOurGoal: True if the agent is witin the threshold to the
   our goal */
class IsCloseToOurGoal: public WorldStatePredicate
{
  
 public:

  
  IsCloseToOurGoal(bool positive): WorldStatePredicate(positive) {}


  bool isSatisfiable() const
    {
      
      /* Also, define the threshold distance as a macro */
      #ifndef NEAR_OURGOAL
      #define NEAR_OURGOAL   20.0
      #endif
      
      /* Retrieve the opponent goal position */
      VecPosition ourgoalPos = MyWorldModel.getPosOwnGoal();

      /* Retrieve the agent position */
      VecPosition agentPos = MyWorldModel.getAgentGlobalPosition();


      /* Calculate the distance from the agent to the opponent goal
	 position */
      double dist = agentPos.getDistanceTo(ourgoalPos);

      
      /* Check if the distance is within the threshold */
      return dist < NEAR_OURGOAL;
    }
};



/* IsBallOwnedByOurTeam: True if our team owns the ball */
class IsBallOwnedByOurTeam: public WorldStatePredicate
{
 public:
  
  IsBallOwnedByOurTeam(bool positive): WorldStatePredicate(positive) { }

  bool isSatisfiable() const
    {
      /* Ball owner */
      rcss::clang::UNum::unum_t *ballOwner;

      /* Return true if the ball is owned by our team */
      return MyWorldModel.getBallOwner(ballOwner);
    }
};


/* IsPlayOnMode: True if the world state is on play-on mode */
class IsPlayOnMode: public WorldStatePredicate
{
 public:
  
  IsPlayOnMode(bool positive): WorldStatePredicate(positive) { }

  bool isSatisfiable() const
    {
      return (  MyWorldModel.getPlayMode() == rcss::clang::PlayOn  );
    }
};

/* IsBallKickable: True if the ball is kickable */
class IsBallKickable: public WorldStatePredicate
{
 public:
  
  IsBallKickable(bool positive): WorldStatePredicate(positive) { }

  bool isSatisfiable() const
    {
      return (  MyWorldModel.isBallKickable());
    }
};

/* FastestToBall: True if the agent can get to the ball the fastest of all */
class FastestToBall: public WorldStatePredicate
{
 public:

  FastestToBall(bool positive): WorldStatePredicate(positive) {}

  bool isSatisfiable() const
    {
      return MyWorldModel.FastestToBall();
    }
};



/* IsNearBall: True if the ball is within 5m */
class IsNearBall: public WorldStatePredicate
{
      /* Also, define the threshold distance as a macro */
      #ifndef NEAR_BALL
      #define NEAR_BALL   5.0
      #endif

 public:
  
  IsNearBall(bool positive): WorldStatePredicate(positive) { }

  bool isSatisfiable() const
    {
      /* First, we need to check the confidence level of the ball. If it is
	 low, then the agent obviously isn't anywhere near the ball */
      if( !MyWorldModel.ballConfidenceGood() )
	return false;


      /* Retrieve the agent position */
      VecPosition agentPos = MyWorldModel.getAgentGlobalPosition();
      
      /* Retrieve the current ball position */
      VecPosition ballPos = MyWorldModel.getBallPos();

      /* Calculate the distance from the agent to the bal */
      double dist = agentPos.getDistanceTo(ballPos);


      //actionlog(40) << "Distance: " << dist << "\n\n";

      return ( dist < NEAR_BALL);
    }
};


/* IsNearBall: True if the ball is within 5m */
class IsNearBallForIntercept: public WorldStatePredicate
{
      /* Also, define the threshold distance as a macro */
      #ifndef NEAR_BALL_INTERCEPT
      #define NEAR_BALL_INTERCEPT   3.0
      #endif

 public:
  
  IsNearBallForIntercept(bool positive): WorldStatePredicate(positive) { }

  bool isSatisfiable() const
    {
      /* First, we need to check the confidence level of the ball. If it is
	 low, then the agent obviously isn't anywhere near the ball */
      if( !MyWorldModel.ballConfidenceGood() )
	return false;


      /* Retrieve the agent position */
      VecPosition agentPos = MyWorldModel.getAgentGlobalPosition();
      
      /* Retrieve the current ball position */
      VecPosition ballPos = MyWorldModel.getBallPos();

      /* Calculate the distance from the agent to the bal */
      double dist = agentPos.getDistanceTo(ballPos);


      //actionlog(40) << "Distance: " << dist << "\n\n";

      return ( dist < NEAR_BALL_INTERCEPT);
    }
};



/* IsBallNearOurGoal: True if the ball is near our goal */
class IsBallNearOurGoal: public WorldStatePredicate
{
      /* Also, define the threshold distance as a macro */
      #ifndef BALL_NEAR_OUR_GOAL
      #define BALL_NEAR_OUR_GOAL    20.0
      #endif

 public:
  
  IsBallNearOurGoal(bool positive): WorldStatePredicate(positive) { }


  bool isSatisfiable() const
    {
      /* First, we need to check the confidence level of the ball. If it is
	 low, then the ball isn't anywhere near our goal */
      if( !MyWorldModel.ballConfidenceGood() )
	return false;
      

      /* Retrieve ball position */
      VecPosition ballPos = MyWorldModel.getBallPos();


      /* Retrieve the position of our goal */
      VecPosition ourGoal = MyWorldModel.getPosOwnGoal();


      /* Calculate the distance from the agent to the bal */
      double dist = ballPos.getDistanceTo(ourGoal);


      return ( dist < BALL_NEAR_OUR_GOAL );
    }
};


#endif
