/*
 * Here we define functions for Rule messages
 * For example, there are functions for turning on/off rules
 *
 * John Davin
 */

#include <iostream.h>
#include <ostream.h>
#include <unistd.h>
#include <netinet/in.h>
#include "msgrulestore.h"
#include "RuleVisitor.h"
#include "CondVisitor.h"
#include "DirVisitor.h"
#include "ActionPicker.h"
#include "WorldModelAccessor.h"
#include "Filter.h"
#include "LevelStorage.h"


// the ActionPicker we'll use to store the action list
Filter filter;
LevelStorage levels;
ActionPicker action_picker;

extern WorldModelAccessor MyWorldModel; // from CondVisitor.C
extern WorldModel *trilearn_wm; // this comes from PlayerInterface.C


bool MsgRuleStore::add(const string &msgstr, const RulePair &rule){

  /* Get the pointer to the Rule object */
  Rule *rptr = rule.ptr;

  if(rptr == NULL){
    /* PR(9, "[MsgStore:add]: ERROR: inserting null pointer for name %s\n", 
       msgstr.c_str());*/
    cout << "[MsgStore:add]: ERROR: inserting null pointer for name %s" <<
      msgstr.c_str() << "\n";
    exit(1);
  }

  // but if a message with this name already exists, then we're overwriting a 
  // rule which may be in the onrules list (and therefore its Rule* pointer 
  // is no good anymore).
  MsgRuleStore::RuleStateIter it;
  RulePair searchpair;
  bool exists = true;
  
  searchpair.rulename = msgstr;
  searchpair.ptr = rptr;
  searchpair.sourceType = rule.sourceType;

  if((it = onrules.find(searchpair)) != onrules.end())  {

    // found rule in onrules
    delete (*it).ptr; // free the memory for the previous Rule. 
    // we do this case because it is a const_iterator (since we're dealing with a set)
    // however, we know that we can update without changing the order in the set
    // because we are not changing the name
    (RulePair&)(*it) = searchpair; // update with the pointer to our new Rule
    exists = true;
  } else {
    /* Before adding to the storage, we need to check if the rule exists in
     the storage... */
    RulePair ruletest;

    exists = lookup(msgstr, &ruletest);

    /* If it does exist, delete the Rule pointer inside the RulePair object */
    if(exists){
      delete (ruletest.ptr);
    }
  }

  //std::cout << "[MsgRuleStore]: added rule " << msgstr << " to map" << endl;
  /* Insert the rule into our message store */
  messages[msgstr] = searchpair;

  /* True if the rule existed in the storage already, false otherwise */
  return exists;
}

/*
 * delete a rule. Returns number of entries deleted (1 if succeeded, 0 if 
 * failed)
 */
int MsgRuleStore::remove(const std::string &name){

  bool success;
  
  if(QueryState(name) != NULL){ // rule is in onrules. 

    // remove it from onrules
    success = TurnOff(name);
    
   /*
    if(!success) 
      PR(9, 
      "[MsgRuleStore:delete]: ERROR! failed to remove rule from onrules\n");
    */

  }

  // now delete it from the messages store
  return messages.erase(name);

}


/*
 * Turn on a rule
 * Input: Rule name as a string
 * Return: true on success, 
 *         false on error (if name does not exist in our list of rules)
 */
bool MsgRuleStore::TurnOn(const std::string &name){

  //PR(10, "[MsgRuleStore]: Turning on rule %s\n", name.c_str());

  RulePair rulepair;
  MsgIterator it;

  rulepair.rulename = name; 
  rulepair.ptr = NULL;

  // check if name is not present in messages
  if( (it = messages.find(name)) == messages.end() ){ // rule does not exist
    //PR(9, "[MsgRuleStore]: Rule not found: %s\n", name.c_str());
    return false;
  }

  rulepair = (*it).second;
  /*
  rulepair.rulename = name;
  rulepair.ptr = (*it).second;
  rulepair.priority = (*it).third;
  */

  onrules.insert(rulepair);
  return true;
}

/*
 * Turn off a rule.
 * Input: Rule name as a string
 * Return: true on success, 
 *         false on error (if name does not exist in our list of rules)
 */
bool MsgRuleStore::TurnOff(const std::string &name){

  RulePair rulepair;

  //PR(10, "[MsgRuleStore]: Turning off rule %s\n", name.c_str());
  if(messages.find(name) == messages.end()){ // rule does not exist
    //PR(10, "[MsgRuleStore]: Rule not found: %s\n", name.c_str());
    return false;
  }

  rulepair.rulename = name;
  rulepair.ptr = NULL;

  onrules.erase(rulepair);
  return true;
}

/*
 * Turn on all rules (insert them into onrules list)
 *
 */
void MsgRuleStore::TurnOnAll(){
  RulePair rulepair;
  MsgIterator it = messages.begin();
  
  while(it != messages.end()){
    //PR(10, "[MsgRuleStore]: Turning on rule %s", (*it).first.c_str());
    
    rulepair = (*it).second;
    onrules.insert(rulepair);

    it++;
  }
}

/*
 * Turn off all rules (remove all from onrules list)
 *
 */
void MsgRuleStore::TurnOffAll(){

  onrules.clear();
  //PR(10, "[MsgRuleStore]: Turned off all rules\n");
}



/*
 * Input: Rule name as a string
 * Return: Pointer to rule if it is on,
 *         NULL if rule is off or rule does not exist
 */
Rule *MsgRuleStore::QueryState(const std::string &name)  {

  RuleStateIter it;
  RulePair searchpair;

  searchpair.rulename = name;
  searchpair.ptr = NULL;

  if((it = onrules.find(searchpair)) == onrules.end()) // did not find rule
    return NULL;

  // if we get here, we found the rule in the onrules list
  return (*it).ptr;

}


/*
 * Consults the world state model to see if any of the 'on' rules
 * are applicable. If so, carry out those actions (call player commands)
 *
 * Additional notes: 
 * Each round, the player will call EvaluateRules (in msgrulestore.C) which 
 * will use a RuleVisitor to evaluate all applicable directives (ie, *only* 
 * directives that had true conditions). Those directives will visit a 
 * DirVisitor, which will place their actions into the ActionPicker. 
 *
 * Before each evaluation cycle we must clear the action list to get rid of 
 * things left over from last round. 
 *
 * Input: none
 * Return: none
 */
void MsgRuleStore::EvaluateRules(){
  RulePair rulepair;
  //Timing evaltimer;
  //evaltimer.restartTime();
  //cout << "Evaluating "<<onrules.size()<<" rules of "<<messages.size()<<" total"<<endl;
  //cout << "time= "<<evaltimer.getElapsedTime()<<endl;

  if(onrules.size() == 0)
    {MY_PLAYER->executeCommand(NULL); return; } // do default action

  action_picker.clear();
  filter.clear();

  RuleStateIter cur  = onrules.begin();

  while(cur != onrules.end()){
    rulepair = *cur;

    Rule *currule = rulepair.ptr; 

    /* PR(10, "[MsgRulesStore:EvaluateRules]: Evaluating rule %s\n", 
       rulepair.rulename.c_str()); */


    //// Now evaluate the rule 
    // RuleVisitor will evaluate the rule's condition and store any directives 
    // that are applicable. 
    RuleVisitor *myrulevisitor;
    
    if(rulelogging)
      myrulevisitor = new RuleVisitor(rulepair.sourceType, rulepair.rulename, rulelogging);
    else
      myrulevisitor = new RuleVisitor(rulepair.sourceType);

    if(Log.isInLogLevel(4)){ 
      *logger << "Visiting rule *******"<<rulepair.rulename<<"*********" <<endl;}

    currule->accept(*myrulevisitor);
    delete myrulevisitor;
    
    cur++;
  }

  if (trilearn_wm->getPlayerSettings()->getUseBackChannelAdvice())
    {
      std::string rule_name = readBackChannelAdvice();
      if (rule_name.length() > 0)
	{
	  RulePair rulepair;
	  if (!lookup(rule_name, &rulepair))
	    {
	      cerr << "Back Channel: I don't have a rule name '" << rule_name << "'"
		   << " " << rule_name.length() << " " << ((int)rule_name[0])
		   << endl;
	    }
	  else
	    {
	      // last one forces the rule visits to match cond as true
	      RuleVisitor myrulevisitor(rulepair.sourceType, rulepair.rulename, rulelogging, true);
	      rulepair.ptr->accept(myrulevisitor);
	    }
	}
    }

  // select and execute an action from the ActionPicker
  action_picker.pickAction();
  //  cout << "done time= "<<evaltimer.getElapsedTime()<<endl;
}




std::ostream& MsgRuleStore::print( std::ostream& o ) const
{
  o << "OnRules: \n";
 
  for( RuleStateIter i = onrules.begin(); i != onrules.end(); i++ ){
    o << (*i).rulename << endl;

  }
  return o << endl;
}

std::string
MsgRuleStore::readBackChannelAdvice()
{
  if (trilearn_wm->getCurrentTime().getTime() <= 0)
    return "";
  
  int fail_open_cnt = 0;
  int fail_cnt = 0;
  int old_cnt = 0;
  
  for (int i=0; i<13; i++)
    {
      ifstream in("/tmp/abmdp_back_channel");

      if (!in)
	{
	  fail_open_cnt++;
	  return "";
	}

      int time;
      in.read((char*)(&time), sizeof(int));
      
      if (in.fail())
	{
	  fail_cnt++;
	  continue;
	}

      time = ntohl(time);
      
      if (time < trilearn_wm->getCurrentTime().getTime())
	{
	  old_cnt++;
	  usleep(1000); // 1ms
	  continue;
	}
      else if (time > trilearn_wm->getCurrentTime().getTime())
	{
	  cerr << "Back Channel: time skew! "
	       << time << " " << trilearn_wm->getCurrentTime().getTime()
	       << endl;
	  return "";
	}
      else
	{
	  std::string rule_name;
	  in >> rule_name;
	  if (in.fail())
	    cerr << "Back Channel: rule read failed" << endl;
	  int len = rule_name.length();
	  if (rule_name[len-1] == 0)
	    rule_name.erase(len - 1, 1);
	  return rule_name;	      
	}
    }
  cerr << "Back Channel: too many iter "
       << "failopen=" << fail_open_cnt
       << ", fail=" << fail_cnt
       << ", old=" << old_cnt
       << endl;
  return "";
}

